#  Copyright (c) 2008 patrick coleman

from Exceptions import *
from geom.Point3 import Point3
from geom.Normal3 import Normal3
from mmath.Vector3 import Vector3
import math


from Exceptions import *

def _boundIndices(knots, t):
	"""returns indices to knots bounding t; they might be equal, in this case 
	t corresponds to a knot exactly"""
	priorIndex = 0
	nextIndex = 0
		
	# t before knots or at start
	if t <= knots[0]:
		return (0, 0)
		
	# t after knots or at end
	lastI = len(knots) - 1
	if t >= knots[lastI]:
		return (lastI, lastI)
		
	# test for on knots
	for i in range(1, len(knots) - 1):
		if knots[i] == t:
			return (i, i)
		
	# general check for prior index
	for i in range(1, len(knots)):
		if knots[i] > t:
			priorIndex = i - 1
			break
		
	# set nextIndex, clamp to end
	nextIndex = priorIndex + 1
	if nextIndex == len(knots):
		nextIndex -= 1
			
	return (priorIndex, nextIndex)

def bezierEval(values, t):
	t2 = t * t
	oneMinusT = 1.0 - t
	oneMinusT2 = oneMinusT * oneMinusT
		
	c0 = oneMinusT2 * oneMinusT
	c1 = 3.0 * t * oneMinusT2
	c2 = 3.0 * t2 * oneMinusT
	c3 = t2 * t
		
	return (
		c0 * values[0] + 
		c1 * values[1] + 
		c2 * values[2] + 
		c3 * values[3]
		)

class Curve(object):

	def value(self, t):
		raise AbstractMethod()

class LinearSpline(Curve):
	
	def __init__(self, values, knots = None):
		"""arbitrary pt number, defaults knots to [0, len(values) - 1]"""
		self.values = values
		if knots != None:
			self.knots = knots
		else:
			self.knots = range(len(values))
	
	def value(self, t):
		pt, nt = _boundIndices(self.knots, t)
		if pt == nt:
			return self.values[pt]
		else:
			subT = (t - self.knots[pt]) / float(self.knots[nt] - self.knots[pt])
			return self.values[pt] + subT * (self.values[nt] - self.values[pt])
		
class SampledSpline(LinearSpline):
	"""Adds numerical approximations to derivitives"""
	
	def __init__(self, values, knots = None):
		LinearSpline.__init__(self, values, knots)
		self.normals = None
		self.kNormals = None
		self.kinkNormals = None
	
	def normal(self, i):
		"Returns a convex normal vector for value i"
		if self.normals == None:
			self._computeNormals()
		return self.normals[i]
	
	def curvatureNormal(self, i):
		"""Returns a curvature-scaled vector for value i"""
		if self.kNormals == None:
			self._computeKNormals()
		return self.kNormals[i]
	
	def kinkNormal(self, i):
		"""Returns a pseudo-curvature value which will be called 'kink'.  It is a nonlinear map
		from curvature such that curvature in [0, inf) maps to kink in [0, 1]"""
		if self.kinkNormals == None:
			self._computeKinkNormals()
		return self.kinkNormals[i]
		
	def _computeNormals(self):
		self.normals = []
		
		# zero normal at first value
		if len(self.values) > 0:
			self.normals.append(Normal3(0.0, 0.0, 0.0))
		
		# general normals as interpolated convex edge normals
		for i in range(1, len(self.values) - 1):
			av = self.values[i - 1] - self.values[i]
			bv = self.values[i + 1] - self.values[i]
			al = av.length()
			bl = bv.length()
			
			if al == 0.0 or bl == 0.0:
				self.normals.append(Normal3(0.0, 0.0, 0.0))
				continue
			
			side = bv ^ av
			an = av ^ side
			bn = side ^ bv
			if an.length() > 0.0:
				an = an.normalize()
			if bn.length() > 0.0:
				bn = bn.normalize()
			
			#interpolate face normals
			t = bl / (al + bl)
			n = an.interpolate(t, bn)
			if n.length() > 0.0:
				n.normalize()
			self.normals.append(Normal3(n.x, n.y, n.z))
			
		# zero normal at end
		if len(self.values) > 1:
			self.normals.append(Normal3(0.0, 0.0, 0.0))
			
		# refine first and last normals for continuity
		# in plane with adjacent normals
		# maintain anle with edge by reflecting about a segment normal (perp to edge)
		if len(self.values) > 2:
			# start
			edge = self.values[1] - self.values[0]
			n1 = Vector3(self.normals[1].x, self.normals[1].y, self.normals[1].z)
			outOfPlane = edge ^ n1
			nSegment = outOfPlane ^  edge
			if nSegment.lengthSquared() > 0.0:
				nSegment = nSegment.normalize()
				n1ProjN = (n1 * nSegment) * nSegment
				if n1ProjN.lengthSquared() != 0.0:
					n1Refl = (2.0 * n1ProjN) - n1
					self.normals[0] = Normal3(n1Refl.x, n1Refl.y, n1Refl.z)
		
			#end
			edge = self.values[len(self.values) - 2] - self.values[len(self.values) - 1]
			nEnd = Vector3(self.normals[len(self.values) - 2].x, 
						   self.normals[len(self.values) - 2].y, 
						   self.normals[len(self.values) - 2].z)
			outOfPlane = edge ^ nEnd
			nSegment = outOfPlane ^ edge
			if nSegment.lengthSquared() > 0.0:
				nSegmnt = nSegment.normalize()
				nEndProjN = (nEnd * nSegment) * nSegment
				if nEndProjN.lengthSquared() != 0.0:
					nEndRefl = (2.0 * nEndProjN) - nEnd
					self.normals[len(self.values) - 1] = Normal3(nEndRefl.x, nEndRefl.y, nEndRefl.z)

	def _computeKNormals(self):
		if self.normals == None:
			self._computeNormals()
		
		self.kNormals = []
		
		# zero kappa at start
		if len(self.values) > 0:
			self.kNormals.append(Normal3(0.0, 0.0, 0.0))
		
		# general curvature as weighted directional curvature per Taubin 95
		for i in range(1, len(self.values) - 1):
			av = self.values[i - 1] - self.values[i]
			bv = self.values[i + 1] - self.values[i]
			av2 = av.lengthSquared()
			bv2 = bv.lengthSquared()
			
			ka = 0.0
			kb = 0.0
			if av2 > 0.0:
				ka = math.fabs(2.0 * self.normals[i] *  av / av2)
			if bv2 > 0.0:
				kb = math.fabs(2.0 * self.normals[i] *  bv / bv2)
			
			# interp
			al = av.length()
			bl = bv.length()
			
			denom = al + bl
			if denom <= 0.0:
				self.kNormals.append(Normal3(0, 0, 0))
				continue
			t = bl / denom
			k = t * ka + (1.0 - t) * kb
			self.kNormals.append(k * self.normals[i])
			
		# zero kappa at end
		if len(self.values) > 1:
			self.kNormals.append(Normal3(0.0, 0.0, 0.0))
		
		# refine first and last curvatures for continuity
		# just copies adjacent values
		if len(self.values) > 2:
			self.kNormals[0] = self.kNormals[1]
			self.kNormals[len(self.kNormals) - 1] = self.kNormals[len(self.kNormals) - 2]

	def _computeKinkNormals(self):
		if self.normals == None:
			self._computeNormals()
		
		self.kinkNormals = []
		
		# zero kink at start
		if len(self.values) > 0:
			self.kinkNormals.append(Normal3(0.0, 0.0, 0.0))
		
		# general curvature as weighted directional curvature per Taubin 95
		for i in range(1, len(self.values) - 1):
			av = self.values[i - 1] - self.values[i]
			bv = self.values[i + 1] - self.values[i]
			avLen = av.length()
			bvLen = bv.length()
			
			kinkA = 0.0
			kinkB = 0.0
			if avLen > 0.0:
				kinkA = math.fabs(self.normals[i] *  av / avLen)
			if bvLen > 0.0:
				kinkB = math.fabs(self.normals[i] *  bv / bvLen)
			
			# interp value and project onto normal
			denom = avLen + bvLen
			if denom <= 0.0:
				self.kinkNormals.append(Normal3(0, 0, 0))
				continue
			t = bvLen / denom
			kink = t * kinkB + (1.0 - t) * kinkA
			self.kinkNormals.append(kink * self.normals[i])
			
		
		# zero kink at end
		if len(self.values) > 1:
			self.kinkNormals.append(Normal3(0.0, 0.0, 0.0))
			
		# refine first and last curvatures for continuity
		# just copies adjacent values
		if len(self.values) > 2:

			self.kinkNormals[0] = self.kinkNormals[1]
			self.kinkNormals[len(self.kinkNormals) - 1] = self.kinkNormals[len(self.kinkNormals) - 2]

class UniformCubicBSpline(Curve):
	
	def __init__(self, values):
		"""Initialize to four values, t in [0, 1]"""
		Curve.__init__(self)
		self.values = values
		
	def value(self, t):
		t2 = t * t
		t3 = t2 * t
		c0 =       -t3 + 3.0 * t2 - 3.0 * t + 1.0
		c1 = 3.0  * t3 - 6.0 * t2           + 4.0
		c2 = -3.0 * t3 + 3.0 * t2 + 3.0 * t + 1.0
		c3 = t3
		
		return (
			c0 * self.values[0] + 
			c1 * self.values[1] + 
			c2 * self.values[2] + 
			c3 * self.values[3]
			) / 6.0

class InterpolatingCubicBSpline(Curve):
	
	def __init__(self, values):
		"""Initialize to four values, t in [0, 1]"""
		Curve.__init__(self)
		self.values = values
		
	def value(self, t):
		return bezierEval(self.values, t)

class PiecewiseCubicBSpline(Curve):
	"""
		like unweighted maya splines.  does piecewse cubic interp
		
		i wrote this in 10 mins with one bug.  it was a typo.  yay!
	"""
	
	def __init__(self, knots, values, inValues, outValues):
		""" 
			all should be the same length, ok to set inValues[0] to None, 
			since it is ignored in current imp
		"""
		self.knots = knots
		self.values = values
		self.inValues = inValues
		self.outValues = outValues
	
	def value(self, t):
		if t <= self.knots[0]:
			return self.values[0]
		elif t >= self.knots[-1]:
			return self.values[-1]
		else:
			pi = 0
			while self.knots[pi + 1] < t:
				pi += 1
			ni = pi + 1
			u = (t - self.knots[pi]) / float(self.knots[ni] - self.knots[pi])
			return bezierEval([self.values[pi], self.outValues[pi], self.inValues[ni], self.values[ni]], u)

class Handle(object):
	def __init__(self, value, inTangentPoint, outTangentPoint):
		self.value = value
		self.inTangentPoint = inTangentPoint
		self.outTangentPoint = outTangentPoint

class HandleSpline(Curve):
	"""User-friendly spline with handles to control the shape.
	A handle consists of:
		value (Point3)
		inTangentPoint (Point3)
		outTangentPoint (Point3)
		parameter (float, stored as a knot vector)
	"""
	
	def __init__(self, values, knots = None, inTangents = [], outTangents = [], format = 'valueSequence'):
		"""
		Handles encode piecewise-cubic bspline/bezier curves, with the joins representing the handles.
		
		There are threee ways to initialize data:
		1) values: format = 'valueSequence'
			just list the values, every 4th is a handle
			len(values) % 3 should be one
			len(knots) should be len(values) / 3 if specified
		2) values and tangents: format = 'handleTangents'
			values should be just the handle values
			inTangents and outTagents should contains vector tangent values, which will be converted to 
				the internal value representation
			len(values) = len(inTangents) = len(outTangents)
				first inTangent and last outTangent don't affect the evaluation, ok to set to None
			len(knots) should be len(values) if specified
		3) values and tnagent values: format = 'handlePoints'
			values should be just the handle values
			inTangents and outTangents are tangent values, (as in the internal rep), not the mathematical
				tangent/first derivative
			length constraints on values, inTangents, outTangents, knots as in option 2
			
		Knots can be specified.  The knots should correspond to the handle values.  The nurbsData call
			will expand this to the more typical representation.  They default to integers starting at 0
		"""
		
		Curve.__init__(self)
		lp = len(values)
		self.handles = []
		oneThird = 1.0 / 3.0
		
		if format == 'valueSequence':
			# interpret as sequence of values, 
			# where every 4th is a handle and those between are handle values
			
			if lp % 3 != 1 or lp < 4:
				raise BadInitilizationData()
			numHandles = lp / 3 + 1
			
			self.handles.append(Handle(values[0], None, values[1]))
			for i in range(1, numHandles - 1):
				self.handles.append(Handle(values[3 * i], values[3 * i - 1], values[3 * i + 1]))
			self.handles.append(Handle(values[3 * (numHandles - 1)], 
									   values[3 * (numHandles - 1) - 1], 
									   None))
			
		elif format == 'handleTangents':
			# interpret as values = values, inTangents are vectors, outTangents are vectors
			
			if lp != len(inTangents) or lp != len(outTangents):
				raise BadInitilizationData()
			
			self.handles.append(Handle(values[0], 
									   None, 
									   values[0] + oneThird * outTangents[0]))
			for i in range(1, lp - 1):
				self.handles.append(Handle(values[i], 
										   values[i] - oneThird * inTangents[i], 
										   values[i] + oneThird * outTangents[i]))
			self.handles.append(Handle(values[lp - 1], 
									   values[lp - 1] - oneThird * inTangents[lp - 1],
									   None))
			
		elif format == 'handlePoints':
			# interpret as values = values, inTangents are values, outTangents are values
			
			if lp != len(inTangents) or lp != len(outTangents):
				raise BadInitilizationData()
			for i in range(lp):
				self.handles.append(Handle(values[i], inTangents[i], outTangents[i]))
		
		else: 
			raise BadInitializationData()
		
		if knots == None:
			self.knots = []
			for i in range(len(self.handles)):
				self.knots.append(float(i))
		else:
			if len(knots) == len(self.handles):
				self.knots = knots
			else:
				raise BadInitializationData()
	
	def numberOfHandles(self):
		return len(self.handles)
	
	def value(self, t):
		pt, nt = _boundIndices(self.knots, t)
		if pt == nt:
			return self.handles[pt].value
		
		tNorm = (t - self.knots[pt]) / float(self.knots[nt] - self.knots[pt])
		return bezierEval([self.handles[pt].value,
						    self.handles[pt].outTangentPoint,
							self.handles[nt].inTangentPoint,
							self.handles[nt].value], tNorm)
	
	def minT(self):
		return self.knots[0]
	
	def maxT(self):
		return self.knots[len(self.knots) - 1]
	
	def nurbsData(self):
		"""Returns a value list and knot vector for traditional nurbs curve representations"""
		
		valuesData = []
		if len(self.handles) > 0:
			valuesData.append(self.handles[0].value)
			valuesData.append(self.handles[0].outTangentPoint)
		for i in range(1, len(self.handles) - 1):
			valuesData.append(self.handles[i].inTangentPoint)
			valuesData.append(self.handles[i].value)
			valuesData.append(self.handles[i].outTangentPoint)
		if len(self.handles) > 1:
			valuesData.append(self.handles[len(self.handles) - 1].inTangentPoint)
			valuesData.append(self.handles[len(self.handles) - 1].value)
		
		
		knotsData = []
		for k in self.knots:
			knotsData.append(float(k))
			knotsData.append(float(k))
			knotsData.append(float(k))
		
		return valuesData, knotsData
		
class DisplacementHandleSpline(HandleSpline):

	def __init__(self, values, knots = None, inTangents = [], outTangents = [], format = 'valueSequence',
				 displacementValues = None, displacementKnots = None):
		"same as HandleSpline, but displacementValues and displacementKnots must be specified too"
		HandleSpline.__init__(self, values, knots, inTangents, outTangents, format)
		if displacementValues == None or displacementKnots == None:
			raise BadInitializationData()
		self.displacementValues = displacementValues
		self.displacementKnots = displacementKnots

	def value(self, t):
		v = HandleSpline.value(self, t)
		v += self._dispEval(t)
		return v
	
	def splineValue(self, t):
		return HandleSpline.value(self, t)
	
	def displacementNurbsData(self):
		"Returns values and a knot sequence for making a linear spline"
		knots = self.displacementKnots[:]
		if self.minT() not in knots:
			knots.append(self.minT())
		if self.maxT() not in knots:
			knots.append(self.maxT())
		for k in self.displacementKnots:
			if k < self.minT():
				knots.remove(k)
			elif k > self.maxT():
				knots.remove(k)
		knots.sort()
		values = []
		for k in knots:
			values.append(self.value(k))
		return values, knots
	
	def setHandleKnot(i, t):
		# TODO implemnt in superclass to just change the parameter value
		# TODO here: riase if it swaps order
		# TODO here: reparameertize the displacements in adjacent regions
		pass
	
	def _dispEval(self, t):
		pt, nt = _boundIndices(self.displacementKnots, t)
		if pt == nt:
			return self.displacementValues[pt]
		else:
			subT = ((t - self.displacemntKnots[pt]) / 
					float(self.displacemntKnots[nt] - self.displacemntKnots[pt]))
			return (self.displacementValues[pt] + 
					subT * (self.displacementValues[nt] - self.displacementValues[pt]))

def testUCBSEval():
	p1 = Point3(1, 1, 1)
	p2 = Point3(2, 2, 2)
	p3 = Point3(3, 3, 3)
	p4 = Point3(4, 4, 4)
	c = UniformCubicBSpline([p1, p2, p3, p4])
	numPts = 35
	for i in range(numPts):
		t = i * 1.0 / (numPts - 1)
		print t, ':', c.value(t)

def testICBSEval():
	p1 = Point3(1, 1, 1)
	p2 = Point3(2, 2, 2)
	p3 = Point3(3, 3, 3)
	p4 = Point3(4, 4, 4)
	c = InterpolatingCubicBSpline([p1, p2, p3, p4])
	numPts = 35
	for i in range(numPts):
		t = i * 1.0 / (numPts - 1)
		print t, ':', c.value(t)

def testLinearBounds():
	p1 = Point3(1, 1, 1)
	p2 = Point3(2, 2, 2)
	p3 = Point3(3, 3, 3)
	p4 = Point3(4, 4, 4)
	c = LinearSpline([p1, p2, p3, p4])
	minT = -5
	maxT = 10
	numPts = 31
	for i in range(numPts):
		t = minT + i * (maxT - minT) / float(numPts - 1)
		print t, ':', _boundIndices(c.knots, t)

def testLinearEval():
	p1 = Point3(1, 1, 1)
	p2 = Point3(2, 2, 2)
	p3 = Point3(3, 3, 3)
	p4 = Point3(4, 4, 4)
	c = LinearSpline([p1, p2, p3, p4])
	minT = -5
	maxT = 10
	numPts = 91
	for i in range(numPts):
		t = minT + i * (maxT - minT) / float(numPts - 1)
		print t, ':', c.value(t)

def testSampledDerivs():
	p1 = Point3(1, 1, 0)
	p2 = Point3(2, 4, 0)
	p3 = Point3(3, 9, 0)
	p4 = Point3(4, 16, 0)
	c = SampledSpline([p1, p2, p3, p4])
	
	numPts = 4
	for i in range(numPts):
		print i, ':', c.values[i], c.normal(i), c.curvatureNormal(i), c.kinkNormal(i)


if __name__ == '__main__':
	testICBSEval()