#  Copyright (c) 2008 patrick coleman

from Exceptions import *

import geom.Rotations
import mmath.Quaternion
import mmath.Vector3

# TODO _qrBezierEval corretly
# TODO test QRInterpolatingCubicSpline
# TODO QRHandle and QRHandleSpline and QRDisplacedHandleSpline classes
# TODO someday: QRSampledSpline for intrinsic analysis instead of external streamframe analysis

def _boundIndices(knots, t):
	"""returns indices to knots bounding t; they might be equal, in this case 
	t corresponds to a knot exactly"""
	priorIndex = 0
	nextIndex = 0
		
	# t before knots or at start
	if t <= knots[0]:
		return (0, 0)
		
	# t after knots or at end
	lastI = len(knots) - 1
	if t >= knots[lastI]:
		return (lastI, lastI)
		
	# test for on knots
	for i in range(1, len(knots) - 1):
		if knots[i] == t:
			return (i, i)
		
	# general check for prior index
	for i in range(1, len(knots)):
		if knots[i] > t:
			priorIndex = i - 1
			break
		
	# set nextIndex, clamp to end
	nextIndex = priorIndex + 1
	if nextIndex == len(knots):
		nextIndex -= 1
			
	return (priorIndex, nextIndex)

def qrBezierEval(values, t):
	"""Bezier interpolation of qrotations by deCasteljau"""
	
	qOne1 = values[0].interpolate(t, values[1])
	qOne2 = values[1].interpolate(t, values[2])
	qOne3 = values[2].interpolate(t, values[3])
	
	qTwo1 = qOne1.interpolate(t, qOne2)
	qTwo2 = qOne2.interpolate(t, qOne3)
	
	return qTwo1.interpolate(t, qTwo2)
	
class QRCurve(object):
	def value(self, t):
		raise AbstractMethod()

class QRLinearSpline(QRCurve):
	def __init__(self, values, knots = None):
		"""arbitrary qr number, defaults knots to [0, len(values) - 1]"""
		self.values = values
		if knots != None:
			self.knots = knots
		else:
			self.knots = range(len(values))
	
	def value(self, t):
		pt, nt = _boundIndices(self.knots, t)
		if pt == nt:
			return self.values[pt]
		else:
			subT = (t - self.knots[pt]) / float(self.knots[nt] - self.knots[pt])
			return self.values[pt] + subT * (self.values[nt] - self.values[pt])

class QRSampledSpline(QRLinearSpline):
	"""Implement this if looking into intrinisic qrotation analysis on SO3"""
	pass

class QRInterpolatingCubicBSpline(QRCurve):
	"""Mostly a test class for _qrBezierEval"""
		
	def __init__(self, values):
		"""Initialize to four values, t in [0, 1]"""
		QRCurve.__init__(self)
		self.values = values
		
	def value(self, t):
		return _qrBezierEval(self.values, t)
		
class PiecewiseCubicQRBSpline(QRCurve):
	"""
		piecewise interpolating cubic with multiple knots
		values and tangent values should be QRotations
	"""
	
	def __init__(self, knots, values, inValues, outValues):
		""" 
			all should be the same length, ok to set inValues[0] to None, 
			since it is ignored in current imp
		"""
		self.knots = knots
		self.values = values
		self.inValues = inValues
		self.outValues = outValues
	
	def value(self, t):
		if t <= self.knots[0]:
			return self.values[0]
		elif t >= self.knots[-1]:
			return self.values[-1]
		else:
			pi = 0
			while self.knots[pi + 1] < t:
				pi += 1
			ni = pi + 1
			u = (t - self.knots[pi]) / float(self.knots[ni] - self.knots[pi])
			return qrBezierEval([self.values[pi], self.outValues[pi], self.inValues[ni], self.values[ni]], u)


def wackyBugTest():
	qr1 = geom.Rotations.QRotation(q = mmath.Quaternion.Quaternion(0.993543338262, mmath.Vector3.Vector3(0.0710683630625, -0.0601262289226, -0.0648518262107)))
	qr2 = geom.Rotations.QRotation(q = mmath.Quaternion.Quaternion(0.993543338262, mmath.Vector3.Vector3(0.0710683630625, -0.0601262289226, -0.0648518262107)))
	qr3 = geom.Rotations.QRotation(q = mmath.Quaternion.Quaternion(0.996792833825, mmath.Vector3.Vector3(0.0516320344519, 0.0470159959175, 0.039085490672)))
	qr4 = geom.Rotations.QRotation(q = mmath.Quaternion.Quaternion(0.996792833825, mmath.Vector3.Vector3(0.0516320344519, 0.0470159959175, 0.039085490672)))
	vals = [qr1, qr2, qr3, qr4]
	res = qrBezierEval(vals, 0.96875)
	
	print 'res:', res.q, res.q.magnitude()
	

if __name__ == '__main__':
	wackyBugTest()