#  Copyright (c) 2008 patrick coleman

import math

import numpy
import scipy.optimize.optimize

import QRCurve
import mmath
import geom.Rotations
import motion.StreamframeFilters
import motion.StreamframeAnalysis
from Exceptions import *

# TODO class for representing a found knot (QRMaximum?)
# TODO QRHandleSplineCeator::spline()
# TODO NOW QRSplineCeator::chooseKnots
# TODO QRHandleSplineCeator::fitTangents(self, knotIndices)
# TODO QRDisplacementHandleSplineCreator::spline

class SFKnot(object):
	def __init__(self, index, magnitude):
		self.index = index
		self.magnitude = magnitude

	def __cmp__(self, other):
		return cmp(self.magnitude, other.magnitude)

class QRSplineCreator(object):
	def __init__(self, sampledSpline, streamframes, smoothingBias = 0.5, smoothingIterations = 50,
				 analyzerType = 'kink'):
		"streamframes is a collection (list/tuple) of streamframes"
		smoothedFrames = []
		for sf in streamframes:
			sfFilter = motion.StreamframeFilters.SmoothingFilter(sf,
																 bias = smoothingBias,
																 iterations = smoothingIterations)
			smoothedFrames.append(sfFilter.output())
	
		
		self.samples = sampledSpline
		self.streamframes = smoothedFrames
		self.trueStreamframes = streamframes
		self.smoothingBias = smoothingBias
		self.smoothingIterations = smoothingIterations
		self.analyzerType = analyzerType
	
	def spline(self):
		raise AbstractMethod()
	
	def chooseKnots(self, numComponents = 2):
		"Returns indices of knots, sorted by importance"
		
		# use a streamframe analyzer to get an event sequence on this streamframe
		# is the poseSignal important?
		analyzers = []
		eventLists = []
		#poseSignals = []
		for sf in self.streamframes:
			analyzer = None
			if self.analyzerType == 'kink':
				analyzer = motion.StreamframeAnalysis.KinkAnalyzer(sf, smoothingIterations = 0)
			else:
				analyzer = motion.StreamframeAnalysis.CurvatureAnalyzer(sf, smoothingIterations = 0)
			analyzers.append(analyzer)
			#poseSignals.append(analyzer.poseSignal())
			eventLists.append(analyzer.eventSequence(includeBounds = False))
	
		# at this point, the eventLists have event.time and event.component (0, 1, 2), event.magnitude 
		# now also event.index
		# make a knot list.  each knot i has a magnitude of:
		#	sum events of index i: sum_comps: (numComponents >= comp) * mag
		# (QRKnot: index & magnitude)
		# then sort the list by magnitude
		# then return indices only
		sfKnots = {}
		for el in eventLists:
			for e in el:
				if e.component < numComponents:
					if sfKnots.has_key(e.index):
						sfKnots[e.index].magnitude += e.magnitude
					else:
						sfKnots[e.index] = SFKnot(e.index, e.magnitude)
		sfKnots = sfKnots.values()
		sfKnots.sort()
		sfKnots.reverse()
		
		# start, end knots only if len(first sf) > 0, > 1, respectively
		knots = []
		if len(self.streamframes) > 0:
			if len(self.streamframes[0]) > 0:
				knots.append(0)
		if len(self.streamframes) > 0:
			if len(self.streamframes[0]) > 1:
				knots.append(len(self.streamframes[0]) - 1)

		for sfk in sfKnots:
			knots.append(sfk.index)
		
		return knots
		
class QRLinearSplineCreator(QRSplineCreator):
	def __init__(self, sampledSpline, streamframes, smoothingBias = 0.5, smoothingIterations = 50,
				 analyzerType = 'kink'):
		QRSplineCreator.__init__(self, sampledSpline, streamframes, smoothingBias, smoothingIterations,
								 analyzerType)
	
	def spline(self, knotIndices = None):
		pass
	
	



class QRHandleSplineCreator(QRSplineCreator):
	"""Assumes the value data are of type QRotation"""
	def __init__(self, sampledSpline, streamframes): # missing lots of params here
		QRSplineCreator.__init__(self, sampledSpline, streamframes)
	
	def spline(self, knotIndices = None):
		pass
	
	def fitTangents(self, knotIndices):
		pass
	
class QRDisplacementHandleSplineCreator(QRHandleSplineCreator):
	def __init__(self, sampledSpline, streamframes, smoothingBias = 0.5, smoothingIterations = 50):
		QRSplineCreator.__init__(self, sampledSpline, streamframes, smoothingBias, smoothingIterations)
		
	def spline(self, knotIndices = None):
		pass

def fitBezierTangentPoints(qr1, qr4, qRotations, knots, fitWeight = 1000.0, unitWeight = 1.0, 
						   maxIterations = 20, errTol = 5.0):
	"""returns qr2, qr3, the bezier interpolation tangents
	knots should be in [0, 1],
	currently uses the nelder mead algorithm
	params are set to run in < 3sec for ~50 samples
	note: a c implementation of the obj func would speed things up a lot"""
	
	iqr2 = qr1.interpolate(0.333, qr4)
	iqr3 = qr1.interpolate(0.667, qr4)
	
	initParams = numpy.array([iqr2.q.s, iqr2.q.v.x, iqr2.q.v.y, iqr2.q.v.z,
							  iqr3.q.s, iqr3.q.v.x, iqr3.q.v.y, iqr3.q.v.z])
	
	def objFunc(params, qr1 = qr1, qr4 = qr4, qRotations = qRotations, knots = knots,
				fitWeight = fitWeight, unitWeight = unitWeight):
		"unitizeError * (2.0 - |q2| -|q3|) + fitError * avg(slerp(knots[i]).angle(qRotations[i]))"
		q2 = mmath.Quaternion.Quaternion(params[0],
					mmath.Vector3.Vector3(params[1], params[2], params[3]))
		q3 = mmath.Quaternion.Quaternion(params[4],
					mmath.Vector3.Vector3(params[5], params[6], params[7]))
		
		unitizeError = unitWeight * math.fabs(2.0 - q2.dot(q2) - q3.dot(q3))
		
		
		q2norm = q2.normalize()
		q3norm = q3.normalize()
		qr2 = geom.Rotations.QRotation(q = q2norm)
		qr3 = geom.Rotations.QRotation(q = q3norm)
		fitError = 0.0
		for i in range(len(knots)):
			qInterp = geom.QRCurve.qrBezierEval((qr1, qr2, qr3, qr4), knots[i])
			cosPhi = qInterp.q.dot(qRotations[i].q)
			# catch slerp inaccuracies
			if cosPhi < -1.0:
				cosPhi = -1.0
			if cosPhi > 1.0:
				cosPhi = 1.0
			fitError += (math.acos(cosPhi) **2)
		
		fitError *= fitWeight
		fitError /= float(len(knots))
	
		return numpy.array([fitError + unitizeError])
	
	finalParams = scipy.optimize.optimize.fmin(objFunc, initParams, 
							maxiter = maxIterations, ftol = errTol, disp = 0)
	
	q2 = mmath.Quaternion.Quaternion(finalParams[0],
					mmath.Vector3.Vector3(finalParams[1], finalParams[2], finalParams[3]))
	q3 = mmath.Quaternion.Quaternion(finalParams[4],
					mmath.Vector3.Vector3(finalParams[5], finalParams[6], finalParams[7]))

	q2Norm = q2.normalize()
	q3Norm = q3.normalize()
	qr2 = geom.Rotations.QRotation(q = q2Norm)
	qr3 = geom.Rotations.QRotation(q = q3Norm)
	return qr2, qr3
		
