#  Copyright (c) 2008 patrick coleman

import math
import random

import numpy

import Exceptions
import geom
from mmath.Vector3 import Vector3
from Point3 import Point3
from mmath.Quaternion import Quaternion
from mmath.Matrix4 import Matrix4
from geom.Transformation import Transformation

import mmath.Matrix
import mmath.Vector

class AbstractMethod(Exception): pass
class TODO(Exception): pass # TODO get rid of this

class Rotation(Transformation):
	def __init__(self, matrixData):
		"""Subclasses should call and pass in none, users should pass in Matrix4"""
		self.matrixData = matrixData

	def matrix(self):
		"""Overridden from Transform to let Rotations support lazy computation.
		Ideally, setattr or whatnot should be overriden to dirty the matrix when params change"""
		if self.matrixData == None:
			self._updateMatrix()
		return self.matrixData
		
	def _updateMatrix(self):
		"Rotation subclasses should implement this"
		raise AbstractMethod()

	def qRotation(self):
		"Numerically robust matrix extraction from Hansen Section 16.1"
		
		mat = self.matrix()
		t = mat.trace() - mat[3][3] # 3x3 rotation submatrix trace
		q0 = 0.5 * math.sqrt(t + 1.0)
		
		if t > 0.0:
			inv4Q0 = 1.0 / (4.0 * q0)
			q1 = inv4Q0 * (mat[2][1] - mat[1][2])
			q2 = inv4Q0 * (mat[0][2] - mat[2][0])
			q3 = inv4Q0 * (mat[1][0] - mat[0][1])
			return QRotation(q = Quaternion(q0, Vector3(q1, q2, q3)))
		else:
			q1 = 0.5 * math.sqrt(1.0 + mat[0][0] - mat[1][1] - mat[2][2])
			q2 = 0.5 * math.sqrt(1.0 + mat[1][1] - mat[2][2] - mat[0][0])
			q3 = 0.5 * math.sqrt(1.0 + mat[2][2] - mat[0][0] - mat[1][1])
			if q1 > q2 and q1 > q3:
				inv4Q1 = 1.0 / (4.0 * q1)
				q0 = inv4Q1 * (mat[2][1] - mat[1][2])
				q2 = inv4Q1 * (mat[1][0] + mat[0][1])
				q3 = inv4Q1 * (mat[0][2] + mat[2][0])
				return QRotation(q = Quaternion(q0, Vector3(q1, q2, q3)))
			elif q2 > q1 and q2 > q3:
				inv4Q2 = 1.0 / (4.0 * q2)
				q0 = inv4Q2 * (mat[0][2] - mat[2][0])
				q1 = inv4Q2 * (mat[0][1] + mat[1][0])
				q3 = inv4Q2 * (mat[1][2] + mat[2][1])
				return QRotation(q = Quaternion(q0, Vector3(q1, q2, q3)))
			else: # q3 >= q1 and q3 >= q2
				inv4Q3 = 1.0 / (4.0 * q3)
				q0 = inv4Q3 * (mat[1][0] - mat[0][1]) 
				q1 = inv4Q3 * (mat[2][0] + mat[0][2])
				q2 = inv4Q3 * (mat[1][2] + mat[2][1])
				return QRotation(q = Quaternion(q0, Vector3(q1, q2, q3)))
	
	def aaRotation(self):
		# TODO get transformation and extract axis angle and return an AARotation
		raise TODO()
	
	def eulerRotation(self, order = 'xyz', local = True):
		"""Rotation matrix factorization.  Based on Dave Eberly's notes at
		http://www.geometrictools.com/Documentation/EulerAngles.pdf
		Many bug fixes to his derivations included.  :-)"""
	
		mat = self.matrix()
		
		rx = 0.0
		ry = 0.0
		rz = 0.0
		halfPi = 0.5 * math.pi
		
		if (local and (order == 'xyz')) or ((not local) and (order == 'zyx')): # MxMyMz factorization
			ry = math.asin(mat[0][2])
			if ry < halfPi:
				if ry > - halfPi:
					rx = math.atan2(-mat[1][2], mat[2][2])
					rz = math.atan2(-mat[0][1], mat[0][0])
				else:
					rx = - math.atan2(mat[1][0], mat[1][1])
					rz = 0.0
			else:
				rx = math.atan2(mat[1][0], mat[1][1])
				rz = 0.0
			
		elif (local and (order == 'xzy')) or ((not local) and (order == 'yzx')): # MxMzMy factorization
			rz = math.asin(-mat[0][1])
			if rz < halfPi:
				if rz > - halfPi:
					rx = math.atan2(mat[2][1], mat[1][1])
					ry = math.atan2(mat[0][2], mat[0][0])
				else:
					rx = math.atan2(-mat[2][0], mat[2][2]) # Eberly bug
					ry = 0.0
			else:
				rx = math.atan2(mat[2][0], mat[2][2]) # Eberly bug
				ry = 0.0
		
		elif (local and (order == 'yxz')) or ((not local) and (order == 'zxy')): # MyMxMz factorization
			rx = math.asin(-mat[1][2])
			if rx < halfPi:
				if rx > - halfPi:
					ry = math.atan2(mat[0][2], mat[2][2])
					rz = math.atan2(mat[1][0], mat[1][1])
				else:
					ry = math.atan2(- mat[0][1], mat[0][0]) # Eberly bug
					rz = 0.0
			else:
				ry = - math.atan2(- mat[0][1], mat[0][0]) # Eberly Bug
				rz = 0.0
		
		elif (local and (order == 'yzx')) or ((not local) and (order == 'xzy')): # MyMzMx factorization
			rz = math.asin(mat[1][0])
			if rz < halfPi:
				if rz > - halfPi:
					rx = math.atan2(- mat[1][2], mat[1][1])
					ry = math.atan2(- mat[2][0], mat[0][0])
				else:
					rx = 0.0
					ry = - math.atan2(mat[2][1], mat[2][2])
			else:
				rx = 0.0
				ry = math.atan2(mat[2][1], mat[2][2])
		
		elif (local and (order == 'zxy')) or ((not local) and (order == 'yxz')): # MzMxMy factorization
			rx = math.asin(mat[2][1])
			if rx < halfPi:
				if rx > - halfPi:
					ry = math.atan2(- mat[2][0], mat[2][2])
					rz = math.atan2(- mat[0][1], mat[1][1])
				else:
					ry = 0.0 
					rz = - math.atan2(mat[0][2], mat[0][0])
			else:
				ry = 0.0
				rz = math.atan2(mat[0][2], mat[0][0])
		
		else: # MzMyMx factorization
			ry = math.asin(-mat[2][0])
			if ry < halfPi:
				if ry > - halfPi:
					rx = math.atan2(mat[2][1], mat[2][2])
					rz = math.atan2(mat[1][0], mat[0][0])
				else:
					rx = 0.0
					rz = - math.atan2(mat[0][1], mat[1][1]) # bug in eberly's code
			else:
				rx = 0.0 
				rz = math.atan2(- mat[0][1], mat[0][2])
	
		return EulerRotation(rx, ry, rz, order, local)
	
	def interpolate(self, t, other):
		raise AbstractMethod()
		
	def _radians(angle, degrees):
		if degrees:
			return math.radians(angle)
		else:
			return angle
	_radians = staticmethod(_radians)
	
class QRotation(Rotation):
	def __init__(self, axis = Vector3(1.0, 0.0, 0.0), angle = 0.0, degrees = False, q = None):
		"""Default to constucting using axis and angle args.  If q is provided, that overrides them/
		REQUIRES NORMALIZED AXIS!!!"""
		Rotation.__init__(self, None)
		if q == None:
			halfAngle = 0.5 * self._radians(angle, degrees)
			self.q = Quaternion(math.cos(halfAngle), math.sin(halfAngle) * axis)
		else:
			self.q = q
		
		self.qInv = self.q.inverse()
	
	def qRotation(self):
		return self
	
	def __call__(self, other):
		t = type(other)
		v = other
		if t == Point3:
			v = Vector3(other.x, other.y, other.z)
		elif t != Vector3:
			raise Exceptions.BadArgumentData()
		
		return (self.q * Quaternion(0, v) * self.qInv).v
	
	def interpolate(self, t, other, closest = True):
		"""linear interpolation of unit quaternions on S^3; from Hansen, with shortest interpolant
		test adapted from Parent"""
		
		otherQ = other.q
		cosPhi = self.q.dot(other.q)
		
		if closest and (cosPhi < 0.0):
				otherQ = - otherQ
				cosPhi = self.q.dot(otherQ)
		
		# numical inaccuracies can percolate this neglibably above 1.0 which causes acos to return nan
		if cosPhi > 1.0:
			cosPhi = 1.0
		
		phi = math.acos(cosPhi)
		
		# test for interpolation against quats with zero angle
		if phi <= 0.0: # in [0, pi] by definition of acos
			return QRotation(q = self.q)
			
		invSinPhi = 1.0 / math.sin(phi)
		unnormQ = (math.sin((1.0 - t) * phi) * invSinPhi * self.q + math.sin(t * phi) * invSinPhi * otherQ) 
		return QRotation(q = unnormQ)
	
	def __add__(self, other):
		"rotation concatenation Rself(Rother(v)), ie. other gets applied first "
		# R-toL multiplication gives R-to-L ordering in expressions calling this operation
		return QRotation(q = other.q * self.q)
	
	def __sub__(self, other):
		"""quaternion rotation difference operation
		Given two rotations, self and other, what is the difference rotation self - other?
		In other words, Rself(v) = Rres(Rother(v))
		In quaternion math qSelf * v * qSelf-1 = qRes * qOther * v * qOther-1 * qRes-1
		v is your arbitrary rotated vector
		Note thate this operation is not commutative!  
		It returns the rotation to apply AFTER other to get self!
		
		Possible todo item, implement converse order and include flag or something
		This is a 4 eqn, 4 unknown system, my derivation.
		See near end of Draw Book 5, page labeled 'Measuring Rotation Disps' for details"""
		
		return QRotation(q = other.q.inverse() * self.q)
		
		# old code
		# i guess when it's 5 in the morning and you figure out how to factorize quaternions, 
		#you forget about simple things like inverses.  oops
		
		r = self.q
		p = other.q
		
		aData = [[  p.s, -p.v.x, -p.v.y, -p.v.z],
				 [p.v.x,    p.s, -p.v.z,  p.v.y],
				 [p.v.y,  p.v.z,    p.s, -p.v.x],
				 [p.v.z, -p.v.y,  p.v.x,    p.s]]
		bData = [r.s, r.v.x, r.v.y, r.v.z]
		
		a = mmath.Matrix.Matrix(aData)
		b = mmath.Vector.Vector(bData)
		x = a.solve(b)
		
		return QRotation(q = Quaternion(x[0], Vector3(x[1], x[2], x[3])))
	
	def __neg__(self):
		"""
			returns opposite rotation (equiv to negating theta or axis but not both)
		"""
		return QRotation(q = mmath.Quaternion.Quaternion(self.q.s, -self.q.v))
	
	def __repr__(self):
		return ('geom.Rotations.QRotation(' + 
				'q = ' + repr(self.q) +
				')')
	
	def _updateMatrix(self):
		"Adapted from Parent"
		x_2 = 2.0 * self.q.v.x # 2x
		y_2 = 2.0 * self.q.v.y # 2y
		z_2 = 2.0 * self.q.v.z # 2z
		x2_2 = x_2 * self.q.v.x # 2x^2
		y2_2 = y_2 * self.q.v.y # 2y^2
		z2_2 = z_2 * self.q.v.z # 2z^2
		xy_2 = x_2 * self.q.v.y # 2xy
		xz_2 = x_2 * self.q.v.z # 2xz
		yz_2 = y_2 * self.q.v.z # 2yz
		sx_2 = x_2 * self.q.s # 2sx
		sy_2 = y_2 * self.q.s # 2sy
		sz_2 = z_2 * self.q.s # 2sz
		
		
		self.matrixData = Matrix4([
			[1.0 - y2_2 -z2_2, xy_2 - sz_2,        xz_2 + sy_2,       0.0],
			[xy_2 + sz_2,      1.0 - x2_2 - z2_2,  yz_2 - sx_2,       0.0],
			[xz_2 - sy_2,      yz_2 + sx_2,        1.0 - x2_2 - y2_2, 0.0],
			[0.0,              0.0,                0.0,               1.0]])
	
class AARotation(Rotation):
	def __init__(self, axis = Vector3(1.0, 0.0, 0.0), angle = 0.0, degrees = False):
		Rotation.__init__(self, None)
		self.axis = axis
		self.angle = self._radians(angle, degrees)

	def aaRotation(self):
		return self
	
	def _updateMatrix(self):
		c = math.cos(self.angle)
		s = math.sin(self.angle)
		
		oneMinusC = 1.0 - c
		oneMinusCX = oneMinusC * self.axis.x
		oneMinusCY = oneMinusC * self.axis.y
		oneMinusCZ = oneMinusC * self.axis.z
		
		oneMinusCXX = oneMinusCX * self.axis.x
		oneMinusCXY = oneMinusCX * self.axis.y
		oneMinusCXZ = oneMinusCX * self.axis.z
		oneMinusCYY = oneMinusCY * self.axis.y
		oneMinusCYZ = oneMinusCY * self.axis.z
		oneMinusCZZ = oneMinusCZ * self.axis.z
		
		sX = s * self.axis.x
		sY = s * self.axis.y
		sZ = s * self.axis.z
		
		self.matrixData = Matrix4([
			[ c + oneMinusCXX, oneMinusCXY - sZ, oneMinusCXZ + sY, 0.0],
			[oneMinusCXY + sZ,  c + oneMinusCYY, oneMinusCYZ - sX, 0.0],
			[oneMinusCXZ - sY, oneMinusCYZ + sX,  c + oneMinusCZZ, 0.0],
			[             0.0,              0.0,              0.0, 1.0]])
		
class EulerRotation(Rotation):
	def __init__(self, rx = 0.0, ry = 0.0, rz = 0.0, order = 'xyz', local = True, degrees = False):
		"to emulate a maya rotation, use order as in maya and set local = False and degrees = True"
		Rotation.__init__(self, None)
		
		self.rx = self._radians(rx, degrees)
		self.ry = self._radians(ry, degrees)
		self.rz = self._radians(rz, degrees)
		
		self.order = order
		self.local = local
		
	def eulerRotation(self, order = 'xyz', local = True):
		"""Returns self if order and local option match; otherwise this will
		return a new factorization of the matrix equivalent"""
		if order == self.order and local == self.local:
			return self
		return Rotation.eulerRotation(self, order, local)
	
	def _updateMatrix(self):
		
		# build per-axis matrices
		cx = math.cos(self.rx)
		sx = math.sin(self.rx)
		rxMat = Matrix4([
			[1.0, 0.0, 0.0, 0.0],
			[0.0,  cx, -sx, 0.0],
			[0.0,  sx,  cx, 0.0],
			[0.0, 0.0, 0.0, 1.0]])
		
		cy = math.cos(self.ry)
		sy = math.sin(self.ry)
		ryMat = Matrix4([
			[ cy, 0.0,  sy, 0.0],
			[0.0, 1.0, 0.0, 0.0],
			[-sy, 0.0,  cy, 0.0],
			[0.0, 0.0, 0.0, 1.0]])
		
		cz = math.cos(self.rz)
		sz = math.sin(self.rz)
		rzMat = Matrix4([
			[ cz, -sz, 0.0, 0.0],
			[ sz,  cz, 0.0, 0.0],
			[0.0, 0.0, 1.0, 0.0],
			[0.0, 0.0, 0.0, 1.0]])
		
		# concatenate per-axis matrices based on order and local/global axis option
		if self.local:
			if self.order == 'xyz':
				self.matrixData = rxMat * ryMat * rzMat
			elif self.order == 'xzy':
				self.matrixData = rxMat * rzMat * ryMat
			elif self.order == 'yxz':
				self.matrixData = ryMat * rxMat * rzMat
			elif self.order == 'yzx':
				self.matrixData = ryMat * rzMat * rxMat
			elif self.order == 'zxy':
				self.matrixData = rzMat * rxMat * ryMat
			else: # default to 'zyx'
				self.matrixData = rzMat * ryMat * rxMat
		else:
			if self.order == 'xyz':
				self.matrixData = rzMat * ryMat * rxMat
			elif self.order == 'xzy':
				self.matrixData = ryMat * rzMat * rxMat
			elif self.order == 'yxz':
				self.matrixData = rzMat * rxMat * ryMat
			elif self.order == 'yzx':
				self.matrixData = rxMat * rzMat * ryMat
			elif self.order == 'zxy':
				self.matrixData = ryMat * rxMat * rzMat
			else: # default to 'zyx'
				self.matrixData = rxMat * ryMat * rzMat

# OTHER OPS

def getQRotation(inDirection, outDirection, normalized = False):
	"""
		computes the qrotation that rotated inDirection to outDirection
		
		set normalized to true if normalizing directions externally
	"""
	theta = inDirection.angle(outDirection)
	axis = inDirection ^ outDirection
	if axis.lengthSquared() < 0.000001:
		return QRotation(mmath.Vector3.Vector3(0, 0, 0), 0.0)
	if not normalized:
		axis = axis.normalize()
	return QRotation(axis, theta)

def _fitQRotationTwoSamples(inDirections, outDirections, normalized = False):
	"""
		Fits two directions in inDirections to corresponding vectors in outDirections 
		in a least squares sense 
	
		set normalized to true if normalizing directions externally	
	
		step 1 aligns mean vectors (this splits dp error between corresponding pairs)
		step 2 spins around aligned mean vector to maximize dp 
	"""
	
	ids = inDirections[:]
	ods = outDirections[:]
	if not normalized:
		ids = [id.normalize() for id in ids]
		ods = [od.normalize() for od in ods]
	
	# 1) find rotation to align mean axes
	
	# get mean vecs
	theta1 = ids[0].angle(ids[1])
	theta2 = ods[0].angle(ods[1])
	axis1 = (ids[0] ^ ids[1]).normalize()
	axis2 = (ods[0] ^ ods[1]).normalize()
	qr1 = QRotation(axis1, 0.5 * theta1)
	qr2 = QRotation(axis2, 0.5 * theta2)
	meanVec1 = qr1(ids[0])
	meanVec2 = qr2(ods[0])
	
	# find rotation that aligns median vectors
	thetaAlign = meanVec1.angle(meanVec2)
	axisAlign = (meanVec1 ^ meanVec2).normalize()
	qrAlign = QRotation(axisAlign, thetaAlign)
		
	# 2 find rotation to spin points about aligned mean vector
	
	# transporm inDirs to align them
	ids = [qrAlign(v) for v in ids]
	
	# find vectors connecting in pair and outpair
	idVec = ids[1] - ids[0]
	odVec = ods[1] - ods[0]
	
	#find rotation that maps idVec to odVec
	axisSpin = meanVec2
	thetaSpin = idVec.angle(odVec)
	cp = (idVec ^ odVec).normalize()
	if (cp * axisSpin) < 0.0:
		thetaSpin *= -1.0	
	qrSpin = QRotation(axisSpin, thetaSpin)
		
	# 3) concatenate the mean vector alignment and spin rotations
	return QRotation(q = qrSpin.q * qrAlign.q)

def _fitQRotationThreeSamples(inDirections, outDirections, normalized = False):
	"""
		Fits three directions in inDirections to corresponding vectors in outDirections 
		in a least squares sense 

		set normalized to true if normalizing directions externally
		
		DOES NOT SUPPORT COLINEAR DATA IN EITHER inDirections of outDirections
		planar alignment becomes degenerate
	"""
	
	ids = inDirections[:]
	ods = outDirections[:]
	if not normalized:
		ids = [id.normalize() for id in ids]
		ods = [od.normalize() for od in ods]

	# 1) align normals of planes containing samples (this code will break down if they are colinear)
	
	# and the cylons have a plane
	# compute normals of planes
	iVec1 = ids[1] - ids[0]
	iVec2 = ids[2] - ids[0]
	oVec1 = ods[1] - ods[0]
	oVec2 = ods[2] - ods[0]
	iN = iVec1 ^ iVec2
	oN = oVec1 ^ oVec2
	
	# reverse normals if needed to point toward majority of points (makes debugging visually easier)
	#iCentroid = 0.33333333 * (ids[0] + ids[1] + ids[2])
	#oCentroid = 0.33333333 * (ods[0] + ods[1] + ods[2])
	#if iN * iCentroid < 0.0:
	#	iN = -iN
	#if oN * oCentroid < 0.0:
	#	oN = -oN
	
	# compute aligning rotation
	axisAlign = (iN ^ oN).normalize()
	thetaAlign = iN.angle(oN)
	qrAlign = QRotation(axisAlign, thetaAlign)
	
	# 2) find best rotation to align planar vectors (they generaly lie in two parallel planes at this point)
	
	axisSpin = oN.normalize()
	
	# aply aligning rotation to in dirs
	ids = [qrAlign(v) for v in ids]
	
	# bring dirs to appropriate local plane
	ids = [v - (v * axisSpin) * axisSpin for v in ids]
	ods = [v - (v * axisSpin) * axisSpin for v in ods]
	
	# compute theta (closed form solution that minimizes angles)
	num = sum([(ids[i] ^ ods[i]) * axisSpin for i in range(3)])
	denom = sum([ids[i] * ods[i] for i in range(3)])
	thetaSpin = math.atan(num/denom)
	qrSpin = QRotation(axisSpin, thetaSpin)
		
	# 3) concatenate the plane alignment alignment and spin rotations
	return QRotation(q = qrSpin.q * qrAlign.q) #, iN, oN, axisAlign

def _fitQRotationMultipleSamples(inDirections, outDirections, normalized = False):
	"""
		Fits multiple directions in inDirections to corresponding vectors in outDirections 
		in a least squares sense 

		set normalized to true if normalizing directions externally
	"""
	# not working; has a z-reflection in the solution for three samples
	
	# compute M matrix
	M = numpy.zeros((3, 3))
	for i in range(len(inDirections)):
		id = inDirections[i]
		od = outDirections[i]
		if not normalized:
			id = id.normalize()
			od = od.normalize()
		print id, '\n', od
		print numpy.outer([id.x, id.y, id.z], [od.x, od.y, od.y])
		M += numpy.outer([id.x, id.y, id.z], [od.x, od.y, od.y])
	
	
	print '\nM'
	print M
	print '\n'
	
	# compute unique entries of N array
	a = M.trace() #OK
	b =  M[0][0] - M[1][1] - M[2][2] #OK
	c = -M[0][0] + M[1][1] - M[2][2] #OK
	d = -M[0][0] - M[1][1] + M[2][2] #OK
	e =  M[1][2] - M[2][1] #OK
	f =  M[0][1] + M[1][0] #OK
	g =  M[1][2] + M[2][1] #OK
	h =  M[2][0] - M[0][2] #OK
	i =  M[0][2] + M[2][0] #OK
	j =  M[0][1] - M[1][0] #OK
	N = numpy.array([[a, e, h, j], [e, b, f, i], [h, f, c, g], [j, i, g, d]])

	evals, evecs = numpy.linalg.eig(N)
	iMax = list(evals).index(max(evals))
	
	print 'evals', evals
	print evecs
	q = evecs[:,iMax]
	qr = QRotation(q = mmath.Quaternion.Quaternion(q[0], mmath.Vector3.Vector3(q[1], q[2], q[3])))
	print qr.q, q
	return qr


def fitQRotation(inDirections, outDirections, normalized = False):
	"""
		given two lists of directions, finds the rotation that does 
		a least squares fit of the first set to the second
		
		set normalized to true if normalizing directions externally
		
		handles colocated point.  does not handle colinear 3 pts
	"""
	l = len(inDirections)
	if l == 1 or geom.colocated(inDirections) or geom.colocated(outDirections):
		return getQRotation(inDirections[0], outDirections[0], normalized)
	elif l == 2:
		return _fitQRotationTwoSamples(inDirections, outDirections, normalized)
	elif l == 3:
		return _fitQRotationThreeSamples(inDirections, outDirections, normalized)
	else:
		raise "general case not implemented, see Horn paper for the solution"


# TEST CODE
	
def _slerpTest():
	# build up rotations from x to y and x to z
	v0 = Vector3(1, 0, 0)
	v1 = Vector3(0, 1, 0)
	v2 = Vector3(0, 0, 1)
	axis1 = v0 ^ v1
	axis2 = v0 ^ v2
	angle1 = v0.angle(v1)
	angle2 = v0.angle(v2)
	print "axis1, angle1:", axis1, angle1
	print "axis2, angle2:", axis2, angle2
	q1 = QRotation(axis = axis1, angle = angle1)
	q2 = QRotation(axis = axis2, angle = angle2)
	print 'q1:', q1.q
	print 'q2:', q2.q
	
	# test applying them
	print 'q1(<1, 0, 0>)', q1(v0)
	print 'q2(<1, 0, 0>)', q2(v0)
	
	# test applying the interpolation between them
	ts = [ i / 10.0 for i in range(11)]
	for t in ts:
		print t, ':',
		qt = q1.interpolate(t, q2)
		print qt(v0)

def _slerpTest2():
	# more like how things would be done in ohlone
	v0 = Vector3(1, 0, 0)
	
	er1 = EulerRotation(0, 0, 90, degrees = True)
	er2 = EulerRotation(0, -90, 0, degrees = True)
	qr1 = er1.qRotation()
	qr2 = er2.qRotation()
	print 'qr1:', qr1.q
	print 'qr2:', qr2.q
	
	# test applying them
	print 'q1(<1, 0, 0>)', qr1(v0)
	print 'q2(<1, 0, 0>)', qr2(v0)
	
	m = qr1.matrix()
	print m[0]
	print m[1]
	print m[2]
	print m[3]
	
	# test applying the interpolation between them
	ts = [ i / 10.0 for i in range(11)]
	for t in ts:
		print t, ':',
		qt = qr1.interpolate(t, qr2)
		print qt(v0), qt.q.magnitude()

def _eulerTest():
	e1 = EulerRotation(-60, 25, -25, order = 'zyx', local = False, degrees = True)
	mat = e1.matrix().transpose() # to compare with maya's left hand matrices
	print mat[0]
	print mat[1]
	print mat[2]
	print mat[3]	
	


def _factorizeTest():
	"""Comment out EulerRotation::eulerRotation() to run this test and
	add back code to return a code path in Rotation::eulerRotation if narrowing down"""
	numTests = 100
	orders = ['xyz', 'xzy', 'yxz', 'yzx', 'zxy', 'zyx']
	localStates = [True, False]
	minAngle = -720.0
	maxAngle = 720.0
	borderTestValues = [-90.0, 90.0, 0.0]
	random.seed(1)
	
	pathTotals = {'a':0, 'b':0, 'c':0, 'd':0, 'e':0, 'f':0, 'g':0, 'h':0, 'i':0, 'j':0, 'k':0,
		'l':0, 'm':0, 'n':0, 'o':0, 'p':0, 'q':0, 'r':0}
	failuresTotals = {'a':0, 'b':0, 'c':0, 'd':0, 'e':0, 'f':0, 'g':0, 'h':0, 'i':0, 'j':0, 'k':0,
		'l':0, 'm':0, 'n':0, 'o':0, 'p':0, 'q':0, 'r':0}
	for order in orders:
		for localState in localStates:
			for i in range(numTests):
				rx = random.uniform(minAngle, maxAngle)
				ry = random.uniform(minAngle, maxAngle)
				rz = random.uniform(minAngle, maxAngle)
				er1 = EulerRotation(rx, ry, rz, local = localState, order = order, degrees = True)
				er2, cp = er1.eulerRotation(local = localState, order = order), 'a'
				erVerify = EulerRotation(er2.rx, er2.ry, er2.rz, local = localState, order = order)
				mat = er1.matrix() - erVerify.matrix()
				rxr = math.radians(rx)
				ryr = math.radians(ry)
				rzr = math.radians(rz)
				if (math.fabs(mat[0][0]) > 0.001 or
					math.fabs(mat[0][1]) > 0.001 or
					math.fabs(mat[0][2]) > 0.001 or
					math.fabs(mat[0][3]) > 0.001 or
					math.fabs(mat[1][0]) > 0.001 or
					math.fabs(mat[1][1]) > 0.001 or
					math.fabs(mat[1][2]) > 0.001 or
					math.fabs(mat[1][3]) > 0.001 or
					math.fabs(mat[2][0]) > 0.001 or
					math.fabs(mat[2][1]) > 0.001 or
					math.fabs(mat[2][2]) > 0.001 or
					math.fabs(mat[2][3]) > 0.001 or
					math.fabs(mat[3][0]) > 0.001 or
					math.fabs(mat[3][1]) > 0.001 or
					math.fabs(mat[3][2]) > 0.001 or
					math.fabs(mat[3][3]) > 0.001
					):
					failuresTotals[cp] += 1
					print '\nTest', order, localState, i, ': rx, ry, rz = ', rxr, ',', ryr, ',', rzr, ';'
					print 'Result: rx, ry, rz = ', er2.rx, er2.ry, er2.rz 
					print 'Code path: ', cp
				pathTotals[cp] += 1
	print 'Random Test Path Totals:'
	ps = pathTotals.keys()
	ps.sort()
	for p in ps:
		print '\t', p, pathTotals[p], ';', failuresTotals[p], 'failures'

	pathTotals = {'a':0, 'b':0, 'c':0, 'd':0, 'e':0, 'f':0, 'g':0, 'h':0, 'i':0, 'j':0, 'k':0,
		'l':0, 'm':0, 'n':0, 'o':0, 'p':0, 'q':0, 'r':0}
	failuresTotals = {'a':0, 'b':0, 'c':0, 'd':0, 'e':0, 'f':0, 'g':0, 'h':0, 'i':0, 'j':0, 'k':0,
		'l':0, 'm':0, 'n':0, 'o':0, 'p':0, 'q':0, 'r':0}
	for order in orders:
		for localState in localStates:
			for i in range(numTests):
				rx = borderTestValues[int(random.uniform(0, 2.999))]
				ry = borderTestValues[int(random.uniform(0, 2.999))]
				rz = borderTestValues[int(random.uniform(0, 2.999))]
				er1 = EulerRotation(rx, ry, rz, local = localState, order = order, degrees = True)
				er2, cp = er1.eulerRotation(local = localState, order = order), 'a'
				erVerify = EulerRotation(er2.rx, er2.ry, er2.rz, local = localState, order = order)
				mat = er1.matrix() - erVerify.matrix()
				rxr = math.radians(rx)
				ryr = math.radians(ry)
				rzr = math.radians(rz)
				if (math.fabs(mat[0][0]) > 0.001 or
					math.fabs(mat[0][1]) > 0.001 or
					math.fabs(mat[0][2]) > 0.001 or
					math.fabs(mat[0][3]) > 0.001 or
					math.fabs(mat[1][0]) > 0.001 or
					math.fabs(mat[1][1]) > 0.001 or
					math.fabs(mat[1][2]) > 0.001 or
					math.fabs(mat[1][3]) > 0.001 or
					math.fabs(mat[2][0]) > 0.001 or
					math.fabs(mat[2][1]) > 0.001 or
					math.fabs(mat[2][2]) > 0.001 or
					math.fabs(mat[2][3]) > 0.001 or
					math.fabs(mat[3][0]) > 0.001 or
					math.fabs(mat[3][1]) > 0.001 or
					math.fabs(mat[3][2]) > 0.001 or
					math.fabs(mat[3][3]) > 0.001):
					failuresTotals[cp] += 1
					print '\nTest', order, localState, i, ': rx, ry, rz = ', rxr, ',', ryr, ',', rzr, ';'
					print 'Result: rx, ry, rz = ', er2.rx, er2.ry, er2.rz 
					print 'Code path: ', cp
					print resControl, resTest
					pm = er1.matrix()
					print pm[0]
					print pm[1]
					print pm[2]
					print pm[3]
					pm = er2.matrix()
					print pm[0]
					print pm[1]
					print pm[2]
					print pm[3]
					
					
				pathTotals[cp] += 1
	print 'Border Test Path Totals:'
	ps = pathTotals.keys()
	ps.sort()
	for p in ps:
		print '\t', p, pathTotals[p], ';', failuresTotals[p], 'failures'

def diffTest():
	axisP = Vector3(1, 0, 0) # 90
	axisR = Vector3(0, 0, 1) # -90
	qrP = QRotation(axisP, 90.0, True)
	qrR = QRotation(axisR, 90.0, True)
	qrDiff = qrR - qrP
	print 'r:', qrR.q, qrR.q.magnitude()
	print 'p:', qrP.q, qrP.q.magnitude()
	
	print 'diff:', qrDiff.q, qrDiff.q.magnitude()
	
	recon = qrP + qrDiff
	print 'recon:', recon.q, recon.q.magnitude()

def testFit2():
	inDirs = [mmath.Vector3.Vector3(4.7325929172356815, 0.68261001921033848, -2.9725025149026116),
			  mmath.Vector3.Vector3(-2.1117953330407389, 0.0, -7.7604768719724255)]
	#inDirs.reverse()
	outDirs = [mmath.Vector3.Vector3(-7.5182124844137599, 4.9838879794229296, -5.1264925930457199),
			   mmath.Vector3.Vector3(-9.684303193202517, 5.7553961596568115e-13, -0.18796901181110215)]
	inNorms = [v.normalize() for v in inDirs]
	outNorms = [v.normalize() for v in outDirs]
	preErrors = [inNorms[i] * outNorms[i] for i in [0, 1]]
	print "pre errors:", str(preErrors), str(sum(preErrors))

	qr = _fitQRotationTwoSamples(inNorms, outNorms)

	print '\nrotation info'
	print 'mag:', qr.q.magnitude()
	theta = 2 * math.acos(qr.q.s)
	print 'theta:', theta
	print 'axis:', qr.q.v / math.sin(theta)
	print '\n'

	print '\n'
	print inDirs[0]
	print outDirs[0]
	print qr(inDirs[0])
	print '\n'

	print '\n'
	print inDirs[1]
	print outDirs[1]
	print qr(inDirs[1])
	print '\n'

	resNorms = [qr(v) for v in inNorms]
	postErrors = [inNorms[i] * resNorms[i] for i in [0, 1]]
	print "post errors:", str(postErrors), str(sum(postErrors))


def testFit3():
	inDirs = [mmath.Vector3.Vector3(4.7325929172356815, 0.68261001921033848, -2.9725025149026116),
			  mmath.Vector3.Vector3(-2.1117953330407389, 0.0, -7.7604768719724255),
			  mmath.Vector3.Vector3(1.9608407306644855, -0.11192304971428513, -6.4813399989075764)]
	outDirs = [mmath.Vector3.Vector3(-7.5182124844137599, 4.9838879794229296, -5.1264925930457199),
			   mmath.Vector3.Vector3(-9.684303193202517, 5.7553961596568115e-13, -0.18796901181110215),
			   mmath.Vector3.Vector3(-8.6497439059052255, 2.9176701495497177, -1.591957724002274)]
	inNorms = [v.normalize() for v in inDirs]
	outNorms = [v.normalize() for v in outDirs]
	preErrors = [inNorms[i] * outNorms[i] for i in [0, 1, 2]]
	print "pre errors:", str(preErrors), str(sum(preErrors))

	qr = _fitQRotationThreeSamples(inNorms, outNorms)

	print '\nrotation info'
	print 'mag:', qr.q.magnitude()
	theta = 2 * math.acos(qr.q.s)
	print 'theta:', theta
	print 'axis:', qr.q.v / math.sin(theta)
	print '\n'

	print '\n'
	print inDirs[0]
	print outDirs[0]
	print qr(inDirs[0])
	print '\n'

	print '\n'
	print inDirs[1]
	print outDirs[1]
	print qr(inDirs[1])
	print '\n'
	
	print '\n'
	print inDirs[2]
	print outDirs[2]
	print qr(inDirs[2])
	print '\n'

	resDirs = [qr(v) for v in inDirs]
	resNorms = [qr(v) for v in inNorms]
	postErrors = [inNorms[i] * resNorms[i] for i in [0, 1, 2]]
	print "post errors:", str(postErrors), str(sum(postErrors))

def testFit2Bug():
	oldKidOffsets = [mmath.Vector3.Vector3(-5.9501015226e-16, -1.37561, 0.0), mmath.Vector3.Vector3(-5.9501015226e-16, -1.37561, 5.13385)]
	newKidOffsets = [mmath.Vector3.Vector3(-1.04083408559e-16, -1.37561, -6.66133814775e-15), mmath.Vector3.Vector3(-2.22044604925e-15, -1.37561, 5.13385)]
	alignQr = fitQRotation(oldKidOffsets, newKidOffsets)
	print alignQr

def testAngleBug():
	meanVec1 = mmath.Vector3.Vector3(-3.43159224019e-16, -0.7933532871, 0.608761498329)
	meanVec2 = mmath.Vector3.Vector3(-3.10981832975e-16, -0.7933532871, 0.608761498329)
	print math.acos((meanVec1 * meanVec2) / (meanVec1.length() * meanVec2.length()))
	print meanVec1.angle(meanVec2)

	
	

if __name__ == '__main__':
	#_slerpTest2()
	#_eulerTest()
	#diffTest()		
	#testFit2()
	#testFit3()
	print '###'
	testFit2Bug()
	print '###'
	testAngleBug()	