#  Copyright (c) 2008 patrick coleman

import Curve
from Point3 import Point3
from mmath.Matrix import Matrix
from Exceptions import *
import msignal.NumericFilters

class SplineCreator(object):
	def __init__(self, sampledSpline):
		self.samples = sampledSpline
	
	def spline(self):
		raise AbstractMethod()

class Maximum(object):
	def __init__(self, i, t, value, measurement):
		self.index = i
		self.t = t
		self.value = value
		self.measurement = measurement
		
	def __cmp__(self, other):
		"The measurement is used for comparison"
		return cmp(self.measurement, other.measurement)

class HandleSplineCreator(SplineCreator):
	"""This class assumes that the value data are of type Point3"""
	def __init__(self, sampledSpline):
		SplineCreator.__init__(self, sampledSpline)
	
	def spline(self, knotIndices = None):
		"""optionally takes a knotIndices argument"""
		if knotIndices == None:
			knotIndices = self.chooseKnots()
			knotIndices.sort()
		tPoints = self.fitTangents(knotIndices)
		
		# buid point sequence
		# len(tPoints) = (len(knotIndices) - 1) * 2 
		allPoints = []
		allPoints.append(self.samples.values[knotIndices[0]])
		for i in range(len(knotIndices) - 1):
			allPoints.append(tPoints[2 * i])
			allPoints.append(tPoints[2 * i + 1])
			allPoints.append(self.samples.values[knotIndices[i + 1]])
		knots = [self.samples.knots[i] for i in knotIndices]
	
		return Curve.HandleSpline(allPoints, knots)
	
	def chooseKnots(self):
		"""Returns indices of knots 
		currently chooses all local maxima and endpoints.  endpoint indices are returned first, 
		followed by the maxima, sorted by value.  that is, stronger features are returned before 
		less strong features"""
		_kink = self._kinkSignal()
		maxima = self._internalMaxima(self._kinkSignal())
		maxima.sort()
		maxima.reverse()
		knotIndices = []
		knotIndices.append(0)
		knotIndices.append(len(self.samples.values) - 1)
		for m in maxima:
			knotIndices.append(m.index)
		return knotIndices
	
	def fitTangents(self, knotIndices):
		"""Returns a list of point pairs for bezier/interpolating b-spline cubic connections
		between knots"""
		tPoints = []
		for i in range(len(knotIndices) - 1):
			k1, k2 = knotIndices[i], knotIndices[i + 1]
			p1, p2 = self._fitSpanTangents((k1, k2))
			tPoints.append(p1)
			tPoints.append(p2)
		return tPoints
		
		# TODO: maybe extend this to constrain tagent continuity at knots
		
	def _kinkSignal(self):
		res = []
		for i in range(len(self.samples.values)):
			res.append(self.samples.kinkNormal(i).length())
		return res
	
	def _internalMaxima(self, signal):
		"Will NOT return maxima at bounds of signal"
		maxima = []
		for i in range(1, len(signal) - 1):
			if signal[i] > signal[i - 1] and signal[i] > signal[i + 1]:
				maxima.append(Maximum(i, self.samples.knots[i], self.samples.values[i], signal[i]))
		return maxima
	
	def _fitSpanTangents(self, knotIndices):
		"""Returns pair of tanget points between the knot indices that matches intermediate samples.
		Does not work in reverse"""

		# check for cases where linear system is meaningless
		if knotIndices[1] < knotIndices[0]:
			raise BadArgumentData()
		elif knotIndices[0] == knotIndices[1]:
			return self.samples.values[knotIndices[0]], self.samples.values[knotIndices[0]]
		elif (knotIndices[1] - knotIndices[0]) == 1:
			tp1 = self.samples.values[knotIndices[0]].interpolate(0.333, 
				self.samples.values[knotIndices[1]])
			tp2 = self.samples.values[knotIndices[0]].interpolate(0.667, 
				self.samples.values[knotIndices[1]])
			return tp1, tp2
		
		# builds least squares system and solves for two intermdiate points
		aData = []
		bData = []
		tDenom = 1.0 / (self.samples.knots[knotIndices[1]] - self.samples.knots[knotIndices[0]])
		for i in range(knotIndices[0] + 1, knotIndices[1]):
			ti =  (self.samples.knots[i] - self.samples.knots[knotIndices[0]]) * tDenom
			oneMinusTi = 1.0 - ti
			oneMinusTi2 = oneMinusTi * oneMinusTi
			
			# Add coefficients for one equation
			aData.append([3.0 * ti * oneMinusTi * oneMinusTi, 
				    3.0 * ti * ti * oneMinusTi])
			
			# Add constant term for one equation
			bPt = (self.samples.values[i]
				   - (oneMinusTi2 * oneMinusTi * self.samples.values[knotIndices[0]])
				   -  (ti ** 3 * self.samples.values[knotIndices[1]]))
			bData.append([bPt.x, bPt.y, bPt.z])

		a = Matrix(aData)
		b = Matrix(bData)
		x = a.leastSquares(b)
		tp1 = Point3(x[0][0], x[0][1], x[0][2])
		tp2 = Point3(x[1][0], x[1][1], x[1][2])
		return tp1, tp2

class DisplacementHandleSplineCreator(HandleSplineCreator):
	
	def __init__(self, sampledSpline, smoothingBias = 0.5, smoothingIterations = 50):
		sFilter = msignal.NumericFilters.SmoothingFilter(sampledSpline.values, 
															 smoothingBias,
															 smoothingIterations)
		HandleSplineCreator.__init__(self, 
									 Curve.SampledSpline(sFilter.output(), 
														 sampledSpline.knots))
		self.trueSamples = sampledSpline.values
		
	def spline(self, knotIndices = None):
		if knotIndices == None:
			knotIndices = self.chooseKnots()
			knotIndices.sort()
		tPoints = self.fitTangents(knotIndices)
		
		# buid point sequence
		# len(tPoints) = (len(knotIndices) - 1) * 2 
		allPoints = []
		
		allPoints.append(self.samples.values[knotIndices[0]])
		for i in range(len(knotIndices) - 1):
			allPoints.append(tPoints[2 * i])
			allPoints.append(tPoints[2 * i + 1])
			allPoints.append(self.samples.values[knotIndices[i + 1]])
		
		knots = [self.samples.knots[i] for i in knotIndices]
	
		hs = Curve.HandleSpline(allPoints, knots)
		
		dispValues = []
		for i in range(len(self.samples.knots)):
			dispValues.append(self.trueSamples[i] - hs.value(self.samples.knots[i]))
		return Curve.DisplacementHandleSpline(allPoints, 
											  knots,
											  displacementValues = dispValues,
											  displacementKnots = self.samples.knots)

def fitBezierTangentPoints(p1, p4, points, knots):
	"""bare bones tangent fitter:
	p1, p4: first and last points
	points: points to fit
	knots: parameter values of the points, these are most likely in [0, 1] (although the math is general) 
	returns p2, p3, the intermediate bezier points"""

		
	# builds least squares system and solves for two intermdiate points
	aData = []
	bData = []
	
	
	for i in range(len(points)):
		ti = knots[i]
		oneMinusTi = 1.0 - ti
		oneMinusTi2 = oneMinusTi * oneMinusTi
			
		# Add coefficients for one equation
		aData.append([3.0 * ti * oneMinusTi * oneMinusTi, 
				3.0 * ti * ti * oneMinusTi])
			
		# Add constant term for one equation
		bPt = (points[i]
				- (oneMinusTi2 * oneMinusTi * p1)
				-  (ti ** 3 * p4))
		bData.append([bPt.x, bPt.y, bPt.z])

	a = Matrix(aData)
	b = Matrix(bData)
	x = a.leastSquares(b)
	tp1 = Point3(x[0][0], x[0][1], x[0][2])
	tp2 = Point3(x[1][0], x[1][1], x[1][2])
	return tp1, tp2

if __name__ == '__main__':
	p1 = Point3(3, 3, 1)
	p2 = Point3(-3, 10.5, 2)
	p3 = Point3(-1.5, 14, 1.5)
	p4 = Point3(3, 12, -.5)
	p5 = Point3(4.5, 7.5, -2.5)

	analyzer = HandleSplineCreator(Curve.SampledSpline([p1, p2, p3, p4, p5]))
	
	
	s = analyzer.spline()
	ki = analyzer.chooseKnots()
	print analyzer.fitTangents([0, 2, 4])
	