#  Copyright (c) 2008 patrick coleman

import numpy
import numpy.linalg
import Vector

import Exceptions

class Matrix(object):
	"Float matrices of arb. size"
	
	def __init__(self, data):
		t = type(data) 
		if t == list:
			self.data = numpy.array(data, float)
			return
		elif t == numpy.ndarray:
			self.data = data
			return
		raise Exceptions.BadInitializationData()
	
	def rowSize(self):
		return self.data.shape[0]
	
	def columnSize(self):
		return self.data.shape[1]
		
	def transpose(self):
		return Matrix(numpy.transpose(self.data))
	
	def inverse(self):
		if self.data.shape[0] != self.data.shape[1]:
			raise Exceptions.DimensionMismatch()
		return Matrix(numpy.linalg.inv(self.data))

	def pseudoInverse(self):
		return Matrix(numpy.linalg.pinv(self.data))

	def determinant(self):
		if self.data.shape[0] != self.data.shape[1]:
			raise Exceptions.DimensionMismatch()
		return numpy.linalg.det(self.data)
	
	def trace(self):
		return numpy.trace(self.data)
	
	def row(self, i):
		return Vector.Vector(self.data[i])
		
	def column(self, i):
		return Vector.Vector(self.data[:,i])
		
	def mean(self, index = None):
		"""index = 0: collapses rows
		index = 1: collapses columns
		index = None: overall mean"""
		if index is None:
			return numpy.mean(self.data)
		else:
			return numpy.mean(self.data, index)
		
	def solve(self, b):
		"""Solves Ax=b and returns x
		self is A, must be square, and b must have same length
		A should also not be singular"""
		if self.data.shape[0] != self.data.shape[1] or self.data.shape[0] != b.data.shape[0]:
			raise Exceptions.DimensionMismatch()
		if self.determinant() == 0.0:
			raise Exceptions.SingularMatrix()
		return Vector.Vector(numpy.linalg.solve(self.data, b.data))
		
	def leastSquares(self, b, verbose = False):
		"""Solves and overdetermined system of equations (tall a, b, short x returned) 
		Will also take a matrix b, in which case a matrix is returned
		Setting verbose to true returns all the details from the numpy library"""
		if self.data.shape[0] != b.data.shape[0]:
			raise Exceptions.DimensionMismatch()
		xArray, resids, rank, s = numpy.linalg.lstsq(self.data, b.data)
		
		x = None
		if len(xArray.shape) == 1:
			x = Vector.Vector(xArray)
		else:
			x = Matrix(xArray)
		
		if verbose:
			return x, Vector.Vector(resids), rank, Vector.Vector(s)
		return x
																																													
	def __neg__(self):
		return Matrix(- self.data)
	
	def __add__(self, other):
		if other.data.shape[0] != self.data.shape[0] or other.data.shape[1] != self.data.shape[1]:
			raise Exceptions.DimensionMismatch()
		return Matrix(self.data + other.data)
	
	def __sub__(self, other):
		if other.data.shape[0] != self.data.shape[0] or other.data.shape[1] != self.data.shape[1]:
			raise Exceptions.DimensionMismatch()
		return Matrix(self.data - other.data)
	
	def __mul__(self, other):
		t = type(other)
		if t == Matrix:
			if self.data.shape[1] != other.data.shape[0]:
				raise Exceptions.DimensionMismatch()
			return Matrix(numpy.dot(self.data, other.data))
		elif t == Vector.Vector:
			if self.data.shape[1] != other.data.shape[0]:
				raise Exceptions.DimensionMismatch()
			return Vector.Vector(numpy.dot(self.data, other.data))
		elif t in [int, float]:
			return Matrix(other * self.data)
		else:
			raise Exceptions.BadArgumentData()
	
	def __rmul__(self, other):
		t = type(other)
		if t in [int, float]:
			return Matrix(other * self.data)
		else:
			raise Exceptions.BadArgumentData()
	
	def __div__(self, other):
		if type(other) in [int, float]:
			invDenom = 0.0
			if other != 0.0:
				invDenom = 1.0 / other
			return Matrix(invDenom * self.data)
		raise Exceptions.BadArgumentData()
	
	def __getitem__(self, i):
		return self.data[i]
		
	def __setitem__(self, i, value):
		self.data[i] = value
	
	def __str__(self):
		s = '['
		for i in range(self.rowSize() - 1):
			s += '['
			for j in range(self.columnSize() - 1):
				s += str(self.data[i][j]) + ',\t'
			s += str(self.data[i][self.columnSize() - 1]) + ']\n '
		s += '['
		for j in range(self.columnSize() - 1):
			s += str(self.data[self.rowSize() - 1][j]) + ',\t'
		s += str(self.data[self.rowSize() - 1][self.columnSize() - 1]) + ']]'
		return s
	
	def __repr__(self):
		s = 'mmath.Matrix.Matrix(['
		for i in range(self.rowSize() - 1):
			s += '['
			for j in range(self.columnSize() - 1):
				s += str(self.data[i][j]) + ', '
			s += str(self.data[i][self.columnSize() - 1]) + '], '
		s += '['
		for j in range(self.columnSize() - 1):
			s += str(self.data[self.rowSize() - 1][j]) + ', '
		s += str(self.data[self.rowSize() - 1][self.columnSize() - 1]) + ']])'
		return s
	
	

if __name__ == '__main__':
	a = Matrix([[2,-3,2],[1,4,-1],[-1,2,1]])
	#a = Matrix([[1, 2, 3], [1,2,3], [1,2,3]])
	b = Vector.Vector([21,1,17])
	print a 
	print b
	print a.determinant()
	print a.leastSquares(b, verbose = True)
	