#  Copyright (c) 2008 patrick coleman

import Exceptions
import Matrix
import math
import numpy

class Vector(object):
	"""Float vectors of arb length"""

	def __init__(self, data):
		t = type(data) 
		if t == list:
			self.data = numpy.array(data, float)
			return
		elif t == numpy.ndarray:
			self.data = data
			return
		raise Exceptions.BadInitializationData()
	
	def __len__(self):
		return len(self.data)
	
	def __getitem__(self, i):
		return self.data[i]
		
	def __setitem__(self, i, value):
		self.data[i] = value
	
	def __mul__(self, other):
		"""dot product and scalar product"""
		t = type(other)
		if t == Vector:
			return numpy.dot(self.data, other.data)
		elif t in [int, float]:
			return Vector(self.data * other)
		elif t == Matrix.Matrix:
			if self.data.shape[0] != other.data.shape[0]:
				raise Exceptions.DimensionMismatch()
			return Vector(numpy.dot(self.data, other.data))
		else:
			raise Exceptions.BadArgumentData()
	
	def __rmul__(self, other):
		"""premultiplied scalar product"""
		t = type(other)
		if t in [int, float]:
			return Vector(self.data * other)
		else:
			raise Exceptions.BadArgumentData()
	
	def __add__(self, other):
		return Vector(self.data + other.data)
	
	def __sub__(self, other):
		return Vector(self.data - other.data)
	
	def __neg__(self):
		return Vector(- self.data)
	
	def __div__(self, other):
		if type(other) in [int, float]:
			invDenom = 0.0
			if other != 0.0:
				invDenom = 1.0 / other
			return Vector(invDenom * self.data)
		raise Exceptions.BadArgumentData()
	
	def length(self):
		"Euclidean length"
		return math.sqrt(self * self)

	def lengthSquared(self):
		return self * self
	
	def normalize(self):
		invLen = 1.0 / self.length()
		return invLen * self
	
	def __str__(self):
		s = '<'
		for i in range(len(self.data) - 1):
			s = s + str(self.data[i]) + ', '
		if len(self.data) > 0:
			s = s + str(self.data[len(self.data) - 1])
		s = s + '>'
		return s
	
	def __repr__(self):
		s = 'mmath.Vector.Vector('
		for i in range(len(self.data) - 1):
			s = s + str(self.data[i]) + ', '
		if len(self.data) > 0:
			s = s + str(self.data[len(self.data) - 1])
		s = s + ')'
		return s
	
	def sum(self):
		return numpy.sum(self.data)
		
if __name__ == '__main__':
	v = Vector([1, 2, 3])
	m = Matrix.Matrix([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])
	print v
	print m
	print v * m
	