#  Copyright (c) 2008 patrick coleman

import math

from motion.EventExtractors import EventExtractor
import motion.Events as Events
import motion.EventSignals as EventSignals

from jakaltek.incrementalRotation import incrementalRotation as incrementalRotation

class RotationExtremaExtractor(EventExtractor):
	def __init__(self, motion, includeBounds = True):
		EventExtractor.__init__(self, motion)
		self.includeBounds = includeBounds
	
	def eventSignal(self):
		es = EventSignals.EventSignal('rotationEvents',
									  self.motion.startTime,
									  self.motion.endTime,
									  self.motion.attributes)
		
		for attr in self.motion.attributes:
			nodename, attrname = attr.split('.')
			if attrname == 'rotate':
				if self.includeBounds:
					es.addEvent(attr, Events.RotationExtremumEvent(es.startTime))
					es.addEvent(attr, Events.RotationExtremumEvent(es.endTime))
				
				# initial increment
				lastTime = es.startTime
				lastValue = self.motion.attributeValue(attr, lastTime)
				thisTime = es.startTime + self.sampleRate
				thisValue = self.motion.attributeValue(attr, thisTime)
				
				lastIncrementAx, lastIncrementAngle = incrementalRotation(lastValue, thisValue)
				
				# prep for second increment in loop
				lastTime = thisTime
				lastValue = thisValue
				thisTime += self.sampleRate
				
				# candidate events are at lastTime
				while thisTime <= es.endTime:
					thisValue = self.motion.attributeValue(attr, thisTime)
						
					incrementAx, incrementAngle = incrementalRotation(lastValue, thisValue)				
					
					# currently this doesn't consider any held rotation bounds
					if incrementAngle != 0.0 and lastIncrementAngle != 0.0:
						dp = incrementAx * lastIncrementAx
						if dp < 0.0:
							es.addEvent(attr,
										Events.RotationExtremumEvent(lastTime))
					
					#prep for next text
					lastTime = thisTime
					lastValue = thisValue
					lastIncrementAx = incrementAx
					lastIncrementAngle = incrementAngle
					thisTime += self.sampleRate
	

class RotationChangeExtractor(EventExtractor):
	def __init__(self, 
				 motion, 
				 changeThreshold = 90,
				 maximumDuration = 6,
				 includeBounds = True):
		EventExtractor.__init__(self, motion)
		self.changeThreshold = changeThreshold
		self.maximumDuration = maximumDuration
		self.includeBounds = includeBounds
	
	def eventSignal(self):
		es = EventSignals.EventSignal('rotationEvents',
									  self.motion.startTime,
									  self.motion.endTime,
									  self.motion.attributes)
		
		sampleTimes = []
		t = es.startTime
		while t <= es.endTime:
			sampleTimes.append(t)
			t +=  self.sampleRate
		
		for attr in self.motion.attributes:
			nodename, attrname = attr.split('.')
			if attrname == 'rotate':
				if self.includeBounds:
					es.addEvent(attr, Events.RotationChangeEvent(es.startTime, es.startTime, es.startTime))
					es.addEvent(attr, Events.RotationChangeEvent(es.endTime, es.endTime, es.endTime))
				
				attrValues = [self.motion.attributeValue(attr, t) for t in sampleTimes]			
				# makes a list of tuples of (axis, angle)
				attrIncrements = [incrementalRotation(attrValues[i], attrValues[i + 1]) 
									for i in range(len(attrValues) - 1)]
				
				lastEventTime = es.startTime
				for i in range(1, len(attrIncrements)):
					e = self.findEvent(attrIncrements, i, lastEventTime)
					if e != None:
						lastEventTime = e.time
						es.addEvent(attr, e)
					else:
						pass 
		return es
	
	def findEvent(self, increments, i, lastEventTime):
		backStep = 1
		backTime = self.sampleRate
		now = self.motion.startTime + i * self.sampleRate
		while (backTime <= self.maximumDuration 
			and backStep <= i 
			and (now - backTime) >= lastEventTime):
			
			# TODO Revise the following code to handle zero starts and ends
			# this idea has been moved to RotationPoseExtractor instead

			if increments[i][1] != 0.0 and increments[i - backStep][1] != 0.0:
				ang = math.degrees(increments[i][0].angle(increments[i - backStep][0]))
				if ang > self.changeThreshold:
					return Events.RotationChangeEvent(now, now - backTime, now)

			backStep += 1
			backTime = backStep * self.sampleRate

class RotationPoseExtractor(EventExtractor):
	def __init__(self, 
				 motion, 
				 changeThreshold = 90,
				 changeMaximumDuration = 6,
				 zeroMinimumDuration = 1, # about of time to pass to have a zero (i.e. 1 frame or whatever unit)
				 noiseTolerance = 0.001): # angular noise tol in degrees
		EventExtractor.__init__(self, motion)
		self.changeThreshold = changeThreshold
		self.changeMaximumDuration = changeMaximumDuration
		self.zeroMinimumDuration = zeroMinimumDuration
		self.noiseTolerance = noiseTolerance
    
	def eventSignal(self):
		es = EventSignals.EventSignal('rotationEvents',
									  self.motion.startTime,
									  self.motion.endTime,
									  self.motion.attributes)
		times = []
		t = es.startTime
		while t <= es.endTime:
			times.append(t)
			t += self.sampleRate
			
		for attr in self.motion.attributes:
			nodename, attrname = attr.split('.')
			if attrname == 'rotate':
				
				es.addEvent(attr, Events.DataStartEvent(es.startTime))
				es.addEvent(attr, Events.DataEndEvent(es.endTime))
				
				orientations = [self.motion.attributeValue(attr, t) for t in times]			
				increments = [incrementalRotation(orientations[i], orientations[i + 1]) 
							  for i in range(len(orientations) - 1)]	#list of tuples of (axis, angle)
				
				zeroState = False
				if self._zeroStartDetect(times, orientations, 0):
					zeroState = True # No zero events at start
				lastEventIndex = 0
				lastEventTime = 0.0
				
				for i in range(1, len(increments) - 1):
					if zeroState:
						# check for end of a zero state (ZeroEndEvent)
						if self._zeroEndDetect(orientations, i):
							zeroState = False
							es.addEvent(attr, Events.ZeroEndEvent(times[i]))
							lastEventTime = times[i]
							lastEventIndex = i
					else:
						# check for start of a zero state (ZeroStartEvent)
						if self._zeroStartDetect(times, orientations, i):
							zeroState = True
							es.addEvent(attr, Events.ZeroStartEvent(times[i]))
							lastEventTime = times[i]
							lastEventIndex = i
						else:
							# check for rotation change (RotationChangeEvent)
							e = self._rotationChangeDetect(times, increments, i, lastEventTime)
							if e != None:
								lastEventTime = times[i]
								lastEventIndex = i - int((times[i] - lastEventTime) / self.sampleRate)
								es.addEvent(attr, e)
		return es
	
	def _zeroStartDetect(self, times, orientations, i):
		"""Determines if index i is the start of a zero state in (times, orientations)
		Zero states have, over a range of time:
			starting with i
			of at least length self.zeroMinimumDuration
		all angles within self.noiseTolerance of each other
		Returns True/False otherwise"""
		
		checkableEndIndex = self._indexOfZeroDurationEnd(times, i)
		if checkableEndIndex == len(times):
			return False
		else:
			return self._orientationsAreEqual(orientations[i:checkableEndIndex + 1])
	
	def _zeroEndDetect(self, orientations, i):
		"""Determines if index i is the last sample in a zero state
		Requires i+1 samples"""
		return (not self._orientationsAreEqual(orientations[i:i + 2]))
	
	def _rotationChangeDetect(self, times, increments, i, lastEventTime):
		"""Determines sample i causes a rotation change event relative to the prior samples
		Doesn't look past lastEventTime
		Doesn't look past self.changeMaximumDuration
		Rotation changes are considered to have an axis change >= self.changeThreshold"""
			
		backStep = 1
		backTime = self.sampleRate
		while (backTime <= self.changeMaximumDuration 
			and backStep <= i 
			and (times[i] - backTime) >= lastEventTime):
			
			if increments[i][1] != 0.0 and increments[i - backStep][1] != 0.0:
				ang = math.degrees(increments[i][0].angle(increments[i - backStep][0]))
				if ang > self.changeThreshold:
					return Events.RotationChangeEvent(times[i] - 0.5 * backTime,
													  times[i] - backTime, 
													  times[i])

			backStep += 1
			backTime = backStep * self.sampleRate
			
	def _orientationsAreEqual(self, orientations):
		"""Determines if all orientations in orientations are within self.noiseTolerance
		of each other"""
		for oi in orientations:
			for oj in orientations:
				if oi != oj:
					inc = incrementalRotation(oi, oj)
					if math.fabs(inc[1]) > self.noiseTolerance:
						return False
		return True

	def _indexOfZeroDurationEnd(self, times, i):
		"""Returns the index of the last entry of times such that the 
		duration from i to that index is >= self.zeroMinimumDuration
		Requires i < len(times)
		Returns len(times) if none"""
		for j in range((i + 1), len(times)):
			if (times[j] - times[i]) >= self.zeroMinimumDuration:
				return j
		return len(times)