# copyright 2008 patrick coleman

import math

import maya.cmds as mc

import mmaya
import msignal.NumericFilters
import mmath.Vector3
import geom.Rotations

# these next two ops should be migrated into a VectorRotation class at some point

def logMap(qr):
	"""maps a QRotation to r3, not numericaly robust"""
	theta = 2.0 * math.acos(qr.q.s)
	vNorm = mmath.Vector3.Vector3()
	if qr.q.v.length() > 0.0:
		vNorm = qr.q.v.normalize()
	return theta * vNorm
	
def expMap(v):
	"""maps an r3 rotation to a QRotation, not numericaly robust"""
	theta = v.length()
	vNorm = mmath.Vector3.Vector3(1.0) # doesn't much matter what it is if theta close to 0
	if theta > 0.0:
		vNorm = v.normalize()
	return geom.Rotations.QRotation(vNorm, theta)

def filterRotation(node, startTime = None, endTime = None, dt = 1.0, filterWidth = 5):
	"""filters a single rotations from start time to end time with a tent filter"""
	
	# defaults
	if startTime is None:
		startTime = mmaya.sceneStartTime()
	if endTime is None:
		endTime = mmaya.sceneEndTime()

	rotateOrder = mmaya.rotateOrderString(mc.getAttr(node + '.rotateOrder'))

	# get signal as R3 logmap parameterization
	t = startTime
	signal = []
	times = []
	while t <= endTime:
		r = mc.getAttr(node + '.rotate', t = t)[0]
		er = geom.Rotations.EulerRotation(r[0], r[1], r[2], order = rotateOrder, 
										  local = False, degrees = True)
		qr = er.qRotation()
		signal.append(logMap(qr))
		times.append(t)
		t += dt
	
	# apply the filter
	f = msignal.NumericFilters.TentFilter(signal, filterWidth, 
										  zero = mmath.Vector3.Vector3())
	signal = f.output()
	
	# write the data back 
	for i in range(len(signal)):
		qr = expMap(signal[i])
		er = qr.eulerRotation(order = rotateOrder, local = False)
		mc.setKeyframe(node + '.rotateX', t = times[i], v = math.degrees(er.rx))
		mc.setKeyframe(node + '.rotateY', t = times[i], v = math.degrees(er.ry))
		mc.setKeyframe(node + '.rotateZ', t = times[i], v = math.degrees(er.rz))

def filterRotations(nodes = None, recursive = False, includeTransforms = False, 
					startTime = None, endTime = None, dt = 1.0, filterWidth = 5):
	"""filters a collection of nodes by selection, with options for recursion and 
	joints only"""
	
	# get defaults
	if nodes is None:
		nodes = mmaya.selectedTransforms()
	if startTime is None:
		startTime = mmaya.sceneStartTime()
	if endTime is None:
		endTime = mmaya.sceneEndTime()
		
	# get full list of nodes to filter
	if recursive:
		allNodes = []
		for n in nodes:
			allNodes = allNodes + mmaya.getHierarchy(n, includeTransforms)
		nodes = allNodes
	
	# filter
	for n in nodes:
		filterRotation(n, startTime, endTime, dt, filterWidth)
	