'''code for plotting static ode's.  could be extended to time domain.  
mostly from baraff/witkin notes.  in practice, this code could be really optimized'''

import math

import mmaya
import geom.Point3

def eulerPlot2dOde(f, p0, h = 1.0, numSteps = 15, z = 0.0, name = 'odePlot'):
	'''
	f evaluates a vector field as a Vector2 at a given x, y
	p0 is a Point2 starting point
	runs euler stepping with step size h and makes a maya curve
	O(h^2) error
	'''
	
	pts = [geom.Point3.Point3(p0.x, p0.y, z)]
	p = p0

	for i in range(numSteps):
		p = p + h * f(p.x, p.y)
		pts.append(geom.Point3.Point3(p.x, p.y, z))
	
	mmaya.makeCurve(pts, name)

def midpointPlot2dOde(f, p0, h = 1.0, numSteps = 15, z = 0.0, name = 'odePlot'):
	'''
	f evaluates a vector field as a Vector2 at a given x, y
	p0 is a Point2 starting point
	runs midpoint method with step size h and makes a maya curve
	O(h^3) error
	'''
	
	pts = [geom.Point3.Point3(p0.x, p0.y, z)]
	p = p0

	for i in range(numSteps):
		pHalf = p + 0.5 * h * f(p.x, p.y)
		p = p + h * f(pHalf.x, pHalf.y)
		pts.append(geom.Point3.Point3(p.x, p.y, z))
	
	mmaya.makeCurve(pts, name)

def symplecticEulerPlot2dOde(f, p0, h = 1.0, numSteps = 15, z = 0.0, name = 'odePlot'):
	'''
	f evaluates a vector field as a Vector2 at a given x, y
	p0 is a Point2 starting point
	runs symplectic Euler method with step size h and makes a maya curve
	error??
	better energy conservation that explicit or implicit
	'''
	
	pts = [geom.Point3.Point3(p0.x, p0.y, z)]
	p = p0

	
	for i in range(numSteps):
		# estimate a as 
		
		
		pHalf = p + 0.5 * h * f(p.x, p.y)
		p = p + h * f(pHalf.x, pHalf.y)
		pts.append(geom.Point3.Point3(p.x, p.y, z))
	
	mmaya.makeCurve(pts, name)
	
def rk4Plot2dOde(f, p0, h = 1.0, numSteps = 15, z = 0.0, name = 'odePlot'):
	'''
	f evaluates a vector field as a Vector2 at a given x, y
	p0 is a Point2 starting point
	runs order 4 Runge-Kutta method with step size h and makes a maya curve
	O(h^5) error
	'''
	
	pts = [geom.Point3.Point3(p0.x, p0.y, z)]
	p = p0
	oneThird = 1.0 / 3.0
	oneSixth = 1.0 / 6.0

	for i in range(numSteps):
		k1 = h * f(p.x, p.y) # t
		k2 = h * f(p.x + 0.5 * k1.x, p.y + 0.5 * k1.y) # t + h/2
		k3 = h * f(p.x + 0.5 * k2.x, p.y + 0.5 * k2.y) # t + h/2
		k4 = h * f(p.x + k3.x, p.y + k3.y) # t + h
		
		p = p + oneSixth * k1 + oneThird * k2 + oneThird * k3 + oneSixth * k4
		pts.append(geom.Point3.Point3(p.x, p.y, z))
	
	mmaya.makeCurve(pts, name)

def adaptiveEulerPlot2dOde(f, p0, h = 1.0, stepTol = 0.01, 
						   numSteps = 15, z = 0.0, name = 'odePlot'):
	'''
	f evaluates a vector field as a Vector2 at a given x, y
	p0 is a Point2 starting point
	runs euler stepping with step size h and makes a maya curve
	O(h^2) error
	'''
	
	pts = [geom.Point3.Point3(p0.x, p0.y, z)]
	p = p0
	lastH = h

	for i in range(numSteps):
		p1 = p + lastH * f(p.x, p.y)
		p2 = p + 0.5 * lastH * f(p.x, p.y)
		err = p1.distance(p2)
		lastH = lastH * math.sqrt(stepTol/err)
		p = p + lastH * f(p.x, p.y)
		pts.append(geom.Point3.Point3(p.x, p.y, z))
	
	mmaya.makeCurve(pts, name)
	
def adaptiveMidpointPlot2dOde(f, p0, h = 1.0, stepTol = 0.01, 
							  numSteps = 15, z = 0.0, name = 'odePlot'):
	'''
	f evaluates a vector field as a Vector2 at a given x, y
	p0 is a Point2 starting point
	runs midpoint method with step size h and makes a maya curve
	O(h^3) error
	'''
	
	pts = [geom.Point3.Point3(p0.x, p0.y, z)]
	p = p0
	hLast = h
	oneThird = 1.0 / 3.0

	for i in range(numSteps):
		p1Half = p + 0.5 * hLast * f(p.x, p.y)
		p1 = p + hLast * f(p1Half.x, p1Half.y)
		p2Half = p + 0.25 * hLast * f(p.x, p.y)
		p2 = p + 0.5 * hLast * f(p2Half.x, p2Half.y)
		err = p1.distance(p2)
		hLast = hLast * math.pow(stepTol/err, oneThird)
		
		pHalf = p + 0.5 * hLast * f(p.x, p.y)
		p = p + hLast * f(pHalf.x, pHalf.y)
		
		pts.append(geom.Point3.Point3(p.x, p.y, z))
	
	mmaya.makeCurve(pts, name)
	
def adaptiveRk4Plot2dOde(f, p0, h = 1.0, stepTol = 0.01,
				 numSteps = 15, z = 0.0, name = 'odePlot'):
	'''
	f evaluates a vector field as a Vector2 at a given x, y
	p0 is a Point2 starting point
	runs order 4 Runge-Kutta method with step size h and makes a maya curve
	O(h^5) error
	'''
	
	pts = [geom.Point3.Point3(p0.x, p0.y, z)]
	p = p0
	oneThird = 1.0 / 3.0
	oneSixth = 1.0 / 6.0
	hLast = h

	for i in range(numSteps):
		k1 = hLast * f(p.x, p.y) # t
		k2 = hLast * f(p.x + 0.5 * k1.x, p.y + 0.5 * k1.y) # t + h/2
		k3 = hLast * f(p.x + 0.5 * k2.x, p.y + 0.5 * k2.y) # t + h/2
		k4 = hLast * f(p.x + k3.x, p.y + k3.y) # t + h
		
		p1 = p + oneSixth * k1 + oneThird * k2 + oneThird * k3 + oneSixth * k4
		
		k1 = 0.5 * hLast * f(p.x, p.y) # t
		k2 = 0.5 * hLast * f(p.x + 0.5 * k1.x, p.y + 0.5 * k1.y) # t + h/2
		k3 = 0.5 * hLast * f(p.x + 0.5 * k2.x, p.y + 0.5 * k2.y) # t + h/2
		k4 = 0.5 * hLast * f(p.x + k3.x, p.y + k3.y) # t + h
		
		p2 = p + oneSixth * k1 + oneThird * k2 + oneThird * k3 + oneSixth * k4
		
		err = p1.distance(p2)
		hLast = hLast * math.pow(stepTol/err, 0.2)
		
		k1 = hLast * f(p.x, p.y) # t
		k2 = hLast * f(p.x + 0.5 * k1.x, p.y + 0.5 * k1.y) # t + h/2
		k3 = hLast * f(p.x + 0.5 * k2.x, p.y + 0.5 * k2.y) # t + h/2
		k4 = hLast * f(p.x + k3.x, p.y + k3.y) # t + h
		
		p = p + oneSixth * k1 + oneThird * k2 + oneThird * k3 + oneSixth * k4
		pts.append(geom.Point3.Point3(p.x, p.y, z))
	
	mmaya.makeCurve(pts, name)
	
	