#  Copyright (c) 2008 patrick coleman

from Exceptions import *
from mmath.Matrix4 import *

def filmFit(value):
	"""Convenience function to convert maya attribute to string for 
	ProjectionMatrix constructor"""
	if value == 0:
		return 'fill'
	elif value == 1:
		return 'horizontal'
	elif value == 2:
		return 'vertical'
	else:
		return 'overscan'

class ProjectionMatrix(Matrix4):
	def __init__(self, 
				 focalLength, 
				 nearClipPlane, 
				 farClipPlane,
				 horizontalFilmAperture,
				 verticalFilmAperture,
				 renderDimensions,
				 filmFit,
				 filmFitOffset,
				 overscan,
				 cameraScale,
				 horizontalFilmOffset,
				 verticalFilmOffset):
		"""Creates a maya projection matrix.  Constructor parameters are named to match the
		corresponding maya camera attributes.  renderDimensions should by a 2-tuple of ints.
		filmFit should be one of 'fill', 'horizontal', 'vertical', or 'overscan'.  The
		rest are floats.  renderDimensions should be set to defaultResolution.{width, height}
		Assumes maya is set to cm (affects near & far)"""
		Matrix4.__init__(self)
		self.focalLength = focalLength / 25.4 # mm to in
		self.nearClipPlane = nearClipPlane
		self.farClipPlane = farClipPlane
		self.horizontalFilmAperture = horizontalFilmAperture
		self.verticalFilmAperture = verticalFilmAperture
		self.renderDimensions = renderDimensions
		self.filmFit = filmFit
		self.filmFitOffset = filmFitOffset
		self.overscan = overscan
		self.cameraScale = cameraScale
		self.horizontalFilmOffset = horizontalFilmOffset
		self.verticalFilmOffset = verticalFilmOffset
		self.t = 0.0
		self.b = 0.0
		self.l = 0.0
		self.r = 0.0
		self.n = self.nearClipPlane
		self.f = self.farClipPlane
		self._computeData()
	
	def _computeData(self):
		"""Subclasses should implement this to write to self.data, self.{l,r,t,b}"""
		raise AbstractMethod()

class MayaProjectionMatrix(ProjectionMatrix):
	def _computeData(self):
		"""This implementation is based on the paper 'Cameras Made Extra Difficult 
		[A White Paper on the Maya Perspective Camera]' by Chris Thorne, assumed to 
		be from Alias, with a nontrivial bunch of extra reverse engineering"""
		windowAspectRatio = (float(self.renderDimensions[0]) /
			float(self.renderDimensions[1]))
		filmAspectRatio = (float(self.horizontalFilmAperture) /
			float(self.verticalFilmAperture))
		focalToNear = self.cameraScale * self.nearClipPlane / self.focalLength
		
		sx = 1.0
		sy = 1.0
		tx = 0.0
		ty = 0.0
		if self.filmFit == 'fill':
			if windowAspectRatio < filmAspectRatio:
				sx = windowAspectRatio / filmAspectRatio
			else:
				sy = filmAspectRatio / windowAspectRatio
		elif self.filmFit == 'horizontal':
			sy = filmAspectRatio / windowAspectRatio
			if sy > 1.0:
				ty = 0.5 * self.filmFitOffset * (self.verticalFilmAperture -
					self.verticalFilmAperture * sy)
		elif self.filmFit == 'vertical':
			sx = windowAspectRatio / filmAspectRatio
			if sx > 1.0:
				tx = 0.5 * self.filmFitOffset * (self.horizontalFilmAperture - 
					self.horizontalFilmAperture * sx)
		else: # 'overscan'
			if windowAspectRatio < filmAspectRatio:
				sy = filmAspectRatio / windowAspectRatio
			else:
				sx = windowAspectRatio / filmAspectRatio
		
		self.l = focalToNear * (-0.5 * self.horizontalFilmAperture  * sx 
			+ self.horizontalFilmOffset + tx)
		self.r = focalToNear * ( 0.5 * self.horizontalFilmAperture  * sx 
			+ self.horizontalFilmOffset + tx)
		self.b = focalToNear * (-0.5 * self.verticalFilmAperture  * sy 
			+ self.verticalFilmOffset + ty)
		self.t = focalToNear * ( 0.5 * self.verticalFilmAperture  * sy 
			+ self.verticalFilmOffset + ty)
		
		# compute matrix
		invRMinusL = 1.0 / (self.r - self.l)
		invTMinusB = 1.0 / (self.t - self.b)
		invFMinusN = 1.0 / (self.f - self.n)
		
		self.data[0][0] = 2.0 * self.n * invRMinusL
		self.data[0][1] = 0.0
		self.data[0][2] = 0.0
		self.data[0][3] = 0.0
		
		self.data[1][0] = 0.0
		self.data[1][1] = 2.0 * self.n * invTMinusB
		self.data[1][2] = 0.0
		self.data[1][3] = 0.0
		
		self.data[2][0] = (self.r + self.l) * invRMinusL
		self.data[2][1] = (self.t + self.b) * invTMinusB
		self.data[2][2] = (self.n + self.f) * invFMinusN
		self.data[2][3] = -1.0
		
		self.data[3][0] = 0.0
		self.data[3][1] = 0.0
		self.data[3][2] = 2.0 * self.f * self.n * invFMinusN
		self.data[3][3] = 0.0

class MayaViewProjectionMatrix(ProjectionMatrix):
	def _computeData(self):
		"""Variant of the maya projection matrix used for interactive views.  It handles
		overscan differently.  More reverse engineering.  renderDimensions should be set
		to M3dView.activeView.{portWidth(), portHeight()}!!"""
		windowAspectRatio = (float(self.renderDimensions[0]) /
			float(self.renderDimensions[1]))
		filmAspectRatio = (float(self.horizontalFilmAperture) /
			float(self.verticalFilmAperture))
		focalToNear = self.cameraScale * self.nearClipPlane / self.focalLength
		
		sx = 1.0
		sy = 1.0
		tx = 0.0
		ty = 0.0
		if self.filmFit == 'fill':
			if windowAspectRatio < filmAspectRatio:
				sx = windowAspectRatio / filmAspectRatio
			else:
				sy = filmAspectRatio / windowAspectRatio
		elif self.filmFit == 'horizontal':
			sy = filmAspectRatio / windowAspectRatio
			if sy > 1.0:
				ty = 0.5 * self.filmFitOffset * (self.verticalFilmAperture -
					self.verticalFilmAperture * sy)
		elif self.filmFit == 'vertical':
			sx = windowAspectRatio / filmAspectRatio
			if sx > 1.0:
				tx = 0.5 * self.filmFitOffset * (self.horizontalFilmAperture - 
					self.horizontalFilmAperture * sx)
		else: # 'overscan'
			if windowAspectRatio < filmAspectRatio:
				sy = filmAspectRatio / windowAspectRatio
			else:
				sx = windowAspectRatio / filmAspectRatio
		
		sx = self.overscan * sx
		sy = self.overscan * sy
		
		self.l = focalToNear * (-0.5 * self.horizontalFilmAperture  * sx 
			+ self.horizontalFilmOffset + tx)
		self.r = focalToNear * ( 0.5 * self.horizontalFilmAperture  * sx 
			+ self.horizontalFilmOffset + tx)
		self.b = focalToNear * (-0.5 * self.verticalFilmAperture  * sy 
			+ self.verticalFilmOffset + ty)
		self.t = focalToNear * ( 0.5 * self.verticalFilmAperture  * sy 
			+ self.verticalFilmOffset + ty)
		
		# compute matrix
		invRMinusL = 1.0 / (self.r - self.l)
		invTMinusB = 1.0 / (self.t - self.b)
		invFMinusN = 1.0 / (self.f - self.n)
		
		self.data[0][0] = 2.0 * self.n * invRMinusL
		self.data[0][1] = 0.0
		self.data[0][2] = 0.0
		self.data[0][3] = 0.0
		
		self.data[1][0] = 0.0
		self.data[1][1] = 2.0 * self.n * invTMinusB
		self.data[1][2] = 0.0
		self.data[1][3] = 0.0
		
		self.data[2][0] = (self.r + self.l) * invRMinusL
		self.data[2][1] = (self.t + self.b) * invTMinusB
		self.data[2][2] = (self.n + self.f) * invFMinusN
		self.data[2][3] = -1.0
		
		self.data[3][0] = 0.0
		self.data[3][1] = 0.0
		self.data[3][2] = 2.0 * self.f * self.n * invFMinusN
		self.data[3][3] = 0.0

class MayaGLProjectionMatrix(MayaViewProjectionMatrix):
	def _computeData(self):
		windowAspectRatio = (float(self.renderDimensions[0]) /
			float(self.renderDimensions[1]))
		filmAspectRatio = (float(self.horizontalFilmAperture) /
			float(self.verticalFilmAperture))
		focalToNear = self.cameraScale * self.nearClipPlane / self.focalLength
		
		sx = 1.0
		sy = 1.0
		tx = 0.0
		ty = 0.0
		if self.filmFit == 'fill':
			if windowAspectRatio < filmAspectRatio:
				sx = windowAspectRatio / filmAspectRatio
			else:
				sy = filmAspectRatio / windowAspectRatio
		elif self.filmFit == 'horizontal':
			sy = filmAspectRatio / windowAspectRatio
			if sy > 1.0:
				ty = 0.5 * self.filmFitOffset * (self.verticalFilmAperture -
					self.verticalFilmAperture * sy)
		elif self.filmFit == 'vertical':
			sx = windowAspectRatio / filmAspectRatio
			if sx > 1.0:
				tx = 0.5 * self.filmFitOffset * (self.horizontalFilmAperture - 
					self.horizontalFilmAperture * sx)
		else: # 'overscan'
			if windowAspectRatio < filmAspectRatio:
				sy = filmAspectRatio / windowAspectRatio
			else:
				sx = windowAspectRatio / filmAspectRatio
		
		sx = self.overscan * sx
		sy = self.overscan * sy
		
		self.l = focalToNear * (-0.5 * self.horizontalFilmAperture  * sx 
			+ self.horizontalFilmOffset + tx)
		self.r = focalToNear * ( 0.5 * self.horizontalFilmAperture  * sx 
			+ self.horizontalFilmOffset + tx)
		self.b = focalToNear * (-0.5 * self.verticalFilmAperture  * sy 
			+ self.verticalFilmOffset + ty)
		self.t = focalToNear * ( 0.5 * self.verticalFilmAperture  * sy 
			+ self.verticalFilmOffset + ty)
		
		# compute matrix
		invRMinusL = 1.0 / (self.r - self.l)
		invTMinusB = 1.0 / (self.t - self.b)
		invFMinusN = 1.0 / (self.f - self.n)
		
		self.data[0][0] = 2.0 * self.n * invRMinusL
		self.data[0][1] = 0.0
		self.data[0][2] = 0.0
		self.data[0][3] = 0.0
		
		self.data[1][0] = 0.0
		self.data[1][1] = 2.0 * self.n * invTMinusB
		self.data[1][2] = 0.0
		self.data[1][3] = 0.0
		
		self.data[2][0] = (self.r + self.l) * invRMinusL
		self.data[2][1] = (self.t + self.b) * invTMinusB
		self.data[2][2] = - (self.n + self.f) * invFMinusN
		self.data[2][3] = -1.0
		
		self.data[3][0] = 0.0
		self.data[3][1] = 0.0
		self.data[3][2] = - 2.0 * self.f * self.n * invFMinusN
		self.data[3][3] = 0.0

if __name__ == '__main__':
	fl = 35
	n = 0.1
	f = 1000
	hfa = 1.41732
	vfa = 0.94488
	rd = (640, 480)
	ff = 1
	ffo = 0.0
	os = 1.3
	cs = 1.0
	hfo = 0.0
	vfo = 0.0
	mpm = MayaProjectionMatrix(fl, n, f, hfa, vfa, rd, filmFit(ff), ffo, os, cs, hfo, vfo)
	print mpm.data[0]
	print mpm.data[1]
	print mpm.data[2]
	print mpm.data[3]
	
	mgpm = MayaViewProjectionMatrix(fl, n, f, hfa, vfa, rd, filmFit(ff), ffo, os, cs, hfo, vfo)
	print mgpm.data[0]
	print mgpm.data[1]
	print mgpm.data[2]
	print mgpm.data[3]
	
	
	