#  Copyright (c) 2008 patrick coleman

from motion import StreamframeCollections as mStreamframeCollections
import maya.cmds as mc
import motion.Streamframe
import mmaya.Streamframe
from mmath.Vector3 import Vector3
from mmath.Vector4 import Vector4
from mmath.Matrix4 import Matrix4
from geom.Point3 import Point3
from mmaya.Node import Node

class MChildSetCreator(mStreamframeCollections.StreamframeSetCreator):
	"""looks for maya children and creates streamframes using the children's 
	pivots as the first frame axis.  frames are created from the pivots at the 
	start time of motion provided"""

	def __init__(self, motion, attribute, normalize = True):
		mStreamframeCollections.StreamframeSetCreator.__init__(self, motion, attribute)
		self.normalize = normalize
		
	def streamframes(self):
		"Returns a list of streamframes"
		transform, attr = self.attribute.split('.')
		if attr != 'rotate':
			raise UnsupportedData()
		
		sfs = []
		
		# children
		kids = Node(transform).children()
		kids = [k.name for k in kids]
		
		for k in kids:
			f = self._constructFrame(k, self.motion.startTime)
			if f is not None:
				eval = mmaya.Streamframe.MRotatingFrameEvaluator(transform, f, normalize = self.normalize)
				sfCreator = motion.Streamframe.StreamframeCreator(self.motion.startTime, 
															  self.motion.endTime,
															  eval)
				sf = sfCreator.streamframe()
				sfs.append(sf)
		return sfs
	
	def _constructFrame(node, time):
		"""returns none if child length is zero"""
		
		# get kid pivot in local space and ue as axis x
		cPivot = mc.xform(node, rp = True, q = True)
		cpVector = Vector4(cPivot + [1])
		cm = mc.getAttr(node + '.m', t = time)
		cmMatrix = Matrix4(cm)
		xformedPivot = cmMatrix.transpose() * cpVector
		axisX = Point3(xformedPivot[0], xformedPivot[1], xformedPivot[2])
		
		# test against zero length, defualt to x,y,z in local space
		axisXVec = Vector3(axisX.x, axisX.y, axisX.z)
		axisLen = axisXVec.length()
		if axisLen == 0.0:
			return motion.Streamframe.Frame(Point3(1, 0, 0), Point3(0, 1, 0), Point3(0, 0, 1))
		
		# get second axis as that cp with local axis with min(abs(dot prod))
		xVec = Vector3(1, 0, 0)
		yVec = Vector3(0, 1, 0)
		zVec = Vector3(0, 0, 1)
		dps = (axisXVec * xVec, axisXVec * yVec, axisXVec * zVec)
		furthestVec = xVec
		if dps[1] < dps[0] and dps[1] < dps[2]:
			furthestVec = yVec
		elif dps[2] < dps[0] and dps[1] < dps[0]:
			furthestVec = zVec
		axisYVec = axisXVec ^ furthestVec
		axisYVec = axisLen * axisYVec.normalize()
		axisY = Point3(axisYVec.x, axisYVec.y, axisYVec.z)
		 
		# third axis as cp of first two
		axisZVec = axisXVec ^ axisYVec
		axisZVec = axisLen * axisZVec.normalize()
		axisZ = Point3(axisZVec.x, axisZVec.y, axisZVec.z)
		
		return motion.Streamframe.Frame(axisX, axisY, axisZ)
	_constructFrame = staticmethod(_constructFrame)