#  Copyright (c) 2008 patrick coleman

import maya.cmds as mc
import motion.Streamline
import mmath.Matrix4
from geom.Point3 import *

import Node
import mmaya

class SpaceEvaluator(motion.Streamline.Evaluator):
	"""streamline evaluator for points in a given space
	   transform is the space the point is provided in
	   point is the point to trace (defaults to the pivot)
	       currently only 'pivot' is supported
	   space is the space to transform the point to, defualts to world
	       currently only supports 'world' and 'parent'"""
		   
	def __init__(self, transform, point = 'pivot', space = 'world'):
		motion.Streamline.Evaluator.__init__(self)
		
		self.parent = None
		if space == 'parent':
			self.parent = Node.Node(transform).parent() # returns None if none
			
		self.transform = transform
				
	def position(self, time):
		if self.parent == None: #world
			rPivot = mc.xform(self.transform, rp = True, q = True)
			rpVector = mmath.Vector4.Vector4(rPivot + [1])
			m = mc.getAttr(self.transform + '.wm[0]', t = time)
			mMatrix = mmath.Matrix4.Matrix4(m)
			xformedPivot = mMatrix.transpose() * rpVector
			return Point3(xformedPivot[0],
									  xformedPivot[1],
									  xformedPivot[2])
		else:
			m = mc.getAttr(self.transform + '.m', t = time)
			pm = mc.getAttr(self.parent.name + '.m', t = time)
			mMatrix = mmath.Matrix4.Matrix4(m)
			pmMatrix = mmath.Matrix4.Matrix4(pm)
			parentSpaceMat = mMatrix * pmMatrix
			m = parentSpaceMat.array()
			return Point3(m[12], m[13], m[14])
		
class ChildEvaluator(motion.Streamline.Evaluator):
	"""This evaluator applies the rotation component of the transformation at time t.  So the point 
	returned is local with respect to translation, but parallel with respect to rotation, so a point
	constraint will carry the point along correctly.  Yeah, a little confusing."""
	
	def __init__(self, 
				 transform, 
				 childIndex = 0, 
				 defaultPoint = Point3(0, -1, 0), 
				 normalize = True, 
				 radius = 1.0):
		
		motion.Streamline.Evaluator.__init__(self)
		self.transform = transform
		self.defaultPoint = defaultPoint
		self.normalize = normalize
		self.radius = radius
		
		self.child = None
		kids = Node.Node(transform).children()
		if len(kids) > childIndex:
			self.child = kids[childIndex]
	
	def position(self, time):
		ptVector = mmath.Vector4.Vector4([self.defaultPoint.x, self.defaultPoint.y, self.defaultPoint.z, 1.0])
		
		if self.child != None:
			# change pt to trace to pivot of child node in local space
			cPivot = mc.xform(self.child.name, rp = True, q = True)
			cpVector = mmath.Vector4.Vector4(cPivot + [1])
			cm = mc.getAttr(self.child.name + '.m', t = time)
			cmMatrix = mmath.Matrix4.Matrix4(cm)
			xformedPivot = cmMatrix.transpose() * cpVector
			ptVector = xformedPivot
		
		# transform based on transformation at time t, using only rotation
		m = mc.getAttr(self.transform + '.m', t = time)
		mMatrix = mmath.Matrix4.Matrix4(m)
		mRotateMatrix = mMatrix.transpose()
		mRotateMatrix[0][3] = 0
		mRotateMatrix[1][3] = 0
		mRotateMatrix[2][3] = 0
		xformedPt = mRotateMatrix * ptVector
		rp = Point3(xformedPt[0], xformedPt[1], xformedPt[2])
		
		if self.normalize:
			# still needs to correct for world space transform on this node
			pivot = mc.xform(self.transform, rp = True, q = True) # local space			
			
			# pivot from local to parallel space w.r.t rotation only
			pivotVec = mmath.Vector4.Vector4(pivot + [1])
			pivotVec = mRotateMatrix * pivotVec
			pivotPt = Point3(pivotVec[0], pivotVec[1], pivotVec[2])
			
			rpVec = rp - pivotPt
			if rpVec.lengthSquared() > 0.0:
				# ignore normalization for no translation away, this creates degenerate streamlines anyway
				rpVec = rpVec.normalize()
				rpVec = self.radius * rpVec
				rp = pivotPt + rpVec
		
		return rp

class PointEvaluator(motion.Streamline.Evaluator):
	"""This evaluator applies the rotation component of the transformation at time t.  So the point 
	returned is local with respect to translation, but parallel with respect to rotation, so a point
	constraint will carry the point along correctly.  Yeah, a little confusing."""
	
	def __init__(self, 
				 transform, 
				 point = Point3(0, -1, 0), 
				 normalize = True, 
				 radius = 1.0):
		
		motion.Streamline.Evaluator.__init__(self)
		self.transform = transform
		self.point = point
		self.normalize = normalize
		self.radius = radius
	
	def position(self, time):
		ptVector = mmath.Vector4.Vector4([self.defaultPoint.x, self.defaultPoint.y, self.defaultPoint.z, 1.0])
		
		# transform based on transformation at time t, using only rotation
		m = mc.getAttr(self.transform + '.m', t = time)
		mMatrix = mmath.Matrix4.Matrix4(m)
		mRotateMatrix = mMatrix.transpose()
		mRotateMatrix[0][3] = 0
		mRotateMatrix[1][3] = 0
		mRotateMatrix[2][3] = 0
		xformedPt = mRotateMatrix * ptVector
		rp = Point3(xformedPt[0], xformedPt[1], xformedPt[2])
		
		if self.normalize:
			# still needs to correct for world space transform on this node
			pivot = mc.xform(self.transform, rp = True, q = True) # local space			
			
			# pivot from local to parallel space w.r.t rotation only
			pivotVec = mmath.Vector4.Vector4(pivot + [1])
			pivotVec = mRotateMatrix * pivotVec
			pivotPt = Point3(pivotVec[0], pivotVec[1], pivotVec[2])
			
			rpVec = rp - pivotPt
			if rpVec.lengthSquared() > 0.0:
				# ignore normalization for no translation away, this creates degenerate streamlines anyway
				rpVec = rpVec.normalize()
				rpVec = self.radius * rpVec
				rp = pivotPt + rpVec
		
		return rp
	
class TranslationEvaluator(motion.Streamline.Evaluator):
	"just translates the given point around by the local translation"
	def __init__(self, transform, point):
		motion.Streamline.Evaluator.__init__(self)
		self.transform = transform
		self.point = point

	def position(self, time):
		t = mc.getAttr(self.transform + '.translate', t = time)[0]
		return Point3(t[0], t[1], t[2]) + self.point
		
							
def test():
	sel = mc.ls(selection = True)
	if sel == None:
		print 'select something!'
		return
	else:
		sel = sel[0]
	
	te = TranslationEvaluator(sel, Point3(0, 0, 0))
	#ce = ChildEvaluator(sel, 1)
	sc = motion.Streamline.StreamlineCreator(1, 25, te)
	sl = sc.streamline()
	crv = mmaya.makeCurve(sl.points, 'bobbo')
	#mc.pointConstraint(sel, crv)
	mc.select(sel)
		
#test()