#  Copyright (c) 2008 patrick coleman

import math

import maya.cmds as mc
import maya.mel as mm
import maya.OpenMaya as om

import mmath
import geom.Rotations
from geom.Point3 import Point3
import Node

class BadSelection(Exception): pass

def makeCurve(pts, name, knots = None, d = 1):
	"makes a maya curve out of a list of {Point3,Vector3,Normal3}"
	mpts = []
	for p in pts:
		mpts.append((p.x, p.y, p.z))
	if knots == None:
		return mc.curve(point = mpts, d = d, n = name)
	else:
		return mc.curve(point = mpts, k = knots, d = d, n = name)
	
def loft(ptPairs, name):
	constructionCurves = []
	for pPair in ptPairs:
		constructionCurves.append(makeCurve(pPair, 'loftConsructionCurve#'))
	surf = mc.loft(constructionCurves, d = 1, constructionHistory = False, n = name)
	mc.delete(constructionCurves)
	return surf

def plotSignal(signal, name = 'signal', x0 = 0.0, y0 = 0.0, xStep = 0.1, yScale = 1.0, z = 0.0):
	pts = []
	for i in range(len(signal)):
		pts.append(Point3(x0 + i * xStep, y0 + yScale * signal[i], z))
	return makeCurve(pts, name)

def selection():
	"Returns a list of the current selection"
	sel = mc.ls(selection = True)
	if sel == None:
		return []
	return sel

def selectedTransforms():
	sel = mc.ls(selection = True)
	if sel == None: return []
	selTrans = []
	for s in sel:
		if mc.nodeType(s) in ['transform', 'joint']:
			selTrans.append(s)
	return selTrans

def getHierarchy(node, includeTransforms = True):
	nodeType = mc.nodeType(node)
	nodes = []
	if (nodeType == 'joint' or (nodeType == 'transform' and includeTransforms)):
		nodes.append(node)
		
		children = mc.listRelatives(node, children = True)
		if children != None:
			for c in children:
				nodes = nodes + getHierarchy(c, includeTransforms)
	return nodes

def isRootTransform(xform):
	return mc.listRelatives(xform, allParents = True) == None

def sceneStartTime():
	return mc.playbackOptions(query = True, minTime = True)
	
def sceneEndTime():
	return mc.playbackOptions(query = True, maxTime = True)

def applyHandleSpline(spline, transform):
	"""Writes handle spline data to the transform node's translation channels.  Clean up
	any pre-existing anim before calling"""
	# write keyframes
	for i in range(spline.numberOfHandles()):
		mc.setKeyframe((transform + '.translateX'), t = spline.knots[i], v = spline.handles[i].value.x)
		mc.setKeyframe((transform + '.translateY'), t = spline.knots[i], v = spline.handles[i].value.y)
		mc.setKeyframe((transform + '.translateZ'), t = spline.knots[i], v = spline.handles[i].value.z)

	# set animCurves to weighted
	mc.keyTangent((transform + '.translateX'), e = True, wt = True)
	mc.keyTangent((transform + '.translateY'), e = True, wt = True)
	mc.keyTangent((transform + '.translateZ'), e = True, wt = True)
	
	for i in range(spline.numberOfHandles()):
		# compute maya tangents
		outVx = 0.0
		outTx = 0.5
		inVx = 0.0
		inTx = 0.5 
		outVy = 0.0
		outTy = 0.5
		inVy = 0.0
		inTy = 0.5 
		outVz = 0.0
		outTz = 0.5
		inVz = 0.0
		inTz = 0.5 
	
		if i != (spline.numberOfHandles() - 1):
			outVx = 3.0 * (spline.handles[i].outTangentPoint.x - spline.handles[i].value.x)
			outVy = 3.0 * (spline.handles[i].outTangentPoint.y - spline.handles[i].value.y)
			outVz = 3.0 * (spline.handles[i].outTangentPoint.z - spline.handles[i].value.z)
	
		if i != 0:
			inVx = 3.0 * (spline.handles[i].value.x - spline.handles[i].inTangentPoint.x)
			inVy = 3.0 * (spline.handles[i].value.y - spline.handles[i].inTangentPoint.y)
			inVz = 3.0 * (spline.handles[i].value.z - spline.handles[i].inTangentPoint.z)
	
		# write the tangents
		mc.keyTangent((transform + '.translateX'), e = True, a = True, lock = False,
					   t = (spline.knots[i], spline.knots[i]),
					   ix = inTx, iy = inVx, ox = outTx, oy = outVx)
		mc.keyTangent((transform + '.translateY'), e = True, a = True, lock = False,
					   t = (spline.knots[i], spline.knots[i]), 
					   ix = inTy, iy = inVy, ox = outTy, oy = outVy)
		mc.keyTangent((transform + '.translateZ'), e = True, a = True, lock = False,
					   t = (spline.knots[i], spline.knots[i]),
					   ix = inTz, iy = inVz, ox = outTz, oy = outVz)

def warning(msg):
	mm.eval('warning(\"' + msg + '\")')
	
def lockHierarchy(node, lock = True):
	kids = Node.Node(node).children()
	for k in kids:
		lockHierarchy(k.name, lock)
	mc.lockNode(node, lock = lock)
	
def setSceneTimeToAnimBounds():
	sel = selectedTransforms();
	if len(sel) < 1:
		raise BadSelection()
	sel = sel[0]
	connectedAnimCurves = mc.listConnections(sel, source = True, type = 'animCurve')
	if len(connectedAnimCurves) < 1:
		return
	st = None
	et = None
	for ac in connectedAnimCurves:
		numKeys = len(mc.listAttr(ac + '.ktv', multi = True)) / 3
		thisSt, thisEt = mc.getAttr(ac + '.ktv[0].kt'), mc.getAttr(ac + '.ktv[' + str(numKeys - 1) + '].kt')
		if st is None:
			st = thisSt
		elif thisSt < st:
			st = thisSt
		if et is None:
			et = thisEt
		elif thisEt < et:
			et = thisEt
	if st is not None and et is not None:
		mc.playbackOptions(minTime = st, maxTime = et)
		
def rotateOrderString(i):
	if i == 0:
		return 'xyz'
	elif i == 1:
		return 'yzx'
	elif i == 2:
		return 'zxy'
	elif i == 3:
		return 'xzy'
	elif i == 4:
		return 'yxz'
	else:
		return 'zyx'

def getMObject(node):
	"""returns an MObject for the given node"""
	selList = om.MSelectionList()
	om.MGlobal.getSelectionListByName(node, selList)
	
	mobj = om.MObject()
	selList.getDependNode(0, mobj)
	return mobj
	
	#dagPath = om.MDagPath()
	#selList.getDagPath(0, dagPath)
	#return dagPath.node()

def getDynamicAttributes(node):
	ua = mc.listAttr(node, ud = True)
	if ua is None:
		ua = []
	return ua

def stripDynamicAttributes(node, recursive = True):
	"""strips user defined attrs from given node"""
	attrs = getDynamicAttributes(node)
	
	for a in attrs:
		mc.deleteAttr(node + '.' + a)
		
	if recursive:
		kids = Node.Node(node).children()
		for k in kids:
			stripDynamicAttributes(k.name, recursive)
	
def stripSelection(recursive = True):
	"""strips user defined attrs from selection"""
	sel = mc.ls(selection = True)
	for s in sel:
		stripDynamicAttributes(s, recursive)
	
def getParent(node):
	"""
		given a node string, returns the parent in the dag, or None if top level
	"""
	par = mc.listRelatives(node, parent = True)
	if par is None:
		return par
	else:
		return par[0]
	
def getChildren(node):
	"""
		given a node string, returns a list of the children in the dag
	"""
	kids = mc.listRelatives(node, children = True)
	if kids is None:
		return []
	else:
		return kids

def getChildJoints(node):
	"""
		given a node string, returns a list of the joint children in the dag
	"""
	kids = mc.listRelatives(node, children = True)
	if kids is None:
		return []
	else:
		res = []
		for k in kids:
			if mc.nodeType(k) == 'joint':
				res.append(k)
		return res
		
def getQRotAttr(attr, t = None):
	"""
		returns a QRotation for the given rotation attribute
	"""
	r = None
	if t is None:
		r = mc.getAttr(attr)[0]
	else:
		r = mc.getAttr(attr, t = t)[0]
		
	order = rotateOrderString(mc.getAttr(attr.split('.')[0] + '.rotateOrder'))
	eRot = geom.Rotations.EulerRotation(r[0], r[1], r[2], order = order, local = False, degrees = True)
	return eRot.qRotation()

def setQRotAttr(attr, qr):
	order = rotateOrderString(mc.getAttr(attr.split('.')[0] + '.rotateOrder'))
	er = qr.eulerRotation(local = False, order = order)
	mc.setAttr(attr + 'X', math.degrees(er.rx))
	mc.setAttr(attr + 'Y', math.degrees(er.ry))
	mc.setAttr(attr + 'Z', math.degrees(er.rz))

def setQRotKeyframe(attr, t, qr):
	order = rotateOrderString(mc.getAttr(attr.split('.')[0] + '.rotateOrder'))
	er = qr.eulerRotation(local = False, order = order)
	mc.setKeyframe(attr + 'X', t = t, v = math.degrees(er.rx))
	mc.setKeyframe(attr + 'Y', t = t, v = math.degrees(er.ry))
	mc.setKeyframe(attr + 'Z', t = t, v = math.degrees(er.rz))

def getSkeletonNodes(root):
	"""	
		get a list of node names in the skeleton such that parents come before children
	"""
	
	nodes = [root]
	
	kids = mc.listRelatives(root, children = True)
	if kids is None:
		return nodes
	
	for k in kids:
		if mc.nodeType(k) == 'joint':
			nodes += getSkeletonNodes(k)
	
	return nodes

def getSkeletonDict(root, parent = None):
	"""	
		returns a dict whose keys are nodes in the skeleton at root and whose values are a 2-item tuple
		the first item is the parent (None if root)
		the second item is a list of kids
	"""
	
	skelDict = {}
	
	# add root
	kids = mc.listRelatives(root, children = True)
	if kids is None:
		kids = []
	skelDict[root] = (parent, kids)
	
	# add kids recursively
	for k in kids:
		if mc.nodeType(k) == 'joint':
			skelDict.update(getSkeletonDict(k, root))
	
	return skelDict

def getSkeletalPose(skeletalNodes, t = None):
	"""
		returns the skeletal pose as a dict of attr:val
		skeletalNodes is a list of nodes in the skeleton
		translations and orientations are included
		orientations are included as QRotations
		translations as vectors
	"""
	pose = {}
	for n in skeletalNodes:
		tAttr = n + '.translate'
		tr = None
		if t is None:
			tr = mc.getAttr(tAttr)[0]
		else:
			tr = mc.getAttr(tAttr, t = t)[0]
		pose[tAttr] = mmath.Vector3.Vector3(tr[0], tr[1], tr[2])
		
		rAttr = n + '.rotate'
		pose[rAttr] = getQRotAttr(rAttr, t)
	return pose

def getCartesianPose(skeletalNodes, t = None):
	"""
		returns the cartesian pose as a dict of attr:val
		skeletalNodes is a list of nodes in the skeleton
		translations are included in world space as Points
	"""
	pose = {}
	for n in skeletalNodes:
		tAttr = n + '.translate'
		wmAttr = n + '.worldMatrix[0]'
		tr = None
		if t is None:
			tr = mc.getAttr(wmAttr)[12:15]
		else:
			tr = mc.getAttr(wmAttr, t = t)[12:15]
		pose[tAttr] = geom.Point3.Point3(tr[0], tr[1], tr[2])

	return pose
	
def setSkeletalPose(skeletalPose):
	"""
		sets the state of skeletalPose to the scene attributes it contains
		only supports QRotations
	"""
	for attr in skeletalPose.keys():
		n, a = attr.split('.')
		if a == 'translate':
			mc.setAttr(attr, skeletalPose[attr].x, skeletalPose[attr].y, skeletalPose[attr].z)
		elif a == 'rotate':
			order = rotateOrderString(mc.getAttr(n + '.rotateOrder'))
			er = skeletalPose[attr].eulerRotation(order, local = False)
			mc.setAttr(attr, math.degrees(er.rx), math.degrees(er.ry), math.degrees(er.rz))

def drawCartesianPose(pose, skelDict, name = 'pose#'):
	crvs = []
	for attr in pose.keys():
		n, a = attr.split('.')
		par = skelDict[n][0]
		if par is not None:
			parAttr = par + '.translate'
			pts = [pose[parAttr], pose[attr]]
			crvs.append(mmaya.makeCurve(pts, n + '_' + par))
	return mc.group(crvs, name = name)	

def getUniqueName(desiredName):
	"""
		returns a globally unique name based on desired name
	"""
	if mc.objExists(desiredName):
		i = 1
		name = desiredName + str(i)
		while mc.objExists(name):
			i += 1
			name = desiredName + str(i)
		return name
	else:
		return desiredName

