# copyright 2008 patrick coleman

import math
import random

import numpy

import mmath.Matrix

class EigenPair(object):
	"""For sorting eigenthingees"""
	def __init__(self, eigenvalue, eigenvector):
		self.eigenvalue = eigenvalue
		self.eigenvector = eigenvector
	
	def __cmp__(self, other):
		return cmp(math.fabs(self.eigenvalue), math.fabs(other.eigenvalue))

	def __repr__(self):
		return 'models.pca.EigenPair(' + repr(self.eigenvalue) + ',' + repr(self.eigenvector) + ')'

class PCAModel(object):
	"""columns are data points.  rows index into dmensional components. 
	machine learning conventions in this module """

	def __init__(self, data):
		"""data is an d x n matrix.  index 1 is dimension comonent, index 2 is the data pt index"""
		self.data = data
		self._basis = None
		self._eigenvalues = None
	
	def mean(self):
		"""returns the d dimensional mean as a vector"""
		return mmath.Vector.Vector(self.data.mean(1))
	
	def basis(self, size = None):
		""" returns the largest size components of the basis as a d x size matrix, all if None"""
		if self._basis is None:
			self._computeModel()
		if size is None:
			return self._basis
		else:
			return mmath.Matrix.Matrix(self._basis.data[:,0:size])
	
	def eigenvalues(self, size = None):
		if self._eigenvalues is None:
			self._computeModel()
		if size is None:
			return self._eigenvalues
		else:
			return self._eigenvalues[0:size]
		
	def basisValue(self, i, size = None):
		"""returns the data entry i projected onto the pca basis"""
		if self._basis is None:
			self._computeModel()
		return self.basis(size).transpose() * (self.data.column(i) - self.mean())
		
	def projectValue(self, value, size = None):
		"""projects a new data point onto the basis"""
		if self._basis is None:
			self._computeModel()
		
		return self.basis(size).transpose() * (value - self.mean())
	
	def evaluate(self, i, size = None):
		return self.basis(size) * self.basisValue(i, size) + self.mean()
		
	def evaluateValue(self, value, size = None):
		return self.basis(size) * self.projectValue(value, size) + self.mean()
		
	def _expandDimension(self, value, d):
		if len(value) != d:
			expanded = numpy.zeros(d)
			for i in range(len(value)):
				expanded[i] = value[i]
			value = expanded
		return value
	
	def _computeModel(self):
		# subtract mean
		mn = self.mean()
		mnMat = [mn for i in range(self.data.columnSize())]
		mnMat = mmath.Matrix.Matrix(mnMat).transpose()
		centeredData = self.data - mnMat
		
		# compute covariance matrix
		cov = numpy.zeros([self.data.rowSize(), self.data.rowSize()])
		for i in range(self.data.columnSize()):
			cov += numpy.multiply.outer(centeredData.data[:,i], centeredData.data[:,i])
		cov /= self.data.columnSize()
		
		# compute and sort evals and evecs
		eVals, eVecs = numpy.linalg.eig(cov)
		ePairs = [EigenPair(eVals[i], eVecs[:,i]) for i in range(len(eVals))]
		ePairs.sort()
		ePairs.reverse()
		
		# build basis matrix and store
		basis = [ep.eigenvector for ep in ePairs]
		basis = numpy.array(basis).transpose()
		self._eigenvalues = [ep.eigenvalue for ep in ePairs]
		self._basis = mmath.Matrix.Matrix(basis)
		
def test():
	numPts = 15
	rbounds = (1, 2, 4)
	random.seed(2319)
	
	data = []
	for i in range(numPts):
		data.append([random.uniform(-rbounds[0], rbounds[0]),
					 random.uniform(-rbounds[1], rbounds[1]),
					 random.uniform(-rbounds[2], rbounds[2])])
	dataMat = mmath.Matrix.Matrix(data).transpose()
	
	pcaModel = PCAModel(dataMat)
	
	xi = pcaModel.basisValue(0)
	print dataMat.column(0)
	print pcaModel.basis() * xi + pcaModel.mean()
	print pcaModel.evaluate(0, size = None)
	
if __name__ == '__main__':
	test()