#  Copyright (c) 2008 patrick coleman

from Exceptions import *
import Events
import EventSignals
import StreamlineFilters
import msignal.NumericFilters
import stats.StatisticEvaluator

class EventExtractor(object):
	def __init__(self, motion, sampleRate = 1):
		self.motion = motion
		self.sampleRate = 1
	
	def eventSignal(self):
		raise NotImplemented()
	
class RotationStreamlineExtractor(EventExtractor):

	def __init__(self, 
				 motion,
				 streamlineDict,
				 weightDict = {},
				 streamlineSmoothingIterations = 50,
				 streamlineSmoothingBias = 0.3,
				 kSmoothingIterations = 0,
				 kSmoothingBias = 0.5,
				 kCullingPercentage = 0.95,
				 kUseKink = True,
				 visualizer = None):
		EventExtractor.__init__(self, motion)
		self.streamlineDict = streamlineDict
		self.weightDict = weightDict
		self.streamlineSmoothingIterations = streamlineSmoothingIterations
		self.streamlineSmoothingBias = streamlineSmoothingBias
		self.kSmoothingIterations = kSmoothingIterations
		self.kSmoothingBias = kSmoothingBias
		self.kCullingPercentage = kCullingPercentage
		self.kUseKink = kUseKink
		self.visualizer = visualizer
		
	def eventSignal(self):
		es = EventSignals.EventSignal('rotationEvents',
									  self.motion.startTime, 
									  self.motion.endTime,
									  self.motion.attributes)
		
		visZ = 0
		for attr in self.motion.attributes:
			nodeName, attrName = attr.split('.')
			if attrName == 'rotate':
				attrSls = self.streamlineDict[attr]
				
				# bail on this node if no streamlines
				if len(attrSls) == 0:
					continue
				
				# wt
				wt = 1.0
				if self.weightDict.has_key(attr):
					wt = self.weightDict[attr]
				
				# smooth sls
				smoothedSls = []
				for sl in attrSls:
					sf = StreamlineFilters.SmoothingFilter(sl,
														   iterations = self.streamlineSmoothingIterations,
														   bias = self.streamlineSmoothingBias)
					smoothedSls.append(sf.output())
				
				# get aggregate curvature signal
				kSignal = []
				for i in range(len(smoothedSls[0])):
					kSignal.append(0.0)
				for sl in smoothedSls:
					for i in range(len(sl)):
						if self.kUseKink:
							kSignal[i] += sl.kinkNormal(i).length()
						else:
							kSignal[i] += sl.curvatureNormal(i).length()
				scale = 1.0 / len(smoothedSls)
				for i in range(len(kSignal)):
					kSignal[i] *= scale
				
				# smooth the aggregate curvature signal
				ksf = msignal.NumericFilters.SmoothingFilter(kSignal,
															 self.kSmoothingIterations,
															 self.kSmoothingBias)
				kSignal = ksf.output()
				
				# find all maxima (except bounds, this could be changed to include them )
				allMaximaIndices = []
				if len(kSignal) > 1:
					if kSignal[0] > kSignal[1]:
						allMaximaIndices.append(0)
				for i in range(1, len(kSignal) - 1):
					if (kSignal[i] > kSignal[i - 1]) and (kSignal[i] > kSignal[i + 1]):
						allMaximaIndices.append(i)
				if len(kSignal) > 1:
					if kSignal[len(kSignal) - 2] > kSignal[len(kSignal) - 1]:
						allMaximaIndices.append(len(kSignal) - 1)
				
				# event check only if maxima remain
				if len(allMaximaIndices) > 0:
				
					# cull the maxima based on the culling percentage
					cullK = (1.0 - self.kCullingPercentage) * max([kSignal[i] for i in allMaximaIndices])
					maximaIndices = []
					for i in allMaximaIndices:
						if kSignal[i] > cullK:
							maximaIndices.append(i)
				
					# add events for the ramaining maxima
					for i in maximaIndices:
						es.addEvent(attr,
								   Events.StreamlineExtremaEvent(smoothedSls[0].times[i],
								   wt * kSignal[i]))
			
				# algorithmic visulization
				if self.visualizer != None:
					for sl in attrSls:
						pass #self.visualizer.createStreamlineCurve(sl, nodeName)
					for sl in smoothedSls:
						self.visualizer.createStreamlineCurve(sl, nodeName)
					se = stats.StatisticEvaluator.StatisticEvaluator(kSignal)
					self.visualizer.plotSignal(kSignal, name = nodeName + 'Ksignal', z = visZ)
					for i in range(11):
						p = 0.1 * i
						pval = se.percentileValue(p)
						line = [pval for i in kSignal]
						self.visualizer.plotSignal(line, name = nodeName + 'kPercent', z = visZ)
					
					#self.visualizer.plotHistogram(kSignal, name = nodeName + 'kHist', z = 5)
					visZ += 1 
			# end if attr = rotate
		# end for each attr
		
		return es		
		

		