#  Copyright (c) 2008 patrick coleman

"Motion classes.  This is a Maya free module"

from Exceptions import *
from motion.Poses import Pose, TimedPose
import motion.Poses
import math

from mmath.Vector3 import Vector3
from geom.Rotations import EulerRotation
import geom.Rotations

class BadInterpolationType(Exception): pass
class NoData(Exception): pass

class Motion(object):
	_subtractionSampleRate = 1
	
	def __init__(self, name, startTime, endTime, attributes = []):
		self.name = name
		self.startTime = startTime
		self.endTime = endTime
		self.attributes = attributes
	
	def attributeValue(self, attr, time):
		raise AbstractMethod()
	
	def pose(self, time):
		values = [self.attributeValue(attr, time) for attr in self.attributes]
		return Pose('p' + str(time), self.attributes, values)
	
	def attribute(self, name):
		return self.attributes[self.attributeIndex(name)]
	
	def attributeIndex(self, name):
		for i in range(len(self.attributes)):
			if self.attributes[i] == name:
				return i
		raise NonExistentAttribute()
	
	def appendAttribute(self, name):
		if name in self.attributes:
			raise PreExistingAttribute()
		self.attributes.append(name)
	
	def __sub__(self, motion2):
		difference = SampledMotion('difference', 
								   self.startTime, 
								   self.endTime, 
								   self.attributes,
								   self._subtractionSampleRate)	
		
		difference.clearSamples()
		
		sampleTimes = [difference.sampleTime(i) for i in range(difference.numSamples())]
		for t in sampleTimes:
			for attr in difference.attributes:
				diff = self.attributeValue(attr, t) - motion2.attributeValue(attr, t)
				difference.appendSample(attr, diff)
	
		return difference
		
class SampledMotion(Motion):
	_initialized = False

	def __init__(self, name, startTime, endTime, attributes = [], sampleRate = 1,
		samples = None, initialized = None):
		"latter two params are for specifying fully specified motions"
		
		Motion.__init__(self, name, startTime, endTime, attributes)
		
		if samples is None:
			self._samples = dict([[attr, []] for attr in attributes])
		else:
			self._samples = samples
		self.sampleRate = sampleRate
		if initialized is None:
			self._initialized = True
		else:
			self._initialized = initialized

	def __setattr__(self, attr, value):
		if attr in ['startTime', 'endTime', 'sampleRate'] and self._initialized:
			raise NotImplemented() # needs to change samples
		else:
			self.__dict__[attr] = value
	
	def appendAttribute(self, name):
		raise NotImplemented() # needs to initialize samples for that attribute
	
	def numSamples(self):
		return int((self.endTime - self.startTime) / self.sampleRate + 1)
	
	def sampleTime(self, index):
		# zero based indexing
		return self.startTime + index * self.sampleRate
		
	def clearSamples(self):
		for k in self._samples.keys():
			self._samples[k] = []
			
	def appendSample(self, attr, value):
		self._samples[attr].append(value)
	
	def attributeValue(self, attr, time):
		pi = self.priorIndex(time)
		
		if pi == -1:
			return self._samples[attr][0]
		elif pi == (self.numSamples() - 1):
			return self._samples[attr][pi]
		else:
			pv = self._samples[attr][pi]
			nv = self._samples[attr][pi + 1]
			pt = self.sampleTime(pi)
			nt = self.sampleTime(pi + 1)
			t = (time - pt) / (nt - pt)
			if type(pv) in [int, float]:
				return pv + t * (nv - pv)
			else:
				if attr.split('.')[1] == 'rotate':
					rType = type(pv)
					if rType is Vector3:
						pvEr = EulerRotation(pv.x, pv.y, pv.z, local = False, order = 'xyz', degrees = True)
						nvEr = EulerRotation(nv.x, nv.y, nv.z, local = False, order = 'xyz', degrees = True)
						pvQr = pvEr.qRotation()
						nvQr = nvEr.qRotation()
						vQr = pvQr.interpolate(t, nvQr)
						vEr = vQr.eulerRotation(local = False, order = 'xyz')
						return Vector3(math.degrees(vEr.rx), math.degrees(vEr.ry), math.degrees(vEr.rz))
					elif rType is geom.Rotations.QRotation:
						return pv.interpolate(t, nv)
					else:
						raise BadInterpolationType()
				else:
					return pv.interpolate(t, nv)
		
	def priorIndex(self, time):
		endCacheTime = self.sampleTime(self.numSamples() - 1)
		
		if time < self.startTime:
			return -1
		elif time >= endCacheTime:
			return self.numSamples() - 1
		else:
			return int((time - self.startTime) / self.sampleRate)
	
	def __repr__(self):
		return ('motion.Motions.SampledMotion(' +
			repr(self.name) + ', ' + 
			repr(self.startTime) + ', ' + 
			repr(self.endTime) + ', ' + 
			repr(self.attributes) + ', ' + 
			repr(self.sampleRate) + ', ' + 
			repr(self._samples) + ', ' + 
			repr(self._initialized) + ', ' + 
			')')

class GeneralSampledMotion(Motion):
	"""supports irregular sample rates. does not support sample insertion! (only at end)
		client responsible for maintaining startTime, endTime in sync with times
		a pretty dumb class but it's not the time for designing this right"""

	def __init__(self, name, startTime, endTime, attributes = [], times = None,
		samples = None):
		"latter two params are for specifying fully specified motions"
		
		Motion.__init__(self, name, startTime, endTime, attributes)
		
		if samples is None:
			self.samples = dict([[attr, []] for attr in attributes])
		else:
			self.samples = samples
		
		if times is None:
			self.times = dict([[attr, []] for attr in attributes])
		else:
			self.times = times
		
	def appendAttribute(self, name):
		raise NotImplemented() # needs to initialize samples for that attribute
	
	def numSamples(self):
		ks = self.times.keys()
		if len(ks) == 0:
			raise NoData()
		return len(self.times[ks[0]])
	
	def sampleTime(self, attr, index):
		return self.times[attr][index]
		
	def clearSamples(self):
		for k in self.samples.keys():
			self.samples[k] = []
		for k in self.times.keys():
			self.times[k] = []
			
	
	def appendSample(self, attr, time, value):
		"""multiple calls to this should be in sync with an appendTime call"""
		self.samples[attr].append(value)
		self.times[attr].append(time)
	
	def attributeValue(self, attr, time):
		pi = self.priorIndex(attr, time)
		
		if pi == -1:
			return self.samples[attr][0]
		elif pi == (self.numSamples() - 1):
			return self.samples[attr][pi]
		else:
			pv = self.samples[attr][pi]
			nv = self.samples[attr][pi + 1]
			pt = self.sampleTime(attr, pi)
			nt = self.sampleTime(attr, pi + 1)
			t = (time - pt) / (nt - pt)
			if type(pv) in [int, float]:
				return pv + t * (nv - pv)
			else:
				if attr.split('.')[1] == 'rotate':
					rType = type(pv)
					if rType is Vector3:
						pvEr = EulerRotation(pv.x, pv.y, pv.z, local = False, order = 'xyz', degrees = True)
						nvEr = EulerRotation(nv.x, nv.y, nv.z, local = False, order = 'xyz', degrees = True)
						pvQr = pvEr.qRotation()
						nvQr = nvEr.qRotation()
						vQr = pvQr.interpolate(t, nvQr)
						vEr = vQr.eulerRotation(local = False, order = 'xyz')
						return Vector3(math.degrees(vEr.rx), math.degrees(vEr.ry), math.degrees(vEr.rz))
					elif rType is geom.Rotations.QRotation:
						return pv.interpolate(t, nv)
					else:
						raise BadInterpolationType()
				else:
					return pv.interpolate(t, nv)
		
	def priorIndex(self, attr, time):
		if len(self.times[attr]) < 1:
			raise InsufficientMotionData()
		i = 0
		while i < len(self.times[attr]) and self.times[attr][i] < time:
			i+=1
		return (i - 1)
		
	def __repr__(self):
		return ('motion.Motions.GeneralSampledMotion(' +
			repr(self.name) + ', ' + 
			repr(self.startTime) + ', ' + 
			repr(self.endTime) + ', ' + 
			repr(self.attributes) + ', ' + 
			repr(self.times) + ', ' + 
			repr(self.samples) + ', ' + 
			')')

	
class PoseMotion(Motion):
	
	def __init__(self, name, startTime, endTime, attributes = []):
		Motion.__init__(self, name, startTime, endTime, attributes)
		self.poses = []
	
	def addPose(self, time, pose):
		self.poses.append(TimedPose(pose.name, time, pose.attributes, pose.values))
		self.poses.sort()
	
	def attributeValue(self, attr, time):
		pi = self.priorIndex(time)
		
		if pi == -1:
			return self.poses[0].attributeValue(attr)
		elif pi == (len(self.poses) - 1):
			return self.poses[len(self.poses) - 1].attributeValue(attr)
		else:
			pv = self.poses[pi].attributeValue(attr)
			nv = self.poses[pi + 1].attributeValue(attr)
			pt = self.poses[pi].time
			nt = self.poses[pi + 1].time
			t = (float(time) - pt) / (nt - pt)
			if type(pv) in [int, float]:
				return pv + t * (nv - pv)
			else:
				if attr.split('.')[1] == 'rotate':
					pvEr = EulerRotation(pv.x, pv.y, pv.z, local = False, order = 'xyz', degrees = True)
					nvEr = EulerRotation(nv.x, nv.y, nv.z, local = False, order = 'xyz', degrees = True)
					pvQr = pvEr.qRotation()
					nvQr = nvEr.qRotation()
					vQr = pvQr.interpolate(t, nvQr)
					vEr = vQr.eulerRotation(local = False, order = 'xyz')
					return Vector3(math.degrees(vEr.rx), math.degrees(vEr.ry), math.degrees(vEr.rz))
				else:
					return pv.interpolate(t, nv)
	
	def priorIndex(self, time):
		if len(self.poses) < 2:
			raise InsufficientMotionData()
		i = 0
		while i < len(self.poses) and self.poses[i].time < time:
			i+=1
		return (i - 1)
		
class RefinedPoseMotion(PoseMotion):
	
	def addPose(self, time, pose):
		"""Expects a RefinedPose"""
		self.poses.append(motion.Poses.TimedRefinedPose(pose.name,
										   time, 
										   pose.attributes, 
										   pose.values, 
										   pose.refinements, 
										   pose.attenuation))
		self.poses.sort()
	
	def attributeValue(self, attr, time):
		pi = self.priorIndex(attr, time)
		ni = self.subsequentIndex(attr, time)
		
		if pi == -1:
			return self.poses[ni].attributeValue(attr)
		elif ni == len(self.poses):
			return self.poses[pi].attributeValue(attr)
		else:
			pv = self.poses[pi].attributeValue(attr)
			nv = self.poses[ni].attributeValue(attr)
			pt = self.poses[pi].refinedTime(attr)
			nt = self.poses[ni].refinedTime(attr)
			
			# catch for refinements aligning:
			if nt == pt:
				return nv
			
			t = (float(time) - pt) / (nt - pt)
			if type(pv) in [int, float]:
				return pv + t * (nv - pv)
			else:
				if attr.split('.')[1] == 'rotate':
					pvEr = EulerRotation(pv.x, pv.y, pv.z, local = False, order = 'xyz', degrees = True)
					nvEr = EulerRotation(nv.x, nv.y, nv.z, local = False, order = 'xyz', degrees = True)
					pvQr = pvEr.qRotation()
					nvQr = nvEr.qRotation()
					vQr = pvQr.interpolate(t, nvQr)
					vEr = vQr.eulerRotation(local = False, order = 'xyz')
					return Vector3(math.degrees(vEr.rx), math.degrees(vEr.ry), math.degrees(vEr.rz))
				else:
					return pv.interpolate(t, nv)
	
	def priorIndex(self, attr, time):
		"""Overridden to use attr refinements
		Returns index of last pose st p.time <= time"""
		if len(self.poses) < 2:
			raise InsufficientMotionData()
			
		largestOkTime = None
		bestIndex = None
		bestPoseTime = None # for overlapping refined time resolution
		
		for i in range(len(self.poses)):
			attrTime = self.poses[i].refinedTime(attr)
			if attrTime <= time:
				if largestOkTime == None:
					largestOkTime = attrTime
					bestIndex = i
					bestPoseTime = self.poses[i].time
				elif attrTime > largestOkTime:
					largestOkTime = attrTime
					bestIndex = i
					bestPoseTime = self.poses[i].time
				elif attrTime == largestOkTime:
					#overlapping refinement, resolve by who has larger pose.time
					#if they're equal, there isn't really a good answer.
					if self.poses[i].time > bestPoseTime:
						bestIndex = i
						bestPoseTime = self.poses[i].time
			
		if bestIndex == None:
			bestIndex = -1
		return bestIndex
	
	def subsequentIndex(self, attr, time):
		"""Returns index of first pose st p.time > time"""
		
		if (len(self.poses) < 2):
			raise InsufficientMotionData()
		
		smallestOkTime = None
		bestIndex = None
		
		bestPoseTime = None # for overlapping refined time resolution
		
		for i in range(len(self.poses)):
			attrTime = self.poses[i].refinedTime(attr)
			if attrTime > time:
				if smallestOkTime == None:
					smallestOkTime = attrTime
					bestIndex = i
					bestPoseTime = self.poses[i].time
				elif attrTime < smallestOkTime:
					smallestOkTime = attrTime
					bestIndex = i
					bestPoseTime = self.poses[i].time
				elif attrTime == smallestOkTime:
					#overlapping refinement, resolve by who has smaller pose.time
					#if they're equal, there isn't really a good answer.
					if self.poses[i].time < bestPoseTime:
						bestIndex = i
						bestPoseTime = self.poses[i].time
		
		if bestIndex == None:
			bestIndex = len(self.poses)
		return bestIndex
		
def _testRefinedPoseMotions():
	attributes = ['a', 'b', 'c']
	p1Values = [10, 110, 210]
	p2Values = [20, 120, 220]
	p3Values = [30, 130, 230]
	p4Values = [40, 140, 240]
	
	rp1 = motion.Poses.RefinedPose('rp1', attributes, p1Values)
	rp2 = motion.Poses.RefinedPose('rp3', attributes, p2Values)
	rp3 = motion.Poses.RefinedPose('rp3', attributes, p3Values)
	rp4 = motion.Poses.RefinedPose('rp4', attributes, p4Values)
	
	rp1.refinements = [1, 7, 2]
	rp2.refinements = [-2, -7, -3]
	rp3.refinements = [5, -2, 3]
	rp4.refinements = [5, -2, -6]
	rp4.attenuation = 0.5

	rpm = RefinedPoseMotion('rpm', 1, 20, attributes)
	rpm.addPose(5, rp1)
	rpm.addPose(10, rp2)
	rpm.addPose(15, rp3)
	rpm.addPose(20, rp4)
	 
	for a in attributes:
		print a, ': '
		for i in range(1, 26):
			print '(', i, ' ', rpm.priorIndex(a, i), ' ', rpm.subsequentIndex(a, i), ' ', rpm.attributeValue(a, i), '), ',
			if i % 5 == 0:
				print ''
		print ''
	 
