#  Copyright (c) 2008 patrick coleman

from Exceptions import *

class Pose(object):
	def __init__(self, name, attributes = [], values = []):
		self.name = name
		self.attributes = attributes
		self.values = values
	
	def attribute(self, name):
		return self.attributes[self.attributeIndex(name)]
	
	def attributeValue(self, name):
		return self.values[self.attributeIndex(name)]
	
	def attributeIndex(self, name):
		for i in range(len(self.attributes)):
			if self.attributes[i] == name:
				return i
		raise NonExistentAttribute()
		
	def appendAttribute(self, name, value):
		if name in self.attributes:
			raise PreExistingAttribute()
		self.attributes.append(name)
		self.values.append(value)

class TimedPose(Pose):
	def __init__(self, name, time, attributes = [], values = []):
		Pose.__init__(self, name, attributes, values)
		self.time = time

	def __cmp__(self, other):
		return cmp(self.time, other.time)

class RefinedPose(Pose):
	# Old version for Siggraph, use staggered poses instead

	def __init__(self, name, attributes = [], values = []):
		Pose.__init__(self, name, attributes, values)
		self.attenuation = 1.0
		self.refinements = []
		for i in range(len(attributes)):
			self.refinements.append(0.0)
		
	def attributeRefinement(self, name):
		return self.refinements[self.attributeIndex(name)] * self.attenuation
	
	def setAttributeRefinement(self, attribute, value):
		self.refinements[self.attributeIndex(attribute)] = value
	
	def appendAttribute(self, name, value, refinement = 0.0):
		Pose.appendAttribute(self, name, value)
		self.refinements.append(refinement)
	
	def maximumRefinement(self):
		return max(self.refinements) * self.attenuation
	
	def minimumRefinement(self):
		return min(self.refinements) * self.attenuation
	
class TimedRefinedPose(RefinedPose):
	# Old version for siggraph, use staggered poses instead

	def __init__(self, name, time, attributes = [], values = [], refinements = [], attenuation = 1.0,
				 weights = [], inTangentValues = {}, outTangentValues = {}):
		RefinedPose.__init__(self, name, attributes, values)
		self.refinements = refinements
		self.attenuation = attenuation
		self.time = time
		
		# accessory data that can be written by outside apps; here to support repr
		self.weights = weights
		self.inTangentValues = inTangentValues
		self.outTangentValues = outTangentValues

	def refinedTime(self, name):
		return self.time + self.attributeRefinement(name)
	
	def maximumRefinedTime(self):
		return self.time + self.maximumRefinement()
		
	def minimumRefinedTime(self):
		return self.time + self.minimumRefinement()
	
	def __cmp__(self, other):
		return cmp(self.time, other.time)
	
	def __repr__(self):
		return ('motion.Poses.TimedRefinedPose(' +
				repr(self.name) + ',' +
				repr(self.time) + ',' +
				repr(self.attributes) + ',' +
				repr(self.values) + ',' + 
				repr(self.refinements) + ',' +
				repr(self.attenuation) + ',' +
				repr(self.weights) + ',' +
				repr(self.inTangentValues) + ',' +
				repr(self.outTangentValues) + ',' +
				')')
	
class StaggeredPose(Pose):
	# Cleanup up version for SCA

	def __init__(self, name, attributes = [], values = []):
		Pose.__init__(self, name, attributes, values)
		self.attenuation = 1.0
		self.refinements = []
		self.editingValues = []
		for i in range(len(attributes)):
			self.refinements.append(0.0)
			self.editingValues.append(None)
		
	def attributeRefinement(self, name):
		return self.refinements[self.attributeIndex(name)] * self.attenuation
	
	def setAttributeRefinement(self, attribute, value):
		# doesn't handle constraints with refinement changes, 
		# since this class knows nothing about transitions.
		i = self.attributeIndex(attribute)
		
		# delete editing knots if setting refinement to zero
		if value == 0.0:
			self.editingValues[i] = None
		
		self.refinements[i] = value
		
	def appendAttribute(self, name, value, refinement = 0.0):
		Pose.appendAttribute(self, name, value)
		self.refinements.append(refinement)
		self.editingValues.append(None) # no editing knot by default
		
	def maximumRefinement(self):
		return max(self.refinements) * self.attenuation
	
	def minimumRefinement(self):
		return min(self.refinements) * self.attenuation		

	def hasEditingValue(self, attribute):
		i = self.attributeIndex(attribute)
		return (self.editingValues[i] != None)
	
	def editingValue(self, name):
		return self.editingValues[self.attributeIndex(name)]


class TimedStaggeredPose(StaggeredPose):
	# Cleaned up version for SCA
	
	def __init__(self, name, time, attributes = [], values = [], refinements = None, 
				 editingValues = None, attenuation = 1.0):
		StaggeredPose.__init__(self, name, attributes, values)
		if refinements is not None:
			self.refinements = refinements
		if editingValues is not None:
			self.editingValues = editingValues
		self.attenuation = attenuation
		self.time = time

	def refinedTime(self, name):
		return self.time + self.attributeRefinement(name)
	
	def maximumRefinedTime(self):
		return self.time + self.maximumRefinement()
		
	def minimumRefinedTime(self):
		return self.time + self.minimumRefinement()
	
	def __cmp__(self, other):
		return cmp(self.time, other.time)
	
	def __repr__(self):
		return ('motion.Poses.TimedStaggeredPose(' +
				repr(self.name) + ',' +
				repr(self.time) + ',' +
				repr(self.attributes) + ',' +
				repr(self.values) + ',' + 
				repr(self.refinements) + ',' +
				repr(self.editingValues) + ',' +
				repr(self.attenuation) + ',' +
				')')
				