# copyrigt 2008 patrick coleman

# note: for evaluation performance, redesign everything to use dictionaries indexed
# by attribute name.  At a low level, these things are making tons of calls to look
# up indices, which is linear in |attrs|.

import Poses
import Motions

class BadArgumentData(Exception): pass
class InsufficientMotionData(Exception): pass

class StaggeredPoseMotion(Motions.Motion):
	"""Cleaned up staggered pose motion
	   Evaluation order: transitions by index, then globalDisps by index, then constriantDisps 
	   by index
	   constraintDisps should be used for reprenting relative ik solutions to meet constraints 
	   
	   If local transitions are not specified, linear interpolation is used to calculate the motion
	"""
	
	def __init__(self, name, startTime, endTime, attributes = [], poses = [],
				 transitions = {}, globalDisplacements = {},
				 constraintDisplacements = {}):
		
		Motions.Motion.__init__(self, name, startTime, endTime, attributes)
		
		 # temporally sorted list of poses
		self.poses = poses
		
		# dict (by attribute) of lists [by temporal index] of spline & disp stacks
		self.transitions = transitions
		
		# dict (by attr) of global displacement stacks
		self.globalDisplacements = globalDisplacements
		self.constraintDisplacements = constraintDisplacements

	def addPose(self, pose):
		"""Adds a TimedStaggeredPose"""
		if type(pose) is not motion.Poses.TimedStaggeredPose:
			raise BadArgumentData()
		self.poses.append(pose)
		self.poses.sort()

	def linearEvaluate(self, attr, time):
		"Does linear evaluation"
	
		pi = self.priorIndex(attr, time)
		ni = self.subsequentIndex(attr, time)
		
		if pi == -1:
			# before first pose; return flat value of min time value
			pn = self.poses[ni]
			if pn.hasEditingValue(attr) and pn.attributeRefinement(attr) > 0.0:
				return pn.editingValue(attr)
			else:
				return pn.attributeValue(attr)
		
		elif ni == len(self.poses):
			# after last pose; return flat value of max time value
			pp = self.poses[pi]
			if pp.hasEditingValue(attr) and pp.attributeRefinement(attr) < 0.0:
				return pp.editingValue(attr)
			else:
				return pp.attributeValue(attr)
		
		else:	
			# interpolate between neighboring knots
			
			if pi == ni:
				# between editing knot and refined knot
				if self.poses[pi].hasEditingValue(attr):
					pv = self.poses[pi].editingValue(attr)
					nv = self.poses[pi].attributeValue(attr)
					pt = self.poses[pi].time
					nt = self.poses[pi].refinedTime(attr)
					if self.poses[pi].attributeRefinement(attr) < 0.0:
						pv, nv = nv, pv
						pt, nt = nt, pt
					if pt == nt:
						return nv
					t = (float(time) - pt) / (nt - pt)
					return self.interpolate(pv, nv, t)
				else: # no editing knot.  If conventions are held, this won't happen
					return self.poses[pi].attributeValue(attr)
			
			else: #between adjacent poses
				
				# get bounding times and values
				pp = self.poses[pi]
				np = self.poses[ni]
				
				pv = None
				pt = None
				if pp.hasEditingValue(attr) and pp.attributeRefinement(attr) < 0.0:
					pv = pp.editingValue(attr)
					pt = pp.time
				else:
					pv = pp.attributeValue(attr)
					pt = pp.refinedTime(attr)
				
				nv = None
				nt = None
				if np.hasEditingValue(attr) and np.attributeRefinement(attr) > 0.0:
					nv = np.editingValue(attr)
					nt = np.time
				else:
					nv = np.attributeValue(attr)
					nt = np.refinedTime(attr)
				
				# interp
				if pt == nt:
					return nv
				t = (float(time) - pt) / (nt - pt)
				return self.interpolate(pv, nv, t)

	def priorIndex(self, attr, time):
		"""
		Overridden to use attr refinements & editing knots
		Returns index of last pose st min(p.time, p.refineedTime(attr)) <= time
		"""
		if len(self.poses) < 2:
			raise InsufficientMotionData()
			
		largestOkTime = None
		bestIndex = None
		bestPoseTime = None # for overlapping refined time resolution
		
		for i in range(len(self.poses)):
			minTime = self.poses[i].refinedTime(attr)
			if self.poses[i].hasEditingValue(attr):
				minTime = min(minTime, self.poses[i].time)
			
			if minTime <= time:
				if largestOkTime == None:
					largestOkTime = minTime
					bestIndex = i
					bestPoseTime = self.poses[i].time
				elif minTime > largestOkTime:
					largestOkTime = minTime
					bestIndex = i
					bestPoseTime = self.poses[i].time
				elif minTime == largestOkTime:
					#overlapping refinement, resolve by who has larger pose.time
					#if they're equal, there isn't really a good answer.
					if self.poses[i].time > bestPoseTime:
						bestIndex = i
						bestPoseTime = self.poses[i].time
			
		if bestIndex == None:
			bestIndex = -1
		return bestIndex
	
	def subsequentIndex(self, attr, time):
		"""
		Returns index of first pose st max(p.time, p.refinedTime(attr)) > time
		"""
		
		if (len(self.poses) < 2):
			raise InsufficientMotionData()
		
		smallestOkTime = None
		bestIndex = None
		
		bestPoseTime = None # for overlapping refined time resolution
		
		for i in range(len(self.poses)):
			maxTime = self.poses[i].refinedTime(attr)
			if self.poses[i].hasEditingValue(attr):
				maxTime = max(maxTime, self.poses[i].time)
			
			if maxTime > time:
				if smallestOkTime == None:
					smallestOkTime = maxTime
					bestIndex = i
					bestPoseTime = self.poses[i].time
				elif maxTime < smallestOkTime:
					smallestOkTime = maxTime
					bestIndex = i
					bestPoseTime = self.poses[i].time
				elif maxTime == smallestOkTime:
					#overlapping refinement, resolve by who has smaller pose.time
					#if they're equal, there isn't really a good answer.
					if self.poses[i].time < bestPoseTime:
						bestIndex = i
						bestPoseTime = self.poses[i].time
		
		if bestIndex == None:
			bestIndex = len(self.poses)
		return bestIndex

	def interpolate(self, a, b, u):
		"""interpolation wrapper to support floats"""
		if type(a) in [int, float]:
			return a + u * (b - a)
		else:
			return a.interpolate(u, b)

def testLinearEvaluate():
	"""
	poses at times 1, 5, 10
	attrs:
		a: no editing knots (v = 1, ev = None, ref = 1), (v = 20, ev = None, ref = 2), 
			(v = 5, ev = None, ref = -1)
		b: editing knots (v = 1, ev = 2, ref = 1), (v = 20, ev = 25, ref = 2), 
			(v = 5, ev = 7, ref = -1)
		c: editing knots on 1st and last (v = 1, ev = 2, ref = 1), (v = 20, ev = None, ref = 2), 
			(v = 5, ev = 7, ref = -1)
	
	
	print evaluation in dt = .25 increments from -5 to 15
	"""
	attrs = ['a', 'b', 'c']
	st = 1
	et = 10
	
	p1 = Poses.TimedStaggeredPose('p1', 1, attrs, [1, 1, 1])
	p1.refinements = [1, 1, 1]
	p1.editingValues = [None, 2, 2]

	p2 = Poses.TimedStaggeredPose('p2', 5, attrs, [20, 20, 20])
	p2.refinements = [2, 2, 2]
	p2.editingValues = [None, 25, None]
	
	p3 = Poses.TimedStaggeredPose('p3', 10, attrs, [5, 5, 5])
	p3.refinements = [-1, -1, -1]
	p3.editingValues = [None, 7, 7]

	poses = [p1, p2, p3]

	m1 = StaggeredPoseMotion('m1', st, et, attrs, poses)
	
	for a in attrs:
		t = -5.0
		dt = 0.25
		print a + ':'
		while t <= 15.0:
			print str(t) + ':\t' + str(m1.linearEvaluate(a, t))
			t += dt

if __name__ == '__main__':
	testLinearEvaluate()
	




"""
we need a transiton class, which will have an evaluate method.  The transition class will
be a curve or a qrcurve defined over [0,1], along with an attenuation attr, [0, inf].  if its 
a disp map, itll be a linear spline.  


we need a displacement map class, which will have irregularly sampled motion, 
like general sampled motion. maybe just use that, or copy it and clean it up.  attenuation control.

staggered pose motion needs evaluate, which evaluates the motion.
linearEvaluate does basis lerping from pose to pose.
transitionEvaluate ignores the global and constraint disp maps

coding breakdown, in 4 parts.
1) linear evaluation (DONE).  test (DONE).
2) transition class & transition evaluate.  test with a maya script.
3) disp map class
4) evaluate method on staggered motion
"""



