#  Copyright (c) 2008 patrick coleman

from geom.Point3 import Point3
from geom.Normal3 import Normal3
from mmath.Vector3 import Vector3
from geom.BoundingBox import BoundingBox
from geom.Rotations import EulerRotation
from Exceptions import *
import math

class Frame(object):
	"coordinate framed efined by three points on axes"
	def __init__(self, 
				 axisX = Point3(1, 0, 0), 
				 axisY = Point3(0, 1, 0),
				 axisZ = Point3(0, 0, 1)):
		self.axisX = axisX
		self.axisY = axisY
		self.axisZ = axisZ

class FrameNormal(object):
	"Instantaneous curve normals of a frame trajectory"
	def __init__(self, 
				 normalX = Normal3(1, 0, 0), 
				 normalY = Normal3(0, 1, 0),
				 normalZ = Normal3(0, 0, 1)):
		self.normalX = normalX
		self.normalY = normalY
		self.normalZ = normalZ

class Streamframe(object):
	"frame trajectory"
	def __init__(self, frames = [], times = []):
		self.frames = frames
		self.times = times
		self.frameNormals = None # Needs more robust computation ported over from Curve.SampledSpline
		self.kinks = None # Needs more robust computation ported over from Curve.SampledSpline
		self.curvatures = None # Needs more robust computation ported over from Curve.SampledSpline
		
		self._bboxes = (None, None, None)
		self._bboxesDirty = True
	
	def addSample(self, time, frame):
		self.times.append(time)
		self.frames.append(frame)
		self._bboxesDirty = True
		
	def clear(self):
		self.times = []
		self.frames = []
		self.normals = []
		self.kinks = []
		self._bboxesDirty = True
	
	def __len__(self):
		return len(self.frames)
	
	def frameNormal(self, i):
		if self.frameNormals == None:
			self._computeFrameNormals()
		return self.frameNormals[i]
	
	def kink(self, i):
		if self.kinks == None:
			self._computeKinks()
		return self.kinks[i]
	
	def curvature(self, i):
		if self.curvatures == None:
			self._computeCurvatures()
		return self.curvatures[i]
	
	def boundingBoxes(self):
		if self._bboxesDirty:
			self._computeBboxes()
		return self._bboxes
	
	def isDegenerate(self, tol = 0.0, numComponents = 3):
		if self._bboxesDirty:
			self._computeBboxes()
		
		if self._bboxes[0].minPoint.distanceSquared(self._bboxes[0].maxPoint) > (tol * tol):
			return False
		if (numComponents > 1 and
			self._bboxes[1].minPoint.distanceSquared(self._bboxes[1].maxPoint) > (tol * tol)):
			return False
		if (numComponents > 2 and
			self._bboxes[2].minPoint.distanceSquared(self._bboxes[2].maxPoint) > (tol * tol)):
			return False
		return True
	
	def lengths(self):
		"Returns the euclidean length of the streamline"
		res = [0.0, 0.0, 0.0]
		for i in range(1, len(self.frames)):
			res[0] += self.frames[i].axisX.distance(self.frames[i - 1].axisX)
			res[1] += self.frames[i].axisY.distance(self.frames[i - 1].axisY)
			res[2] += self.frames[i].axisZ.distance(self.frames[i - 1].axisZ)
		return res
	
	def _computeFrameNormals(self):
		# TODO Needs more robust computation ported over from Curve.SampledSpline
	
		self.frameNormals = []
		
		# zero normal at first point
		if len(self.frames) > 0:
			self.frameNormals.append(
				FrameNormal(Normal3(0.0, 0.0, 0.0), Normal3(0.0, 0.0, 0.0), Normal3(0.0, 0.0, 0.0)))
		
		# general normals as interpolated convex edge normals
		for i in range(1, len(self) - 1):
			nx = None
			ny = None
			nz = None
			
			# edge vectors
			avX = self.frames[i - 1].axisX - self.frames[i].axisX
			bvX = self.frames[i + 1].axisX - self.frames[i].axisX
			avY = self.frames[i - 1].axisY - self.frames[i].axisY
			bvY = self.frames[i + 1].axisY - self.frames[i].axisY
			avZ = self.frames[i - 1].axisZ - self.frames[i].axisZ
			bvZ = self.frames[i + 1].axisZ - self.frames[i].axisZ
			
			# edge lengths
			alX = avX.length()
			blX = bvX.length()
			alY = avY.length()
			blY = bvY.length()
			alZ = avZ.length()
			blZ = bvZ.length()
			
			# normal on axis x
			if alX == 0.0 or blX == 0.0:
				nx = Normal3(0.0, 0.0, 0.0)
			else:
				sideX = bvX ^ avX
				anX = avX ^ sideX
				bnX = sideX ^ bvX
				if anX.length() > 0.0:
					anX = anX.normalize()
				if bnX.length() > 0.0:
					bnX = bnX.normalize()
			
				#interpolate edge normals
				tX = blX / (alX + blX)
				nX = anX.interpolate(tX, bnX)
				if nX.length() > 0.0:
					nX = nX.normalize()
				nx = Normal3(nX.x, nX.y, nX.z)
				
			# normal on axis y
			if alY == 0.0 or blY == 0.0:
				ny = Normal3(0.0, 0.0, 0.0)
			else:
				sideY = bvY ^ avY
				anY = avY ^ sideY
				bnY = sideY ^ bvY
				if anY.length() > 0.0:
					anY = anY.normalize()
				if bnY.length() > 0.0:
					bnY = bnY.normalize()
			
				#interpolate edge normals
				tY = blY / (alY + blY)
				nY = anY.interpolate(tY, bnY)
				if nY.length() > 0.0:
					nY = nY.normalize()
				ny = Normal3(nY.x, nY.y, nY.z)
		
			# normal on axis z
			if alZ == 0.0 or blZ == 0.0:
				nz = Normal3(0.0, 0.0, 0.0)
			else:
				sideZ = bvZ ^ avZ
				anZ = avZ ^ sideZ
				bnZ = sideZ ^ bvZ
				if anZ.length() > 0.0:
					anZ = anZ.normalize()
				if bnZ.length() > 0.0:
					bnZ = bnZ.normalize()
			
				#interpolate edge normals
				tZ = blZ / (alZ + blZ)
				nZ = anZ.interpolate(tZ, bnZ)
				if nZ.length() > 0.0:
					nZ.normalize()
				nz = Normal3(nZ.x, nZ.y, nZ.z)
			
			# append frame
			self.frameNormals.append(FrameNormal(nx, ny, nz))
			
		# zero normal at end
		if len(self.frames) > 1:
			self.frameNormals.append(
				FrameNormal(Normal3(0.0, 0.0, 0.0), Normal3(0.0, 0.0, 0.0), Normal3(0.0, 0.0, 0.0)))
	
	def _computeKinks(self):
		# TODO Needs more robust computation ported over from Curve.SampledSpline
		
		if self.frameNormals == None:
			self._computeFrameNormals()
	
		self.kinks = []
		
		# zero kink at start
		if len(self.frames) > 0:
			self.kinks.append((0.0, 0.0, 0.0))
		
		# general pseduo-curvature based on weighted directional curvature per Taubin 95
		for i in range(1, len(self.frames) - 1):
			kx = None
			ky = None
			kz = None
			
			avX = self.frames[i - 1].axisX - self.frames[i].axisX
			bvX = self.frames[i + 1].axisX - self.frames[i].axisX
			avY = self.frames[i - 1].axisY - self.frames[i].axisY
			bvY = self.frames[i + 1].axisY - self.frames[i].axisY
			avZ = self.frames[i - 1].axisZ - self.frames[i].axisZ
			bvZ = self.frames[i + 1].axisZ - self.frames[i].axisZ
		
			avLenX = avX.length()
			bvLenX = bvX.length()
			avLenY = avY.length()
			bvLenY = bvY.length()
			avLenZ = avZ.length()
			bvLenZ = bvZ.length()
					
			kinkAX = 0.0
			kinkBX = 0.0
			kinkAY = 0.0
			kinkBY = 0.0
			kinkAZ = 0.0
			kinkBZ = 0.0
			
			if avLenX > 0.0:
				kinkAX = math.fabs(self.frameNormals[i].normalX *  avX / avLenX)
			if bvLenX > 0.0:
				kinkBX = math.fabs(self.frameNormals[i].normalX *  bvX / bvLenX)
			if avLenY > 0.0:
				kinkAY = math.fabs(self.frameNormals[i].normalY *  avY / avLenY)
			if bvLenY > 0.0:
				kinkBY = math.fabs(self.frameNormals[i].normalY *  bvY / bvLenY)
			if avLenZ > 0.0:
				kinkAZ = math.fabs(self.frameNormals[i].normalZ *  avZ / avLenZ)
			if bvLenZ > 0.0:
				kinkBZ = math.fabs(self.frameNormals[i].normalZ *  bvZ / bvLenZ)
			
			# interp between kappa on neighbors normals; this gives a meaningful scalar inter, but
			# use tht local normal.  i'm sure there's a better way to do this. 
			kavX = kinkAX * self.frameNormals[i - 1].normalX
			kbvX = kinkBX * self.frameNormals[i + 1].normalX
			kavY = kinkAY * self.frameNormals[i - 1].normalY
			kbvY = kinkBY * self.frameNormals[i + 1].normalY
			kavZ = kinkAZ * self.frameNormals[i - 1].normalZ
			kbvZ = kinkBZ * self.frameNormals[i + 1].normalZ
			
			denomX = avLenX + bvLenX
			denomY = avLenY + bvLenY
			denomZ = avLenZ + bvLenZ
			
			if denomX <= 0.0:
				kx = 0.0
			else: 
				tX = bvLenX / denomX
				kvX = kavX.interpolate(tX, kbvX)
				kx = kvX.length()
			
			if denomY <= 0.0:
				ky = 0.0
			else: 
				tY = bvLenY / denomY
				kvY = kavY.interpolate(tY, kbvY)
				ky = kvY.length()
			
			if denomZ <= 0.0:
				kz = 0.0
			else: 
				tZ = bvLenZ / denomZ
				kvZ = kavZ.interpolate(tZ, kbvZ)
				kz = kvZ.length()
			
			self.kinks.append((kx, ky, kz))
		
		# zero kink at end
		if len(self.frames) > 1:
			self.kinks.append((0.0, 0.0, 0.0))
	
	def _computeCurvatures(self):
		# TODO Needs more robust computation ported over from Curve.SampledSpline
		
		if self.frameNormals == None:
			self._computeFrameNormals()
	
		self.curvatures = []
		
		# zero curvature at start
		if len(self.frames) > 0:
			self.curvatures.append((0.0, 0.0, 0.0))
		
		# general pseduo-curvature based on weighted directional curvature per Taubin 95
		for i in range(1, len(self.frames) - 1):
			kx = None
			ky = None
			kz = None
			
			avX = self.frames[i - 1].axisX - self.frames[i].axisX
			bvX = self.frames[i + 1].axisX - self.frames[i].axisX
			avY = self.frames[i - 1].axisY - self.frames[i].axisY
			bvY = self.frames[i + 1].axisY - self.frames[i].axisY
			avZ = self.frames[i - 1].axisZ - self.frames[i].axisZ
			bvZ = self.frames[i + 1].axisZ - self.frames[i].axisZ
		
			avLenX = avX.length()
			bvLenX = bvX.length()
			avLenY = avY.length()
			bvLenY = bvY.length()
			avLenZ = avZ.length()
			bvLenZ = bvZ.length()
					
			curvatureAX = 0.0
			curvatureBX = 0.0
			curvatureAY = 0.0
			curvatureBY = 0.0
			curvatureAZ = 0.0
			curvatureBZ = 0.0
			
			if avLenX > 0.0:
				curvatureAX = math.fabs(self.frameNormals[i].normalX *  avX / (avLenX * avLenX))
			if bvLenX > 0.0:
				curvatureBX = math.fabs(self.frameNormals[i].normalX *  bvX / (bvLenX * bvLenX))
			if avLenY > 0.0:
				curvatureAY = math.fabs(self.frameNormals[i].normalY *  avY / (avLenY * avLenY))
			if bvLenY > 0.0:
				curvatureBY = math.fabs(self.frameNormals[i].normalY *  bvY / (bvLenY * bvLenY))
			if avLenZ > 0.0:
				curvatureAZ = math.fabs(self.frameNormals[i].normalZ *  avZ / (avLenZ * avLenZ))
			if bvLenZ > 0.0:
				curvatureBZ = math.fabs(self.frameNormals[i].normalZ *  bvZ / (bvLenZ * bvLenZ))
			
			# interp between kappa on neighbors normals; this gives a meaningful scalar inter, but
			# use tht local normal.  i'm sure there's a better way to do this. 
			kavX = curvatureAX * self.frameNormals[i - 1].normalX
			kbvX = curvatureBX * self.frameNormals[i + 1].normalX
			kavY = curvatureAY * self.frameNormals[i - 1].normalY
			kbvY = curvatureBY * self.frameNormals[i + 1].normalY
			kavZ = curvatureAZ * self.frameNormals[i - 1].normalZ
			kbvZ = curvatureBZ * self.frameNormals[i + 1].normalZ
			
			denomX = avLenX + bvLenX
			denomY = avLenY + bvLenY
			denomZ = avLenZ + bvLenZ
			
			if denomX <= 0.0:
				kx = 0.0
			else: 
				tX = bvLenX / denomX
				kvX = kavX.interpolate(tX, kbvX)
				kx = kvX.length()
			
			if denomY <= 0.0:
				ky = 0.0
			else: 
				tY = bvLenY / denomY
				kvY = kavY.interpolate(tY, kbvY)
				ky = kvY.length()
			
			if denomZ <= 0.0:
				kz = 0.0
			else: 
				tZ = bvLenZ / denomZ
				kvZ = kavZ.interpolate(tZ, kbvZ)
				kz = kvZ.length()
			
			self.curvatures.append((kx, ky, kz))
		
		# zero curvature at end
		if len(self.frames) > 1:
			self.curvatures.append((0.0, 0.0, 0.0))
			
	def _computeBboxes(self):
		axMin = Point3()
		ayMin = Point3()
		azMin = Point3()
		axMax = Point3()
		ayMax = Point3()
		azMax = Point3()
		
		for i in range(len(self.frames)):
			if i == 0:
				axMin = Point3(self.frames[i].axisX.x, self.frames[i].axisX.y, self.frames[i].axisX.z)
				ayMin = Point3(self.frames[i].axisY.x, self.frames[i].axisY.y, self.frames[i].axisY.z)
				azMin = Point3(self.frames[i].axisZ.x, self.frames[i].axisZ.y, self.frames[i].axisZ.z)
				axMax = Point3(self.frames[i].axisX.x, self.frames[i].axisX.y, self.frames[i].axisX.z)
				ayMax = Point3(self.frames[i].axisY.x, self.frames[i].axisY.y, self.frames[i].axisY.z)
				azMax = Point3(self.frames[i].axisZ.x, self.frames[i].axisZ.y, self.frames[i].axisZ.z)				
			else:
				# axis x
				if self.frames[i].axisX.x < axMin.x:
					axMin.x = self.frames[i].axisX.x
				if self.frames[i].axisX.y < axMin.y:
					axMin.y = self.frames[i].axisX.y
				if self.frames[i].axisX.z < axMin.z:
					axMin.z = self.frames[i].axisX.z
				if self.frames[i].axisX.x > axMax.x:
					axMax.x = self.frames[i].axisX.x
				if self.frames[i].axisX.y > axMax.y:
					axMax.y = self.frames[i].axisX.y
				if self.frames[i].axisX.z > axMax.z:
					axMax.z = self.frames[i].axisX.z
				
				# axis y
				if self.frames[i].axisY.x < ayMin.x:
					ayMin.x = self.frames[i].axisY.x
				if self.frames[i].axisY.y < ayMin.y:
					ayMin.y = self.frames[i].axisY.y
				if self.frames[i].axisY.z < ayMin.z:
					ayMin.z = self.frames[i].axisY.z
				if self.frames[i].axisY.x > ayMax.x:
					ayMax.x = self.frames[i].axisY.x
				if self.frames[i].axisY.y > ayMax.y:
					ayMax.y = self.frames[i].axisY.y
				if self.frames[i].axisY.z > ayMax.z:
					ayMax.z = self.frames[i].axisY.z
					
				# axis z
				if self.frames[i].axisZ.x < azMin.x:
					azMin.x = self.frames[i].axisZ.x
				if self.frames[i].axisZ.y < azMin.y:
					azMin.y = self.frames[i].axisZ.y
				if self.frames[i].axisZ.z < azMin.z:
					azMin.z = self.frames[i].axisZ.z
				if self.frames[i].axisZ.x > azMax.x:
					azMax.x = self.frames[i].axisZ.x
				if self.frames[i].axisZ.y > azMax.y:
					azMax.y = self.frames[i].axisZ.y
				if self.frames[i].axisZ.z > azMax.z:
					azMax.z = self.frames[i].axisZ.z
				
		self._bboxes = (BoundingBox(axMin, axMax), BoundingBox(ayMin, ayMax), BoundingBox(azMin, azMax))
		self._bboxesDirty = False

class StreamframeCreator(object):
	def __init__(self, startTime, endTime, evaluator, step = 1):
		self.startTime = startTime
		self.endTime = endTime
		self.evaluator = evaluator
		self.step = step
	
	def streamframe(self):
		sf = Streamframe()
		sf.clear()
		t = self.startTime
		while t <= self.endTime:
			sf.addSample(t, self.evaluator.frame(t))
			t += self.step
		return sf

class FrameEvaluator(object):
	def frame(self, time):
		raise AbstractMethod()
	
class RotatingFrameEvaluator(FrameEvaluator):
	def __init__(self, 
				 motion, 
				 attribute, 
				 localFrame = Frame(Point3(1, 0, 0), Point3(0, 1, 0), Point3(0, 0, 1)), # what's rotated
				 normalize = False,
				 radius = 1.0,
				 degrees = False):
		self.motion = motion
		self.attribute = attribute
		self.localFrame = localFrame
		self.normalize = normalize
		self.radius = radius
		self.degrees = degrees
	
	def frame(self, time):
		attrVal = self.motion.attributeValue(self.attribute, time)
		if type(attrVal) != Vector3:
			raise UnsupportedData()
		rotation = EulerRotation(attrVal.x, attrVal.y, attrVal.z, 
								 order = 'xyz', local = False, 
								 degrees = self.degrees)
		fList = [rotation(self.localFrame.axisX), 
				 rotation(self.localFrame.axisY), 
				 rotation(self.localFrame.axisZ)]
		if self.normalize:
			for i in range(3):
				v = Vector3(fList[i].x, fList[i].y, fList[i].z)
				if v.lengthSquared() > 0.0:
					v = v.normalize()
					v = self.radius * v
					fList[i] = Point3(v.x, v.y, v.z)
		return Frame(fList[0], fList[1], fList[2]) 
	
def _testStreamframe():
	sf = Streamframe()
	for i in range(10):
	
		p1 = Point3(i*i, i, i)
		p2 = Point3(i, i*i, i)
		p3 = Point3(i, i, i*i)
		sf.addSample(0.1 * i, Frame(p1, p2, p3))
	for i in range(len(sf)):
		print sf.times[i], ': ', 
		print sf.frames[i].axisX, sf.frames[i].axisY, 
		print sf.frames[i].axisZ
		print '\t', sf.frameNormal(i).normalX, sf.frameNormal(i).normalY, sf.frameNormal(i).normalZ
		print '\t', sf.kink(i)[0], sf.kink(i)[1], sf.kink(i)[2]
	
if __name__ == '__main__':
	
	print 'nada'