#  Copyright (c) 2008 patrick coleman

from Exceptions import *
import Events
import motion.StreamframeFilters

class StreamframeAnalyzer(object):
	def __init__(self, streamframe):
		self.streamframe = streamframe
	
	def poseSignal(self):
		"""Returns a continuous signal, representing a belief about posiness for the streamframe"""
		raise AbstractMethod()
		
	def eventSequence(self):
		"""Returns a list of events representing significant things that happen"""
		raise AbstractMethod()

class KinkAnalyzer(StreamframeAnalyzer):
	def __init__(self, streamframe, smoothingBias = 0.5, smoothingIterations = 50):
		StreamframeAnalyzer.__init__(self, streamframe)
		filter = motion.StreamframeFilters.SmoothingFilter(self.streamframe, 
														   smoothingBias, 
														   smoothingIterations)
		self.smoothedSf = filter.output()
		
	def poseSignal(self):
		"""This class returns the kink measurement of the streamframe, the 3-comp kink among the components is returned"""
		signal = []
		for i in range(len(self.smoothedSf)):
			k = self.smoothedSf.kink(i)
			signal.append(k)
		return signal

	def eventSequence(self, includeBounds = True):
		"""Returns events for the local maxima of the streamframe.  Local maxima are per component"""
		signals = []
		for i in range(len(self.smoothedSf)):
			signals.append(self.smoothedSf.kink(i))
		
		xMaxima = []
		yMaxima = []
		zMaxima = []
		
		# get all maxima
		if len(signals) > 1 and includeBounds:
			if signals[0][0] > signals[1][0]:
				xMaxima.append(0)
			if signals[0][1] > signals[1][1]:
				yMaxima.append(0)
			if signals[0][2] > signals[1][2]:
				zMaxima.append(0)
		for i in range(1, len(signals) - 1):
			if (signals[i][0] > signals[i - 1][0]) and (signals[i][0] > signals[i + 1][0]):
				xMaxima.append(i)
			if (signals[i][1] > signals[i - 1][1]) and (signals[i][1] > signals[i + 1][1]):
				yMaxima.append(i)
			if (signals[i][2] > signals[i - 1][2]) and (signals[i][2] > signals[i + 1][2]):
				zMaxima.append(i)
		if len(signals) > 1 and includeBounds:
			if signals[len(signals) - 2][0] < signals[len(signals) - 1][0]:
				xMaxima.append(len(signals) - 1)
			if signals[len(signals) - 2][1] < signals[len(signals) - 1][1]:
				yMaxima.append(len(signals) - 1)
			if signals[len(signals) - 2][2] < signals[len(signals) - 1][2]:
				zMaxima.append(len(signals) - 1)
		
		events = []
		for i in xMaxima:
			events.append(Events.StreamframeExtremaEvent(i, self.smoothedSf.times[i],
														 signals[i][0], 0))
		for i in yMaxima:
			events.append(Events.StreamframeExtremaEvent(i, self.smoothedSf.times[i],
														 signals[i][1], 1))
		for i in zMaxima:
			events.append(Events.StreamframeExtremaEvent(i, self.smoothedSf.times[i],
														 signals[i][2], 2))
		events.sort()
		
		return events

class CurvatureAnalyzer(StreamframeAnalyzer):
	def __init__(self, streamframe, smoothingBias = 0.5, smoothingIterations = 50):
		StreamframeAnalyzer.__init__(self, streamframe)
		filter = motion.StreamframeFilters.SmoothingFilter(self.streamframe, 
														   smoothingBias, 
														   smoothingIterations)
		self.smoothedSf = filter.output()
		
	def poseSignal(self):
		"""This class returns the curvature measurement of the streamframe, the max curvature among the components is returned"""
		signal = []
		for i in range(len(self.smoothedSf)):
			k = self.smoothedSf.curvature(i)
			signal.append(max(k))
		return signal

	def eventSequence(self, includeBounds = True):
		"""Returns events for the local maxima of the streamframe.  Local maxima are per component"""
		signals = []
		for i in range(len(self.smoothedSf)):
			signals.append(self.smoothedSf.curvature(i))
		
		xMaxima = []
		yMaxima = []
		zMaxima = []
		
		# get all maxima
		if len(signals) > 1 and includeBounds:
			if signals[0][0] > signals[1][0]:
				xMaxima.append(0)
			if signals[0][1] > signals[1][1]:
				yMaxima.append(0)
			if signals[0][2] > signals[1][2]:
				zMaxima.append(0)
		for i in range(1, len(signals) - 1):
			if (signals[i][0] > signals[i - 1][0]) and (signals[i][0] > signals[i + 1][0]):
				xMaxima.append(i)
			if (signals[i][1] > signals[i - 1][1]) and (signals[i][1] > signals[i + 1][1]):
				yMaxima.append(i)
			if (signals[i][2] > signals[i - 1][2]) and (signals[i][2] > signals[i + 1][2]):
				zMaxima.append(i)
		if len(signals) > 1 and includeBounds:
			if signals[len(signals) - 2][0] < signals[len(signals) - 1][0]:
				xMaxima.append(len(signals) - 1)
			if signals[len(signals) - 2][1] < signals[len(signals) - 1][1]:
				yMaxima.append(len(signals) - 1)
			if signals[len(signals) - 2][2] < signals[len(signals) - 1][2]:
				zMaxima.append(len(signals) - 1)
		
		events = []
		for i in xMaxima:
			events.append(Events.StreamframeExtremaEvent(i, self.smoothedSf.times[i],
														 signals[i][0], 0))
		for i in yMaxima:
			events.append(Events.StreamframeExtremaEvent(i, self.smoothedSf.times[i],
														 signals[i][1], 1))
		for i in zMaxima:
			events.append(Events.StreamframeExtremaEvent(i, self.smoothedSf.times[i],
														 signals[i][2], 2))
		events.sort()
		
		return events
		
if __name__ == '__main__':
	sf = None
	sa = StreamframeAnalyzer(sf)
	print sa.poseSignal()
	print sa.eventSequence()