#  Copyright (c) 2008 patrick coleman

from Exceptions import *
from geom.Normal3 import Normal3

import math

# for test code
from geom.Point3 import Point3
from mmath.Vector3 import Vector3
from geom.BoundingBox import BoundingBox
from geom.Rotations import EulerRotation
from random import uniform

class Streamline(object):

	def __init__(self, points = [], times = []):
		self.points = points
		self.times = times
		self.normals = None # Needs more robust computation ported over from Curve.SampledSpline
		self.kNormals = None # Needs more robust computation ported over from Curve.SampledSpline
		self.kinks = None# Needs more robust computation ported over from Curve.SampledSpline
		self.kinkNormals = None # Needs more robust computation ported over from Curve.SampledSpline
		
		self._bbox = None
		self._bboxDirty = True
		
	def addPoint(self, time, point):
		self.points.append(point)
		self.times.append(time)
		self._bboxDirty = True
		
	def clear(self):
		self.points = []
		self.times = []
		self._bboxDirty = True
	
	def value(self, time):
		"""
			evaluates the sl at time t by piecewise linear interpolation
			first searches for a sample match, then looks for bounds
		"""
		
		# time is a sample
		if time in self.times:
			i = self.times.index(time)
			return self.points[i]
		
		#interp bewteen bounding samples
		pi, ni = self.boundTime(time)
		if pi < 0:
			return self.points[0]
		elif ni >= len(self.times):
			return self.points[len(self.points) - 1]
		else:
			u = float(time - self.times[pi]) / (self.times[ni] - self.times[pi])
			return self.points[pi] + u * (self.points[ni] - self.points[pi])
			
		
	def boundTime(self, time):
		"""
			returns (pi, ni) s.t. time is in (self.times[pi], self.times[ni]]
		"""
		if time <= self.times[0]:
			return (-1, 0)
		elif time > self.times[-1]:
			return (len(self.times) - 1, len(self.times))
		else:
			# binary search
			iStart = 0
			iEnd = len(self.times) - 1
			iMid = (iEnd + iStart) / 2

			while (iEnd - iStart) > 1:
				if self.times[iMid] < time:
					iStart = iMid
				else:
					iEnd = iMid
				iMid = (iEnd + iStart) / 2
			return (iStart, iEnd)
			
		
	def __len__(self):
		return len(self.points)

	def tangentDirection(self, i):
		"Returns a nomalized vector in the tangent direction"
		np = len(self.points)
		if i == 0:
			return (self.points[1] - self.points[0]).normalize()
		elif i == np - 1:
			return (self.points[np - 1] - self.points[np - 2]).normalize()
		else:
			return (self.points[i + 1] - self.points[i - 1]).normalize()
			
	def normal(self, i):
		"Returns a convex normal vector for point i"
		if self.normals == None:
			self._computeNormals()
		return self.normals[i]
	
	def curvatureNormal(self, i):
		"""Returns a curvature-scaled vector for point i"""
		if self.kNormals == None:
			self._computeKNormals()
		return self.kNormals[i]
	
	def kinkNormal(self, i):
		"""Returns a pseudo-curvature value which will be called 'kink'.  It is a nonlinear map
		from curvature such that curvature in [0, inf) maps to kink in [0, 1]"""
		if self.kinkNormals == None:
			self._computeKinkNormals()
		return self.kinkNormals[i]
	
	def kink(self, i):
		if self.kinks == None:
			self._computeKinks()
		return self.kinks[i]
				
	def boundingBox(self):
		if self._bboxDirty:
			self._computeBbox()
		return self._bbox
	
	def isDegenerate(self, tol = 0.0):
		if self._bboxDirty:
			self._computeBbox()
		return self._bbox.minPoint.distanceSquared(self._bbox.maxPoint) <= (tol * tol)
	
	def length(self):
		"Returns the euclidean length of the streamline"
		res = 0.0
		for i in range(1, len(self.points)):
			res += self.points[i].distance(self.points[i - 1])
		return res
		
	def _computeNormals(self):
		# TODO Needs more robust computation ported over from Curve.SampledSpline
		self.normals = []
		
		# zero normal at first point
		if len(self.points) > 0:
			self.normals.append(Normal3(0.0, 0.0, 0.0))
		
		# general normals as interpolated convex edge normals
		for i in range(1, len(self.points) - 1):
			av = self.points[i - 1] - self.points[i]
			bv = self.points[i + 1] - self.points[i]
			al = av.length()
			bl = bv.length()
			
			if al == 0.0 or bl == 0.0:
				self.normals.append(Normal3(0.0, 0.0, 0.0))
				continue
			
			side = bv ^ av
			an = av ^ side
			bn = side ^ bv
			if an.length() > 0.0:
				an = an.normalize()
			if bn.length() > 0.0:
				bn = bn.normalize()
			
			#interpolate face normals
			t = bl / (al + bl)
			n = an.interpolate(t, bn)
			if n.length() > 0.0:
				n.normalize()
			self.normals.append(Normal3(n.x, n.y, n.z))
		
		# zero normal at end
		if len(self.points) > 1:
			self.normals.append(Normal3(0.0, 0.0, 0.0))

	def _computeKNormals(self):
		#  TODO Needs more robust computation ported over from Curve.SampledSpline
		
		if self.normals == None:
			self._computeNormals()
		
		self.kNormals = []
		
		# zero kappa at start
		if len(self.points) > 0:
			self.kNormals.append(Normal3(0.0, 0.0, 0.0))
		
		# general curvature as weighted directional curvature per Taubin 95
		for i in range(1, len(self.points) - 1):
			av = self.points[i - 1] - self.points[i]
			bv = self.points[i + 1] - self.points[i]
			av2 = av.lengthSquared()
			bv2 = bv.lengthSquared()
			
			ka = 0.0
			kb = 0.0
			if av2 > 0.0:
				ka = math.fabs(2.0 * self.normals[i] *  av / av2)
			if bv2 > 0.0:
				kb = math.fabs(2.0 * self.normals[i] *  bv / bv2)
			
			# interp between kappa on neighbors normals; this gives a meaningful scalar inter, but
			# use tht local normal.  i'm sure there's a better way to do this. 
			kav = ka * self.normals[i + 1]
			kbv = kb * self.normals[i - 1]
			
			al = av.length()
			bl = bv.length()
			denom = al + bl
			if denom <= 0.0:
				self.kNormals.append(Normal3(0.0, 0.0, 0.0))
				continue
			t = bl / (al + bl)
			kv = kav.interpolate(t, kbv)
			self.kNormals.append(kv.length() * self.normal(i))
		
		# zero kappa at end
		if len(self.points) > 1:
			self.kNormals.append(Normal3(0.0, 0.0, 0.0))

	def _computeKinkNormals(self):
		# TODO Needs more robust computation ported over from Curve.SampledSpline
		
		if self.normals == None:
			self._computeNormals()
		
		self.kinkNormals = []
		
		# zero kink at start
		if len(self.points) > 0:
			self.kinkNormals.append(Normal3(0.0, 0.0, 0.0))
		
		# general curvature as weighted directional curvature per Taubin 95
		for i in range(1, len(self.points) - 1):
			av = self.points[i - 1] - self.points[i]
			bv = self.points[i + 1] - self.points[i]
			avLen = av.length()
			bvLen = bv.length()
			
			kinkA = 0.0
			kinkB = 0.0
			if avLen > 0.0:
				kinkA = math.fabs(self.normals[i] *  av / avLen)
			if bvLen > 0.0:
				kinkB = math.fabs(self.normals[i] *  bv / bvLen)
			
			# interp between kappa on neighbors normals; this gives a meaningful scalar inter, but
			# use tht local normal.  i'm sure there's a better way to do this. 
			kav = kinkA * self.normals[i - 1]
			kbv = kinkB * self.normals[i + 1]
			
			denom = avLen + bvLen
			if denom <= 0.0:
				self.kinkNormals.append(Normal3(0.0, 0.0, 0.0))
				continue
			t = bvLen / denom
			kv = kav.interpolate(t, kbv)
			self.kinkNormals.append(kv.length() * self.normal(i))
		
		# zero kink at end
		if len(self.points) > 1:
			self.kinkNormals.append(Normal3(0.0, 0.0, 0.0))
		
	def _computeKinks(self):
		# TODO Needs more robust computation ported over from Curve.SampledSpline
		
		if self.normals == None:
			self._computeNormals()
	
		self.kinks = []
		
		# zero kink at start
		if len(self.points) > 0:
			self.kinks.append(0.0)
		
		# kink based on weighted directional curvature per Taubin 95
		for i in range(1, len(self.points) - 1):
			av = self.points[i - 1] - self.points[i]
			bv = self.points[i + 1] - self.points[i]
			avLen = av.length()
			bvLen = bv.length()
			
			kinkA = 0.0
			kinkB = 0.0
			if avLen > 0.0:
				kinkA = math.fabs(self.normals[i] *  av / avLen)
			if bvLen > 0.0:
				kinkB = math.fabs(self.normals[i] *  bv / bvLen)
			
			# interp between kappa on neighbors normals; this gives a meaningful scalar inter, but
			# use tht local normal.  i'm sure there's a better way to do this. 
			kav = kinkA * self.normals[i - 1]
			kbv = kinkB * self.normals[i + 1]
			
			denom = avLen + bvLen
			if denom <= 0.0:
				self.kinks.append(0.0)
				continue
			t = bvLen / denom
			kv = kav.interpolate(t, kbv)
			self.kinks.append(kv.length())
		
		# zero kink at end
		if len(self.points) > 1:
			self.kinks.append(0.0)
		
	def _computeBbox(self):
		pMin = Point3()
		pMax = Point3()
		for i in range(len(self.points)):
			if i == 0:
				pMin = Point3(self.points[i].x, self.points[i].y, self.points[i].z)
				pMax = Point3(self.points[i].x, self.points[i].y, self.points[i].z)
			else:
				if self.points[i].x < pMin.x:
					pMin.x = self.points[i].x
				if self.points[i].y < pMin.y:
					pMin.y = self.points[i].y
				if self.points[i].z < pMin.z:
					pMin.z = self.points[i].z
				if self.points[i].x > pMax.x:
					pMax.x = self.points[i].x
				if self.points[i].y > pMax.y:
					pMax.y = self.points[i].y
				if self.points[i].z > pMax.z:
					pMax.z = self.points[i].z
		self._bbox = BoundingBox(pMin, pMax)
		self._bboxDirty = False
		
	def __repr__(self):
		return 'motion.Streamline.Streamline(' + \
			repr(self.points) + ',' + \
			repr(self.times) + ')'
		

class StreamlineCreator(object):
	def __init__(self, startTime, endTime, evaluator, step = 1):
		self.startTime = startTime
		self.endTime = endTime
		self.evaluator = evaluator
		self.step = step
	
	def streamline(self):
		s = Streamline()
		s.clear()
		
		t = self.startTime
		while t <= self.endTime:
			s.addPoint(t, self.evaluator.position(t))
			t += self.step

		return s


class Evaluator(object):
	""""Abstract base class for evaluating points.  Subclass and override
			position(), plug into the StreamlineCreator to get a Streamline"""
	def position(self, time):
		raise AbstractMethod()

class RotatingPointEvaluator(Evaluator):
	"""position evaluator that takes a motion, attribute, and local space point,
	and produces a streamline of that local space point undergoing the rotation.
	Only supports Vector3 rotation attributes, which are assumes to have an xyz 
	ordering and global axes""" 
	def __init__(self, motion, attribute, point = Point3(0.0, 0.0, 1.0), degrees = False):
		Evaluator.__init__(self)
		self.motion = motion
		self.attribute = attribute
		self.point = point
		self.degrees = degrees
	
	def position(self, time):
		attrVal = self.motion.attributeValue(self.attribute, time)
		if type(attrVal) != Vector3:
			raise UnsupportedData()
		rotation = EulerRotation(attrVal.x, attrVal.y, attrVal.z, 
								 order = 'xyz', local = False, 
								 degrees = self.degrees)
		return rotation(self.point)

def _testEval():
	class TestEval(Evaluator):
		def position(self, time):
			return Point3(time * time, time * time, time * time)
	sc = StreamlineCreator(1, 25, TestEval())
	sl = sc.streamline()
	print [t for t in sl.times]
	print [p for p in sl.points]

	print sl.isDegenerate(1080)
	print sl.isDegenerate(1081)
	
def _testNormals():
	sl = Streamline()
	p1 = Point3(-5, -5, -5)
	p2 = Point3(5, 5, 5)
	numPts = 51
	step = (p2 - p1) / float(numPts - 1)
	for i in range(numPts):
		if i in range(20, 30):
			sl.addPoint(i, p1 + i * step)
		else:
			ramt = 1
			sl.addPoint(i, p1 + i * step + Vector3(uniform(-ramt, ramt), 
				uniform(-ramt, ramt), uniform(-ramt, ramt)))
	for i in range(len(sl.points)):
		print i, sl.points[i], sl.normal(i), sl.curvatureNormal(i).length()

def _testKink():
	sl = Streamline()
	p1 = Point3(-5, -5, -5)
	p2 = Point3(5, 5, 5)
	numPts = 51
	step = (p2 - p1) / float(numPts - 1)
	for i in range(numPts):
		if i in range(20, 30):
			sl.addPoint(i, p1 + i * step)
		else:
			ramt = 1
			sl.addPoint(i, p1 + i * step + Vector3(uniform(-ramt, ramt), 
				uniform(-ramt, ramt), uniform(-ramt, ramt)))
	for i in range(len(sl.points)):
		print i, sl.points[i], sl.normal(i), sl.kinkNormal(i).length(), sl.kinkNormal(i).length()

def _testBounds():
	numPts = 20
	dv = 0.1
	a = [0.1 * i for i in range(numPts)]
	sl = Streamline(a, a)
	values = [0.5, 0.55, 0.5999, 0.0, -0.1, a[-1] + 0.1]
	for v in values:
		bounds = sl.boundTime(v)
		times = [None, None]
		if bounds[0] >= 0:
			times[0] = sl.times[bounds[0]]
		if bounds[1] < len(sl.times):
			times[1] = sl.times[bounds[1]]
		
		print v, bounds, times

def _testValue():
	numPts = 20
	dv = 0.1
	a = [0.1 * i for i in range(numPts)]
	sl = Streamline(a, a)
	values = [0.5, 0.55, 0.5999, 0.0, -0.1, a[-1] + 0.1]
	for v in values:
		print v, sl.value(v)
	
		
	
	


if __name__ == '__main__':
	_testValue()
	#_testBounds()
	#pass
	#_testEval()
	#_testNormals()
	#_testKink()