#  Copyright (c) 2008 patrick coleman

import random

from Exceptions import *
class BadArgumentData(Exception): pass

import Streamline
import geom.Point3

class StreamlineFilter(object):
	def __init__(self, streamline):
		self.streamline = streamline
	
	def output(self):
		raise AbstractMethod()
	
class SmoothingFilter(StreamlineFilter):
	def __init__(self, streamline, bias = 0.5, iterations = 1):
		StreamlineFilter.__init__(self, streamline)
		self.bias = bias
		self.iterations = iterations
	
	def output(self):
		last = self.streamline
		if self.iterations == 0:
			return Streamline.Streamline(last.points[:], last.times)
			
		for i in range(self.iterations):
			smoothed = Streamline.Streamline(last.points[:], last.times)
			for j in range(1, len(last) - 1):
				smoothed.points[j] = (self.bias * last.points[j] + 
					0.5 * (1.0 - self.bias) * (last.points[j - 1] + last.points[j + 1]))
			last = smoothed
			
		return last

class TentFilter(StreamlineFilter):
	def __init__(self, streamline, filterWidth):
		"filter width must be odd and positive"
		if filterWidth <= 0 or filterWidth % 2 != 1:
			raise BadArgumentData()
		StreamlineFilter.__init__(self, streamline)
		self.filterWidth = filterWidth
		self.halfWidth = (self.filterWidth - 1) / 2
	
	def output(self):
		if self.filterWidth == 1:
			return Streamline.Streamline(self.streamline.points[:], self.streamline.times)
		else:
			res = []
			for i in range(len(self.streamline)):
				offset = min(i, self.halfWidth)
				mask = self._weightMask(i)
				resp = geom.Point3.Point3(0.0, 0.0, 0.0)
				
				for j in range(-self.halfWidth, self.halfWidth + 1):
					si = i + j
					if si >= 0 and si < len(self.streamline):
						resp += mask[j + offset] * self.streamline[si]
				res.append(resp)
			return res
		
		
	def _weightMask(self, index):
		"returns a list of length halfWidth + 1 that lerps from 1 down to 0 jsut past bounds"
		
		
		# construct basic tent mask
		res = [1.0]
		for i in range(1, self.halfWidth + 1):
			v = 1.0 - float(i) / (self.halfWidth + 1)
			res = [v] + res + [v]
		
		# clip off based on signal
		res = res[max(self.halfWidth - index, 0):self.filterWidth - \
			max(0, self.halfWidth - len(self.streamline) + index + 1)]
		
		#normalize
		invDenom = 1.0 / sum(res)
		res = [r * invDenom for r in res]
		
		return res

if __name__ == '__main__':
	numPoints = 10
	limit = 10
	iters = 100
	
	#initialize a random signal
	sl = Streamline.Streamline()
	for i in range(numPoints):
		sl.addPoint(i, random.uniform(0, limit))
	print 'initial sl:'
	for i in range(len(sl)):
		print sl.times[i], ':', sl.points[i]
		
	# test
	sf = SmoothingFilter(sl, iterations = iters)
	fsl = sf.output()
	print 'smoothed sl:'
	for i in range(len(fsl)):
		print fsl.times[i], ':', fsl.points[i]
	