#  Copyright (c) 2008 patrick coleman

from Exceptions import *
import random

class NumericFilter(object):
	def __init__(self, signal, zero = 0.0):
		self.signal = signal
		self.zero = zero
		
	def output(self):
		raise AbstractMethod()

class SmoothingFilter(NumericFilter):
	def __init__(self, signal, bias = 0.5, iterations =1, zero = 0.0):
		NumericFilter.__init__(self, signal, zero)
		self.bias = bias
		self.iterations = iterations
	
	def output(self):
		last = self.signal
		
		for i in range(int(self.iterations)):
			smoothed = last[:]
			for j in range(1, len(smoothed) - 1):
				smoothed[j] = (self.bias * last[j] + 
					0.5 * (1.0 - self.bias) * (last[j - 1] + last[j + 1]))
			last = smoothed
		
		return last[:]

class TentFilter(NumericFilter):
	def __init__(self, signal, filterWidth, zero = 0.0):
		"filter width must be odd and positive"
		if filterWidth <= 0 or filterWidth % 2 != 1:
			raise BadArgumentData()
		NumericFilter.__init__(self, signal, zero)
		self.filterWidth = filterWidth
		self.halfWidth = (self.filterWidth - 1) / 2
	
	def output(self):
		if self.filterWidth == 1:
			return self.signal[:]
		else:
			res = []
			for i in range(len(self.signal)):
				offset = min(i, self.halfWidth)
				mask = self._weightMask(i)
				resp = self.zero
				
				for j in range(-self.halfWidth, self.halfWidth + 1):
					si = i + j
					if si >= 0 and si < len(self.signal):
						resp += mask[j + offset] * self.signal[si]
				res.append(resp)
			return res
		
		
	def _weightMask(self, index):
		"returns a list of length halfWidth + 1 that lerps from 1 down to 0 jsut past bounds"
		
		
		# construct basic tent mask
		res = [1.0]
		for i in range(1, self.halfWidth + 1):
			v = 1.0 - float(i) / (self.halfWidth + 1)
			res = [v] + res + [v]
		
		# clip off based on signal
		res = res[max(self.halfWidth - index, 0):self.filterWidth - \
			max(0, self.halfWidth - len(self.signal) + index + 1)]
		
		#normalize
		invDenom = 1.0 / sum(res)
		res = [r * invDenom for r in res]
		
		return res

class DerivativeFilter(object):
	def __init__(self, signal, zero = 0.0):
		self.signal = signal
		self.zero = zero
		
	def output(self):
		res = []
		l = len(self.signal)
		
		if l == 1:
			return [zero]
		
		if l > 1:
			res.append(self.signal[1] - self.signal[0])
		
		for i in range(1, l - 1):
			res.append(0.5 * (self.signal[i + 1] - self.signal[i - 1]))
		
		if l > 2:
			res.append(self.signal[l - 1] - self.signal[l - 2])
		
		return res

class MaxFilter(NumericFilter):
	def __init__(self, signal, filterWidth, zero = 0.0):
		"filter width must be odd and positive"
		if filterWidth <= 0 or filterWidth % 2 != 1:
			raise BadArgumentData()
		NumericFilter.__init__(self, signal, zero)
		self.filterWidth = filterWidth
		self.halfWidth = (self.filterWidth - 1) / 2
	
	def output(self):
		if self.filterWidth == 1:
			return self.signal[:]
		elif len(self.signal) <= self.halfWidth:
			return [max(self.signal)] * len(self.signal)
		else:
			res = []
			for i in range(0, self.halfWidth):
				res.append(max(self.signal[0:i + self.halfWidth + 1]))
			for i in range(self.halfWidth, len(self.signal) - self.halfWidth):
				res.append(max(self.signal[i - self.halfWidth:i + self.halfWidth + 1]))
			for i in range(len(self.signal) - self.halfWidth, len(self.signal)):
				res.append(max(self.signal[i - self.halfWidth:]))
			return res

def testMaxFilter():
	numPts = 1000
	#import random
	#import mmaya
	#import math
	#signal = [math.sin(0.25 * i) + 0.25 * random.random() for i in range(numPts)]
	#maxResp = MaxFilter(signal, 3).output()
	#mmaya.plotSignal(signal, 'randomSig')
	#mmaya.plotSignal(maxResp, 'maxFilter')
	#print len(signal), len(maxResp)
	



if __name__ == '__main__':
	numPts = 10
	iters = 100
	limit = 10
	
	signal = []
	for i in range(numPts):
		signal.append(random.uniform(0, limit))
	print 'initial signal:'
	for i in signal:
		print i
	
	#sf = SmoothingFilter(signal, iterations = iters)
	tf1 = TentFilter(signal, 1)
	tf3 = TentFilter(signal, 3)
	tf5 = TentFilter(signal, 5)
	tf7 = TentFilter(signal, 7)
	
	#import mmaya
	#mmaya.plotSignal(signal, name = 'sig')
	#mmaya.plotSignal(tf1.output(), name = 'tf1')
	#mmaya.plotSignal(tf3.output(), name = 'tf3')
	#mmaya.plotSignal(tf5.output(), name = 'tf5')
	#mmaya.plotSignal(tf7.output(), name = 'tf7')
	
	smoothed = tf.output()
	print 'smoothed signal:'
	for i in smoothed:
		print i
	