#  Copyright (c) 2008 patrick coleman

import math
from maya import cmds as mc
import maya.OpenMaya as om
import maya.OpenMayaAnim as oma
from mmath.Vector3 import Vector3
import mmaya.Node
import geom.Rotations

import Maya
import motion.Motions as Motions
import time

class UnknownDataType(Exception): pass
class NoUniqueNode(Exception): pass

class Character(object):
	def __init__(self, name = u"unknown", root = u"unknown", attributes = []):
		self.name = name
		self.root = root
		self.attributes = attributes
		self.poses = []
		self.motions = []
	
	
	def sampleMotion(self, startTime = Maya.sceneStartTime(), endTime = Maya.sceneEndTime()):
		"""creates a sampled motion from current motion and stores at end of character motions
		stores QRotations as of jan 20th 08"""
		
		sm = Motions.SampledMotion(self.name + 'SM', startTime, endTime, self.attributes)
		sm.clearSamples()
		
		for i in range(sm.numSamples()):
			t = sm.sampleTime(i)
			for attr in self.attributes:
				# this assumes translation or rotation, which are a list of 1 tuple 
				v = mc.getAttr(attr, time = t)
				value = Vector3(v[0][0], v[0][1], v[0][2])
				
				n, a = attr.split('.')
				if a == 'rotate':
					order = mmaya.rotateOrderString(mc.getAttr(n + '.rotateOrder'))
					er = geom.Rotations.EulerRotation(value.x, value.y, value.z,
														local = False, order = order, degrees = True)
					value = er.qRotation()
				sm.appendSample(attr, value)
		self.motions.append(sm)
		
	def applyMotion(self, motion, sceneBounds = True, interactive = True):
		
		for attr in self.attributes:
			for compAttr in Maya.componentAttributes(attr):
				oldAnimNode = Maya.sourceNode(compAttr)
				if oldAnimNode != None and not interactive:
					mc.delete(oldAnimNode)
		
		times = None
		if sceneBounds:
			times = range(int(mmaya.sceneStartTime()), int(mmaya.sceneEndTime()) + 1)
		else:
			times = range(int(math.floor(motion.startTime)), int(math.ceil(motion.endTime)) + 1)
		
		for t in times:
			if interactive:
				mc.currentTime(t)
		
			for attr in self.attributes:
				n, a = attr.split('.')
				val = motion.attributeValue(attr, t)
				compattrs = Maya.componentAttributes(attr)
				
				vType = type(val)
				if vType is Vector3:
					# bug fix, this could be better
					if a == 'rotate':
						order = mmaya.rotateOrderString(mc.getAttr(n + '.rotateOrder'))
						if order != 'xyz':
							er = geom.Rotations.EulerRotation(val.x, val.y, val.z,
														local = False, order = 'xyz', degrees = True)
							erReal = er.eulerRotation(local = False, order = order)
							val.x = math.degrees(erReal.rx)
							val.y = math.degrees(erReal.ry)
							val.z = math.degrees(erReal.rz)
							
					for i in range(3):
						mc.setKeyframe(compattrs[i], time = t, value = val[i])
				elif vType is geom.Rotations.QRotation:
					order = mmaya.rotateOrderString(mc.getAttr(n + '.rotateOrder'))
					eRot = val.eulerRotation(local = False, order = order)
					rVals = [math.degrees(eRot.rx), math.degrees(eRot.ry), math.degrees(eRot.rz)]
					for i in range(3):
						mc.setKeyframe(compattrs[i], time = t, value = rVals[i])
				else:
					raise UnknownDataType()
	
	def apiApplyMotion(self, motion):
		"""time is wrong on this, but it seems to be python conversion/interp bound, so i'm 
		leaving it broken today.  check the MTime units if looking to fix"""
		for attr in self.attributes:
			for compAttr in Maya.componentAttributes(attr):
				oldAnimNode = Maya.sourceNode(compAttr)
				if oldAnimNode != None:
					mc.delete(oldAnimNode)
		times = range(int(math.floor(motion.startTime)), int(math.ceil(motion.endTime)) + 1)
		mtimes = om.MTimeArray()
		for t in times:
			mtimes.append(om.MTime(t))
		
		for attr in self.attributes:
			compattrs = Maya.componentAttributes(attr)
			
			xValues = om.MDoubleArray()
			yValues = om.MDoubleArray()
			zValues = om.MDoubleArray()
			
			for t in times:
				val = motion.attributeValue(attr, t)
				vType = type(val)
				if vType is Vector3:
					xValues.append(val.x)
					yValues.append(val.y)
					zValues.append(val.z)
				elif vType is geom.Rotations.QRotation:
					eRot = val.eulerRotation(local = False, order = 'xyz')
					xValues.append(eRot.rx)
					yValues.append(eRot.ry)
					zValues.append(eRot.rz)
				else:
					raise UnknownDataType()
			compValues = [xValues, yValues, zValues]
			
			for i in range(3):
				locked = mc.getAttr(compattrs[i], lock = True)
				if not locked:
					curveObj, curveFs = self._apiCreateAnimCurveObj(compattrs[i])

					 # TODO CONTINUE HERE
					 # todo set value for each time
					curveFs.addKeys(mtimes, compValues[i])
					
					
					
		
	
	def backupMotion(self):
		"backs up current anim curves/values to backup attributes"
		
		self._verifyBackupAttributes()
		
		for attr in self.attributes:
			for compAttr in Maya.componentAttributes(attr):
				backupAttr = self._backupAttribute(compAttr)
				
				oldBackupNode = Maya.sourceNode(backupAttr)
				if oldBackupNode != None:
					mc.delete(oldBackupNode)
				
				animNode = Maya.sourceNode(compAttr)
				if animNode != None:
					newAnimNode = mc.duplicate(animNode)
					mc.connectAttr(newAnimNode[0] + ".output", backupAttr)
				else:
					mc.setAttr(backupAttr, mc.getAttr(compAttr))
	
	def applyBackupMotion(self):
		"applies the stored backup motion to the character" 
		
		self._verifyBackupAttributes()
		
		for attr in self.attributes:
			for compAttr in Maya.componentAttributes(attr):
				backupAttr = self._backupAttribute(compAttr)
				
				oldAnimNode = Maya.sourceNode(compAttr)
				if oldAnimNode != None:
					mc.delete(oldAnimNode)
				
				animNode = Maya.sourceNode(backupAttr)
				if animNode != None:
					newAnimNode = mc.duplicate(animNode)
					mc.connectAttr(newAnimNode[0] + ".output", compAttr)
				else:
					locked = False
					if mc.getAttr(compAttr, lock = True):
						locked = True
						mc.setAttr(compAttr, lock = False)
					mc.setAttr(compAttr, mc.getAttr(backupAttr))
					if locked:
						mc.setAttr(compAttr, lock = True)
	
	def _verifyBackupAttributes(self):
		"checks for presence of all backup attributes"
		
		for attr in self.attributes:
			nodename, attrname = self._backupAttribute(attr).split('.')
			if not mc.attributeQuery(attrname, node = nodename, exists = True):
				mc.addAttr(nodename, 
						   longName = attrname, 
						   numberOfChildren = 3, 
						   attributeType = 'double3')
						   
				comptype = 'double'
				if attrname == 'o_rotate':
					comptype = 'doubleAngle'
					
				for suffix in ('X', 'Y', 'Z'):
					mc.addAttr(nodename,
							   longName = attrname + suffix, 
							   attributeType = comptype, 
							   parent = attrname)
	
	def _backupAttribute(self, attribute):
		"returns the backup attribute name"
		return attribute.replace('.', '.o_')
		
	def attributeDepth(self, attribute):
		"""Returns the depth of the attribute in the character hierarchy, root is 0"""
		return mmaya.Node.Node(attribute.split('.')[0]).depth()
	
	def descendants(self):
		"Returns descendent node names as list of strings"
		return mmaya.Node.Node(self.root).descendants()
	
	def nodes(self):
		selfNode = mmaya.Node.Node(self.root)
		return [selfNode] + selfNode.descendants()
		
	def _apiCreateAnimCurveObj(self, attr):
		node, attr = attr.split('.')
	
		# get a selection list with the specificed node
		selList = om.MSelectionList()
		om.MGlobal.getSelectionListByName(node, selList)
		if selList.length() != 1:
			raise NoUniqueNode()
		
		# get the node object
		nodeObj = om.MObject()
		selList.getDependNode(0, nodeObj)
		
		# get the attr obj
		depNodeFs = om.MFnDependencyNode(nodeObj)
		attrObj = depNodeFs.attribute(attr)
		
		# create an anim curve for it
		animCurveFs = oma.MFnAnimCurve()
		animCurveObj = animCurveFs.create(nodeObj, attrObj)
		
		return animCurveObj, animCurveFs
	